/****************************************************************************
 *
 *  Copyright (c) 2002 by LSI Logic Corporation, all rights reserved.
 *
 *                         LSI Logic confidential
 *  The information contained herein is proprietary of LSI Logic Corporation.
 *
 *  This software is provided "as is" without any express or implied warranty.
 *  LSI Logic disclaims all representations and warranties with regard to
 *  this software, including all implied warranties of merchantability and
 *  fitness for a particular purpose. In no event shall LSI Logic be liable
 *  for any special, indirect or consequential damages or any damages
 *  whatsoever resulting from loss of use, data or profits, whether in an
 *  action of contract, negligence or other tortious act or omission,
 *  arising out of or in connection with the use or performance of this
 *  software.
 *
 *  : $
 *
 *  Revision Details
 *  :  $
 *
 *
 *  File Description:
 *
 *
 ****************************************************************************/
 
/****************************************************************************
 * standard net-snmp include files
 ****************************************************************************/
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include "../util_funcs.h"

/****************************************************************************
 * standard linux include files
 ****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#include <pwd.h>
#include <syslog.h>
#include <sys/stat.h>
#include <crypt.h>

/****************************************************************************
 * module include files
 ****************************************************************************/
#include "lsilAdslMib.h"

/****************************************************************************
 * shared libs project include files
 ****************************************************************************/

/****************************************************************************
 * private routines
 ****************************************************************************/
static void syncPoolHandler(unsigned int clientreg, void *clientarg);

/****************************************************************************
 * public routines
 ****************************************************************************/
inline int startSyncPool (int nsec);
inline void stopSyncPool (void);

/****************************************************************************
 * extern routines
 ****************************************************************************/
extern void natGroup_Sync(void);
extern void firewallGroup_Sync(void);

/****************************************************************************
 * private variables 
 ****************************************************************************/
 
static int syncPoolControl = 1;

static void (*FuncSuccessHandler)(void) = NULL;
static void (*FuncErrorHandler)(void) = NULL;

/****************************************************************************
 * public (exported) variables
 ****************************************************************************/

/****************************************************************************
 * Function Name  : init_lsilAdslMib
 * Arguments      :
 * Returns        : 
 * Variables      : 
 * Modifies       :  
 * Description    : Initializes the lsilAdslMib MIB modules
 * +--lsiAdslMib(100)
 *   +--lsiAdslObjects(1)
 *   |  +--systemMgmt(1)
 *   |  |  +--system(1)
 *   |  |  +--config(2)
 *   |  |  +--sysResources(3)
 *   |  |  +--fileTransfer(4)
 *   |  +--networkServices(2)
 *   |     +--adsl(1)
 *   |     +--networkConnections(2)
 *   |     |  |
 *   |     |  +-- -R-- INTEGER   activeConnections(1)
 *   |     |  +-- -RW- INTEGER   maxActiveConnections(2)
 *   |     |  |        Range: 1..8
 *   |     |  +-- -R-- INTEGER   ethernetInterfaces(3)
 *   |     |  +-- -R-- INTEGER   bridgeInterfaces(4)
 *   |     |  +-- -R-- INTEGER   ipoaInterfaces(5)
 *   |     |  +-- -R-- INTEGER   pppoaInterfaces(6)
 *   |     |  +-- -R-- INTEGER   pppoeInterfaces(7)
 *   |     |  |
 *   |     |  +--bridgeConfigTable(8)
 *   |     |  +--bridgeEtherPortTable(9)
 *   |     |  +--bridgeAtmPortTable(10)
 *   |     |  +--ipoaConfigTable(11)
 *   |     |  +--pppoaConfigTable(12)
 *   |     |  +--pppoeConfigTable(13)
 *   |     +--firewall(3)
 *   |     +--nat(4)
 *   |     +--dhcpServer(5)
 *   |     +--dnsServer(6)
 ****************************************************************************/

#ifdef INCLUDE_LSIL_ADSL_MIB

void
init_lsilAdslMib(void)
{
#ifdef USING_MIBII_SYSORTABLE_MODULE
    static oid      reg[] = { 1, 3, 6, 1, 4, 1, 1372, 100 };
    register_sysORTable(reg, 8, "The MIB module for managing LSI-Logic Adsl devices.");
#endif

	   /* systemMgmt modules */
	  init_system();
	  init_config();
	  init_sysResources();
	  init_fileTransfer();

	  /* networkServices modules */
	  init_adsl();
	  init_networkConnections();
	  init_firewall();
	  init_nat();
	  init_dhcpServer();
	  init_dnsServer();
	  init_dhcpRelay();	  
}
#endif

/***************************************************************************
 * POOLING SYSTEM/AGENT SYNC ROUTINES
 ***************************************************************************/

/****************************************************************************
 * Function Name  : syncPoolHandler
 * Arguments      :
 * Returns        : 
 * Variables      : 
 * Modifies       :  
 * Description    : Handler that will be called every xx seconds that will 
 *                  call sync handlers between system and Enterpise MIB 
 *                  objects 
 ****************************************************************************/
void 
init_syncMibPooling(int cache_time)/* daft Portugese - you mean "polling" */
{
	  /* start sync Pool Handler */
#ifdef BEFORE_BEN
	  startSyncPool(cache_time);
#endif
}

#ifndef BEFORE_BEN
void do_mib_sync(void)
{
	/*
	 * MIB-II Sync handlers
	 */
#ifdef IPFORWARD_MIB_MIBII
	ipFoward_Sync();
#endif

	/*
	 * Call ATM-MIB Sync Handlers
	 */
#ifdef INCLUDE_ATM_MIB
	sarDriver_AtmMIB_sync();
#endif
	  
	/*
	 * Call ILMI-MIB Sync Handlers
	 */
#ifdef INCLUDE_ILMI_MIB
	ilmiMIB_Sync();
#endif
    		    
	/*
	 * Call Enterpise MIB SYS/AGENT Sync Handlers
	 */
#ifdef INCLUDE_LSIL_ADSL_MIB
	adslGroup_sync();   
	networkConnectionsSync();
	firewallGroup_Sync();
	natGroup_Sync();    
	dhcpServerGroup_Sync();
	dnsServerGroup_Sync();
	dhcpRelayGroup_Sync();    
#endif /*INCLUDE_LSIL_ADSL_MIB*/     
}
#endif

/****************************************************************************
 * Function Name  : syncPoolHandler
 * Arguments      :
 * Returns        : 
 * Variables      : 
 * Modifies       :  
 * Description    : Handler that will be called every xx seconds that will 
 *                  call sync handlers between system and Enterpise MIB 
 *                  objects 
 ****************************************************************************/
static void 
syncPoolHandler(unsigned int clientreg, void *clientarg)
{
    /*DEBUGMSGTL(("lsilAdslMib", "syncPoolHandler\n"));*/
    
    if(syncPoolControl)
    {
      /* restart timer */
	  	if( !snmp_alarm_register(*((int*)clientarg), 0, syncPoolHandler, clientarg ))
      {
        DEBUGMSGTL(("lsilAdslMib", "syncPool -> snmp_alarm_register Error\n"));
      }
    }
    else
    {
	    DEBUGMSGTL(("config", "Pool canceled\n"));
	    return;
    }

#ifdef BEFORE_BEN
	/*
	 * MIB-II Sync handlers
	 */
#ifdef IPFORWARD_MIB_MIBII
	ipFoward_Sync();
#endif

	/*
	 * Call ATM-MIB Sync Handlers
	 */
#ifdef INCLUDE_ATM_MIB
	sarDriver_AtmMIB_sync();
#endif
	  
	/*
	 * Call ILMI-MIB Sync Handlers
	 */
#ifdef INCLUDE_ILMI_MIB
	ilmiMIB_Sync();
#endif
    		    
	/*
	 * Call Enterpise MIB SYS/AGENT Sync Handlers
	 */
#ifdef INCLUDE_LSIL_ADSL_MIB
	adslGroup_sync();   
	networkConnectionsSync();
	firewallGroup_Sync();
	natGroup_Sync();    
	dhcpServerGroup_Sync();
	dnsServerGroup_Sync();
	dhcpRelayGroup_Sync();    
#endif /*INCLUDE_LSIL_ADSL_MIB*/     
#else
		do_mib_sync();
#endif
}

/****************************************************************************
 * Function Name  : startSyncPool
 * Arguments      : int nsec -> number of seconds to next Sync
 * Returns        : -1 on error
 *                   0 on success
 * Variables      : 
 * Modifies       :  
 * Description    : routine that will perform a sync
 ****************************************************************************/
inline int 
startSyncPool (int nsec)
{
#if 0
    static int poolTime;

    DEBUGMSGTL(("lsilAdslMib", "startSyncPool Pool Time -> %d seconds\n", nsec));
    
    poolTime=nsec;
        
		if( !snmp_alarm_register(poolTime, 0, syncPoolHandler, (void *)&poolTime ))
			return -1;

    syncPoolControl=1;
#endif
    return 0;
}

/****************************************************************************
 * Function Name  : stopSyncPool
 * Arguments      : int nsec -> number of seconds to next Sync
 * Returns        : -1 on error
 *                   0 on success
 * Variables      : 
 * Modifies       :  
 * Description    : routine that will perform a sync
 ****************************************************************************/
inline void
stopSyncPool (void)
{
    DEBUGMSGTL(("lsilAdslMib", "stopSyncPool\n"));
    syncPoolControl=0;  
}


/***************************************************************************
 * SYSTEM/AGENT UTIL ROUTINES
 ***************************************************************************/
/****************************************************************************
 * Function Name  : cleanNonValidEntriesTable
 * Arguments      :
 * Returns        : 
 * Variables      : 
 * Modifies       :  
 * Description    : handler for the get of pppoaInterfaces scalar
 *                  contents and how it's structured
 ****************************************************************************/
void 
cleanNonValidTableEntries(netsnmp_table_data_set *table_set,
                          netsnmp_table_row **validRows, int rowsToTest)
{
    netsnmp_table_row *row, *auxRow;
    int i, foundRow;

    /* clean any non-valid row */
    for( row=table_set->table->first_row ; row ; )
    {
      foundRow=0;
      for (i=0; i<rowsToTest; i++)
        if(validRows[i]==row)
	        foundRow=1;

      if(!foundRow)
      {
        /* delete row */
      	auxRow=row;
        row=row->next;
        DEBUGMSGTL(("lsilAdslMib", "deleting row \n"));
     	  netsnmp_table_data_remove_and_delete_row(table_set->table, auxRow);      
      }
      else
        row=row->next;
    }
}

/****************************************************************************
 * Function Name  : sig_chld
 * Arguments      : filename
 * Returns        : int
 * Variables      :
 * Modifies       :
 * Description    : The signal handler function -- only gets called when a 
 *                  SIGCHLD is received, ie when a child terminates
 ****************************************************************************/
void sig_chld(int signo)
{
    int status, child_val;


    DEBUGMSGTL(("lsilAdslMib", "sig_chld\n"));

    /* Wait for any child without blocking */
    if (waitpid(-1, &status, WNOHANG) < 0) 
    {
        /*
         * calling standard I/O functions like fprintf() in a 
         * signal handler is not recommended, but probably OK 
         * in toy programs like this one.
         */
      DEBUGMSGTL(("lsilAdslMib", "sig_chld - waitpid failed\n"));
      return;
    }

    /*
     * We now have the info in 'status' and can manipulate it using
     * the macros in wait.h.
     */
    if (WIFEXITED(status))                 /* did child exit normally? */
    {
      child_val = WEXITSTATUS(status);     /* get child's exit status */
      DEBUGMSGTL(("lsilAdslMib", 
                  "sig_chld - child's exited normally with status %d\n", child_val));
	
	    switch(child_val)
	    {
	      case 127:
  	      DEBUGMSGTL(("lsilAdslMib", "sig_chld - FAILED\n"));
  	      if(FuncErrorHandler!=NULL)
	        {
		        FuncErrorHandler();	      
			      FuncErrorHandler=NULL; FuncSuccessHandler=NULL;
			    }
	        break;
	      case 126:
  	      DEBUGMSGTL(("lsilAdslMib", "sig_chld - SUCCESS\n"));
  	      if(FuncSuccessHandler!=NULL)
	        {
		        FuncSuccessHandler();	      
			      FuncErrorHandler=NULL; FuncSuccessHandler=NULL;
			    }
	        break;
      }    
    }
}

/****************************************************************************
 * Function Name  : execRoutineNoWait
 * Arguments      : filename
 * Returns        : int
 * Variables      :
 * Modifies       :
 * Description    : Fork and exec. Ripped from man system.
 ****************************************************************************/
int
execRoutineNoWait(int (*handler)(void),
                  void (*successHandler)(void),
		              void (*failHandler)(void))
{
    struct sigaction act;
    pid_t pid;

    DEBUGMSGTL(("lsilAdslMib", "fork_handler_noWait\n"));

    /* Assign sig_chld as our SIGCHLD handler */
    act.sa_handler = sig_chld;

    /* We don't want to block any other signals in this example */
    sigemptyset(&act.sa_mask);

    /*
     * We're only interested in children that have terminated, not ones
     * which have been stopped (eg user pressing control-Z at terminal)
     */
    act.sa_flags = SA_NOCLDSTOP;

    /*
     * Make these values effective. If we were writing a real 
     * application, we would probably save the old value instead of 
     * passing NULL.
     */
    if (sigaction(SIGCHLD, &act, NULL) < 0) 
    {
      DEBUGMSGTL(("lsilAdslMib", "fork_handler_noWait - sigaction failed\n"));
      return 0;
    }
    
    /**/
    FuncSuccessHandler = successHandler;
    FuncErrorHandler = failHandler;
    
    
    /* Fork */
    switch (pid = fork())
    {
      case -1:
        DEBUGMSGTL(("lsilAdslMib", "fork_handler_noWait - fork failed\n"));
        return 0;

      case 0:                         /* child -- finish straight away */
        if(handler())
          _exit(126);                 /* exit status = 126 SUCCESS */
	      else
          _exit(127);                 /* exit status = 127 ERROR */
	  
      default:                        /* parent */
        return 1;
    } 
}

/****************************************************************************
 * Function Name  : execLineCommand
 * Arguments      : filename
 * Returns        : int
 * Variables      :
 * Modifies       :
 * Description    : Fork and exec. Ripped from man system.
 ****************************************************************************/
int 
execLineCommand(char *command)
{
    struct extensible extmp;
    int ret=1;
   
    sprintf(extmp.command, "%s", command);	    
    extmp.type = EXECPROC;
    extmp.next = NULL;
    extmp.output[0] = '\0';

    #ifdef NETSNMP_FOR_TARGET
    ret=exec_command(&extmp);
    ret=(ret==0)?1:0;
    #endif
    
    DEBUGMSGTL(("lsilAdslMib\n",
                "execLineCommand command[ %s ]\n"
		            "                output [ %s ]\n"
		            "                return [ %d ]\n", extmp.command, extmp.output, ret));

    return ret;
}

/****************************************************************************
 * Function Name  : snmp_oid_min_compare
 * Arguments      : ...
 * Returns        : -1 if name1 < name2
 *                   0 if name1 = name2
 *                   1 if name1 > name2
 * Variables      : 
 * Modifies       :  
 * Description    : Lexicographical compare two object identifiers.
 *                  This method differs from snmp_oid_compare in that the 
 *                  comparison stops at the length of the smallest 
 *                  object identifier.
 ****************************************************************************/
int
snmp_oid_min_compare(const oid * in_name1,
                     size_t len1, const oid * in_name2, size_t len2)
{
    register int    len;
    register const oid *name1 = in_name1;
    register const oid *name2 = in_name2;

    /*
     * len = minimum of len1 and len2 
     */
    if (len1 < len2)
        len = len1;
    else
        len = len2;
    /*
     * find first non-matching OID 
     */
    while (len-- > 0) 
    {
      /*
       * these must be done in seperate comparisons, since
       * subtracting them and using that result has problems with
       * subids > 2^31. 
       */
      if (*(name1) < *(name2))
        return -1;
      if (*(name1++) > *(name2++))
        return 1;
    }
    /*
     * both OIDs equal up to length of shorter OID 
     */

    return 0;
}

/****************************************************************************
 * Function Name  : strMacToMac
 * Arguments      : 
 * Returns        : 0 if no modification in the file
 *                  1 if modifications in the file
 * Variables      : 
 * Modifies       :  
 * Description    : converts a string xx:xx:xx:xx:xx:xx MAC address into a
 *                  array[6]
 ****************************************************************************/
void
strMacToMac(char *strMac, char *mac)
{
    unsigned int macAddress_t[6];

    sscanf(strMac,"%2x:%2x:%2x:%2x:%2x:%2x",
                   &macAddress_t[0], &macAddress_t[1], &macAddress_t[2],
		               &macAddress_t[3], &macAddress_t[4], &macAddress_t[5] );

    memcpy(mac, macAddress_t, 6);
}


/****************************************************************************
 * Function Name  : fileChanged
 * Arguments      : 
 * Returns        : 0 if no modification in the file
 *                  1 if modifications in the file
 * Variables      : 
 * Modifies       :  
 * Description    :  to be merged in the connections_config.c
 ****************************************************************************/
int 
fileChanged(char *file, time_t *lastConnectionsChange)
{
  struct stat testFile;

  if( lstat(file, &testFile)==0 &&
      !difftime(*lastConnectionsChange, testFile.st_mtime) )
  {
    return 0;
  }

  *lastConnectionsChange = testFile.st_mtime;
  return 1;
}



/****************************************************************************
 * Function Name  : setStateServiceServer
 * Arguments      : int state -> 1 enable
 *                               0 disable
 * Returns        : int          1 OK
 *                               0 ERROR
 * Variables      : 
 * Modifies       :  
 * Description    :
 ****************************************************************************/
int 
setStateServiceServer(int service, int state)
{
    program_load_config(NULL);
    if(state)
      pr_activate(service);
    else
      pr_deactivate(service);
    program_save_config(NULL);

    if(state)
      return execLineCommand(pr_get_program_run(service)); 
    else
      return execLineCommand(pr_get_program_stop(service));
}

/****************************************************************************
 * Function Name  : getStateServiceServer
 * Arguments      :
 * Returns        : int          1 enable
 *                               0 disable
 * Variables      : 
 * Modifies       :  
 * Description    :
 ****************************************************************************/
int 
getStateServiceServer(int service)
{
    program_load_config(NULL);
    if(pr_is_program_activated(service))
      return 1;
    return 0;
}


