#include <linux/version.h>

#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <asm/bootinfo.h>
#include <asm/addrspace.h>
#include <asm/mipsregs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/reboot.h>
#include <asm/time.h>
#include <asm/page.h>
#include <asm/tc3162/tc3162.h>

char arcs_cmdline[256];

extern int __imem, __dmem;

static void tc3162_component_setup(void)
{
	unsigned int controlReg;
	unsigned long flags;
	int i;

	/* setup bus timeout value */
	VPint(CR_AHB_AACS) = 0xffff;
	//VPint(CR_AHB_AACS) = 0x1010;

#ifdef CONFIG_CPU_IMEM
	/* setup imem start address */
	VPint(CR_IMEM) = PHYSADDR(&__imem);

	/* clear internal imem */
	save_and_cli(flags);
	controlReg = read_32bit_cp0_register(CP0_CCTL);
	write_32bit_cp0_register(CP0_CCTL, (controlReg & ~CCTL_IMEMOFF));
	write_32bit_cp0_register(CP0_CCTL, (controlReg | CCTL_IMEMOFF));
	write_32bit_cp0_register(CP0_CCTL, controlReg);
	restore_flags(flags);

	/* refill internal imem */
	save_and_cli(flags);
	controlReg = read_32bit_cp0_register(CP0_CCTL);
	write_32bit_cp0_register(CP0_CCTL, (controlReg & ~CCTL_IMEMFILL4));
	write_32bit_cp0_register(CP0_CCTL, (controlReg | CCTL_IMEMFILL4));
	write_32bit_cp0_register(CP0_CCTL, controlReg);
	restore_flags(flags);
	
	printk("Enable IMEM addr=%x\n", PHYSADDR(&__imem));
#endif

#ifdef CONFIG_CPU_DMEM
	/* setup dmem start address */
	VPint(CR_DMEM) = PHYSADDR(&__dmem);

	memcpy(0xa0001000, KSEG1ADDR(&__dmem), 0x800);

	/* clear internal dmem */
	save_and_cli(flags);
	controlReg = read_32bit_cp0_register(CP0_CCTL);
	write_32bit_cp0_register(CP0_CCTL, (controlReg & ~CCTL_DMEMOFF));
	write_32bit_cp0_register(CP0_CCTL, (controlReg | CCTL_DMEMOFF));
	write_32bit_cp0_register(CP0_CCTL, controlReg);
	restore_flags(flags);

	/* internal dmem on */
	save_and_cli(flags);
	controlReg = read_32bit_cp0_register(CP0_CCTL);
	write_32bit_cp0_register(CP0_CCTL, (controlReg & ~CCTL_DMEMON));
	write_32bit_cp0_register(CP0_CCTL, (controlReg | CCTL_DMEMON));
	write_32bit_cp0_register(CP0_CCTL, controlReg);
	restore_flags(flags);

	printk("Enable DMEM addr=%x\n", PHYSADDR(&__dmem));

	memcpy(KSEG1ADDR(&__dmem), 0xa0001000, 0x800);
#endif
}

const char *get_system_type(void)
{
	return "Trendchip TC3162 SoC";
}

/* Do basic initialization (called after cpu detection, from init_arch()) */
void __init prom_init(int argc, char **argv, unsigned long magic, int *prom_vec)
{
	unsigned long mem_size = CONFIG_TC3162_RAMSIZE;

#ifdef CONFIG_TC3162_SIO_CONSOLE
	tc3162_sio_console_init();
#endif
	tc3162_component_setup();

	/* Playing with MTD partitions, but no "root=" parm yet (use initrd for that) */
	//strcpy(arcs_cmdline, "console=/dev/ttyS0 mtdparts=phys_mapped_flash:64k(boot)ro,64k(romfile),512k(kernel),-(rootfs)");
//	strcpy(arcs_cmdline, "console=/dev/ttyS0 root=/dev/mtdb3 rootfstype=squashfs");
	//strcpy(arcs_cmdline, "console=/dev/ttyS0 mtdparts=phys_mapped_flash:64k(boot)ro,64k(romfile),512k(kernel),-(rootfs) root=/dev/mtdblock3 rootfstype=squashfs");
	//strcpy(arcs_cmdline, "console=/dev/ttyS0"); 
strcpy(arcs_cmdline, "console=/dev/ttyS0 root=/dev/mtd6 rootfstype=squashfs init=/sbin/init");

	mips_machgroup = MACH_GROUP_TRENDCHIP;
	mips_machtype = MACH_TC3162_SOC;

	add_memory_region(0, mem_size, BOOT_MEM_RAM); 
}

void __init prom_free_prom_memory (void)
{
}
