/*  
 * Implementations used to TC3162 console.
 *
 * Written by Robert Wu, Billion Co.
 * Support for TrendChip TC3162
 */


/*
 * Wrap-around code for a console on the
 * TrendChip TC3160 SIO.
 */


#include <linux/tty.h>
#include <linux/major.h>
#include <linux/ptrace.h>
#include <linux/init.h>

#ifdef CONFIG_TC3162_SIO_CONSOLE
#include <linux/console.h>
#endif /* CONFIG_TC3162_SIO_CONSOLE */

#include <linux/fs.h>
#include <linux/timer.h>
#include <linux/serial.h>
#include <linux/serialP.h>
#include <linux/interrupt.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/compatmac.h>
#include <linux/mm.h>
#include <linux/circ_buf.h>
#include <asm/serial.h>
#include <asm/tc3162/tc3162.h>
#include <asm/tc3162/TCIfSetQuery.h>	

/************************************************************************
*     A D S L   R E L A T E D    F U N C T I O N S
*************************************************************************
*/
#define RX_QUE_SIZE	PAGE_SIZE

static struct circ_buf rx_que;

static uint8 tc_console_flag = 0; 

void TCConsole(uint8 mode)
{
	tc_console_flag = mode;
}
EXPORT_SYMBOL(TCConsole);

void uartMacPutchar(int ch)
{
	unsigned long flags;

	if (!rx_que.buf) {
		unsigned long page;

		page = get_zeroed_page(GFP_KERNEL);
		if (!page) {
			printk("startup free page failure\n");
			return;
		}

		save_flags(flags); cli();
		rx_que.buf = (unsigned char *) page;
		rx_que.head = rx_que.tail = 0;
		restore_flags(flags);
	}

	save_flags(flags); cli();
	if (CIRC_SPACE(rx_que.head, rx_que.tail, RX_QUE_SIZE) == 0) {
		restore_flags(flags);
		return;
	}
	rx_que.buf[rx_que.head] = (char) ch;
	rx_que.head = (rx_que.head + 1) & (RX_QUE_SIZE-1);
	restore_flags(flags);
}
EXPORT_SYMBOL(uartMacPutchar);

static unsigned long  uclk_65000[13]={
			357500, 	// uclk 5.5     Baud Rate 115200
			175500, 	// uclk 2.7     Baud Rate 57600
			119808, 	// uclk 1.8432  Baud Rate 38400
			89856,	// uclk	1.3824	Baud Rate 28800
			59904,	// uclk 0.9216	Baud Rate 19200
			44928,	// uclk 0.6912	Baud Rate 14400
			29952,	// uclk 0.4608	Baud Rate 9600
			14976,	// uclk 0.2304	Baud Rate 4800
			7488,	// uclk 0.1152	Baud Rate 2400 
			3744,	// uclk 0.0576	Baud Rate 1200
			1872,	// uclk 0.0288	Baud Rate 600
			936,		// uclk 0.0144	Baud Rate 300
			343		// uclk 0.00528	Baud Rate 110
};


void tc3162_sio_putc(char data)
{
	while (!(VPchar(CR_UART_LSR) & LSR_THRE))
			;
	VPchar(CR_UART_THR) = data; 

#ifdef CONFIG_TC3162_ADSL
	if (tc_console_flag) {
		char buf[2];
		short length = 1;

		buf[0] = data;
		buf[1] = 0x0;

		if (adsl_dev_ops) {
			adsl_dev_ops->set(ADSL_SET_CONSOLE_DATA, (void *)buf, (void *)&length);
		}
	}
#endif
}

char tc3162_sio_getc(void)
{
	int i;
	char ch;

#ifdef CONFIG_TC3162_ADSL
	if (tc_console_flag) {
		if (rx_que.head != rx_que.tail) {
			ch = rx_que.buf[rx_que.tail];
			rx_que.tail = (rx_que.tail + 1) & (RX_QUE_SIZE-1);
			return ch;
		}
	}
#endif
    while (1)
	{
		if 	(VPchar(CR_UART_LSR) & 0x1)
			break;		
	}	
	i=VPchar(CR_UART_RBR);

	return (i & 0xff);
}

int tc3162_sio_tstc(void)
{
#ifdef CONFIG_TC3162_ADSL
	if (tc_console_flag) {
		if (rx_que.head != rx_que.tail)
			return 0x1;
	}
#endif
	return (VPchar(CR_UART_LSR) & 0x1);
}

#ifdef CONFIG_TC3162_ADSL

void tcconsole_sio_putc(char data)
{
	if (tc_console_flag) {
		char buf[2];
		short length = 1;

		buf[0] = data;
		buf[1] = 0x0;

		if (adsl_dev_ops) {
			adsl_dev_ops->set(ADSL_SET_CONSOLE_DATA, (void *)buf, (void *)&length);
		}
	}
}

unsigned int tcconsole_sio_getc(void)
{
	int i;
	char ch;

	if (tc_console_flag) {
		if (rx_que.head != rx_que.tail) {
			ch = rx_que.buf[rx_que.tail];
			rx_que.tail = (rx_que.tail + 1) & (RX_QUE_SIZE-1);
			return ch;
		}
	}
	return 0xffff;
}

int tcconsole_sio_tstc(void)
{
	if (tc_console_flag) {
		if (rx_que.head != rx_que.tail)
			return 0x1;
	}
	return 0x0;
}
#endif

#ifdef CONFIG_TC3162_SIO_CONSOLE

void sio_console_write(struct console *co, const char *s,
			       unsigned count)
{
	register int i;

  	for (i = 0; i < count; i++) {
    	if (s[i] == '\n')
      		tc3162_sio_putc('\r');     
    	tc3162_sio_putc(s[i]);     
  	}
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,19))
/* not needed in post-2.4.18 versions */

static int sio_console_wait_key(struct console *co)
{
	int res;

	res = tc3162_sio_getc();
	return res;
}
#endif

static int __init sio_console_setup(struct console *co, char *options)
{
	int i;
	unsigned long	div_x,div_y;
	unsigned long	word;

//	pause(100);
// Set FIFO controo enable, reset RFIFO, TFIFO, 16550 mode, watermark=0x00 (1 byte)
	VPchar(CR_UART_FCR) = UART_FCR|UART_WATERMARK;

// Set modem control to 0
	VPchar(CR_UART_MCR) = UART_MCR;

// Disable IRDA, Disable Power Saving Mode, RTS , CTS flow control
	VPchar(CR_UART_MISCC) = UART_MISCC;

// Set interrupt Enable to, enable Tx, Rx and Line status
	VPchar(CR_UART_IER) = UART_IER;	
//	

	/* access the bardrate divider */

	VPchar(CR_UART_LCR) = UART_BRD_ACCESS;

	div_y = UART_XYD_Y;
	div_x = (unsigned int)(uclk_65000[0]/SYS_HCLK)*2;
	word = (div_x<<16)|div_y;
	VPint(CR_UART_XYD) = word;

/* Set Baud Rate Divisor to 3*16		*/
	VPchar(CR_UART_BRDL) = UART_BRDL;

	VPchar(CR_UART_BRDH) = UART_BRDH;

/* Set DLAB = 0, clength = 8, stop =1, no parity check 	*/
	VPchar(CR_UART_LCR) = UART_LCR;

	return 0;
}

static kdev_t sio_console_device(struct console *cons)
{
	return(MKDEV(TTY_MAJOR, 64 + cons->index));
}

static struct console sercons =
{
	name:		"ttyS",
	write:		sio_console_write,
	device:		sio_console_device,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,19))
/* removed in post-2.4.18 versions */
	wait_key:	sio_console_wait_key,
#endif
	setup:		sio_console_setup,
	flags:		CON_PRINTBUFFER,
	index:		-1,
};

/*
 * Register console.
 */

void __init tc3162_sio_console_init(void)
{
	register_console(&sercons);
}

#endif /* CONFIG_TC3162_SIO_CONSOLE */

