/*
 *  linux/arch/mips/dec/time.c
 *
 *  Copyright (C) 1991, 1992, 1995  Linus Torvalds
 *  Copyright (C) 2000  Maciej W. Rozycki
 *
 * Implementations used for system timer controller configuration.
 *
 * Written by Robert Wu, Billion Co.
 * Support for TrendChip TC3162
 *
 */


#include <linux/errno.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>

#include <asm/cpu.h>
#include <asm/bootinfo.h>
#include <asm/mipsregs.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <linux/timex.h>

#include <asm/tc3162/interrupts.h>
#include <asm/tc3162/timer.h>
#include <asm/tc3162/tc3162.h>

extern rwlock_t xtime_lock;
extern struct task_struct *current;

#define ENABLE 1


#if 0 /* Show bar progress */
static char bars[4] = { '-', '\\', '|', '/' };
static int barcnt = 0;
#endif

static inline int timer_intr_valid(void) 
{
  unsigned long stat;

  /* Read cleared the status */
  stat = *(volatile unsigned long *)(TC3162_TIMERCTLR_REG);

	stat &=~((TIMERCTLR_Timer0_Int)|
		 (TIMERCTLR_Timer1_Int)|
		 (TIMERCTLR_Timer2_Int)|
		 (TIMERCTLR_Timer3_Int)|
		 (TIMERCTLR_Timer4_Int)|
		 (TIMERCTLR_Timer5_Int)
		);
	/* Clear the bit */
	stat |= TIMERCTLR_Timer1_Int; 
	*(volatile unsigned long *)(TC3162_TIMERCTLR_REG) = stat; 

  return(1);
}

/*
 * Call the "do_timer()" routine every clocktick
 */
static void inline
timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  if (timer_intr_valid()) {
#if 0
    if (!user_mode(regs)) {
      if (prof_buffer && current->pid) {
        extern int _stext;
        unsigned long pc = regs->cp0_epc;

        pc -= (unsigned long) &_stext;
        pc >>= prof_shift;

	/*
	 * Dont ignore out-of-bounds pc values silently,
	 * put them into the last histogram slot, so if
	 * present, they will show up as a sharp peak.
	 */
	if (pc > prof_len - 1)
	  pc = prof_len - 1;
        atomic_inc((atomic_t *) & prof_buffer[pc]);
      }
    }

#if 0 /* Show bar progress */
    printk("%c\b", bars[barcnt]);
    barcnt = (barcnt + 1) % 4;
#endif
#endif

    do_timer(regs);
  }
}

struct irqaction timer_irq = { timer_interrupt, SA_INTERRUPT, 0,
			       "timer", NULL, NULL};

void __init tc3162_time_init(void)
{
  unsigned int year, mon, day, hour, min, sec;
  unsigned long flags;
   
  sec = min = hour = 0;
  day = 1;
  mon = 1;
  year = 2003;

  /* Save the time */
  write_lock_irqsave(&xtime_lock, flags);
  xtime.tv_sec = mktime(year, mon, day, hour, min, sec);
  xtime.tv_usec = 0;
  write_unlock_irqrestore(&xtime_lock, flags);
}

static void setIntPriority(char source, char level)
{
  int ipRegNum,ipLevNum;
  int word;
  
  ipRegNum = level/4;
  ipLevNum = level%4;
  word =*(volatile unsigned int*)(TC3162_IPSR0_REG+ipRegNum*4);
  word |= source << ((3-ipLevNum)*8);
  *(volatile unsigned int*)(TC3162_IPSR0_REG+ipRegNum*4) =word;

}

static void setIntMask(
	int int_no,
	int enable
)
{
  if(enable == ENABLE)
     *(volatile unsigned int*)(TC3162_IMR_REG) |= (0x01 << int_no);
  else
     *(volatile unsigned int*)(TC3162_IMR_REG) &= ~((0x01) << int_no);

}

static void setTimer(
	int timer_no,
	int timerTime,
	int enable,
	int mode,
	int halt
)
{
	int word,word1;

  word = (timerTime * SYS_HCLK) * 1000/2;
  *(volatile unsigned int*)(TC3162_TIMER0LDR_REG+timer_no*0x08) = word;

  word = *(volatile unsigned int*)(TC3162_TIMERCTLR_REG);
  word1 = (enable << timer_no) |(mode << (timer_no + 8))|(halt << (timer_no + 26));
  word |= word1;
  *(volatile unsigned int*)(TC3162_TIMERCTLR_REG) = word;
}


static void tc3162_init_timers(void)
{
  /* Initialize timer */  
  setTimer(1,TIMERTICKS_10MS,ENABLE,TIMER_TOGGLEMODE,TIMER_HALTDISABLE);
  setIntMask(INTR_TIMER1,ENABLE);
}

void __init tc3162_time_setup(struct irqaction *irq)
{
  irq->handler = timer_interrupt;
  irq->flags = SA_INTERRUPT;
  irq->name = "timer";
  irq->mask = 0x20;

  setup_irq(INTR_TIMER1, irq);

  /* Initialize timers and use Timer 1 as system timer */
  tc3162_init_timers();
}

