/*
** $Id: sdl_subr.c,v 1.1.1.1 2005/09/15 13:37:48 root Exp $
*/

/*! \file sdl_subr.c
    \brief IPN2220 AP supporting subroutines for SDL

    This should be the only place within the SDL set of .C source codes which
    know about the IPN2220 specific implementation, especially 'i2220mgmt_t'
*/

/******************************************************************************
Copyright (c) 2002-2003 Integrated Programmable Communications, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with
Integrated Programmable Communications, Inc.
******************************************************************************/

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/
#define WEP_TEST_KEY \
        { \
            0x00, 0x00, 0x00, 0x98, 0x76, 0x54, 0x32, 0x10, \
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 \
        }
#define MAP_TEST_KEY \
        { \
            0x00, 0x00, 0x00, 0x33, 0x33, 0x33, 0x33, 0x33, \
            0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33 \
        }
#define TEST_MAC_ADDR   {0x00, 0x11, 0x22, 0x33, 0x44, 0x55}

#define PSK_PHRASE      {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
                         0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/


/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/
TypMacAddr bcstAddr = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF }; /* broadcast address */
TypMacAddr nullAddr = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }; /* null address */

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/
// KenWang added 2005/06/24
extern int fac_pid;

/* Initial values for IPN2220's 802.11 MIB */
static IEEE_802_11_MIB initIeee80211Mib = {
    /* Station ManagemenT (SMT) Attributes (ieee802dot11 1) */
    /* dot11StationConfigTable (dot11smt 1) */
//    NULL_MAC_ADDR,              /* dot11StationID */
//    100,        /* dot11MediumOccupancyLimit */
//    FALSE,      /* dot11CFPollable */
//    0,          /* dot11CFPPeriod */
//    0,          /* dot11CFPMaxDuration */
//  500,        /* dot11AuthenticationResponseTimeOut */
//  TRUE,       /* dot11PrivacyOptionalImplemented, already implemented by S/W and H/W */
//  MIB_PWR_MGT_MODE_ACTIVE,    /* dot11PowerManagementMode */
    {
        DEFAULT_SSID_LEN,     /* length of dot11DesiredSSID */
        DEFAULT_SSID          /* dot11DesiredSSID */
    },
//  BSS_TYPE_INFRASTRUCTURE,    /* dot11DesiredBSSType */
    {
        12,         /* size of dot11OperationalRateSet */
        {           /* dot11OperationalRateSet */
            RATE_1M | RATE_BASIC_BIT,       /* 1M (basic rate) */
            RATE_2M | RATE_BASIC_BIT,       /* 2M (basic rate) */
            RATE_5_5M | RATE_BASIC_BIT,     /* 5.5M (basic rate) */
            RATE_11M | RATE_BASIC_BIT,      /* 11M (basic rate) */
            RATE_6M,                        /* 6M */
            RATE_12M,                       /* 12M */
            RATE_24M,                       /* 24M */
            RATE_54M,                       /* 54M */
            RATE_9M,                        /* 9M */
            RATE_18M,                       /* 18M */
            RATE_36M,                       /* 36M */
            RATE_48M,                       /* 48M */
            0, 0, 0, 0
        }
    },
    100,        /* dot11BeaconPeriod */
    1,          /* dot11DTIMPeriod */
//  500,        /* dot11AssociationResponseTimeOut */
//  0,          /* dot11DisassociateReason */
//  NULL_MAC_ADDR,  /* dot11DisassociateStation */
//  0,              /* dot11DeauthenticateReason */
//  NULL_MAC_ADDR,  /* dot11DeauthenticateStation */
//  0,              /* dot11AuthenticateFailStatus */
//  NULL_MAC_ADDR,  /* dot11AuthenticateFailStation */
#if SUPPORT_802_11D
    //TRUE,             /* dot11StationConfigEntry 21 */
    TRUE,               /* dot11StationConfigEntry 22 */
    //{'U', 'S', ' '}   /* dot11StationConfigEntry 23 */
#endif
#if SUPPORT_IPNRSN
//  TRUE,           /* dot11RSNOptionImplemented */
#endif /* SUPPORT_RSN */

    MIB_CWMIN,      /* aCWmin  */

    /* dot11AuthenticationAlgorithmsTable (dot11smt 2) */
//    2,          /* number of dot11AuthenticationAlgorithms */
    {
        { 1, MIB_AUTH_ALGORITHM_OPEN_SYSTEM, TRUE },    /* open system */
        { 2, MIB_AUTH_ALGORITHM_SHARED_KEY,  TRUE }     /* shared key */
    },

#if SUPPORT_IPNRSN | SUPPORT_WPA2
    /* dot11WEPDefaultKeysTable (dot11smt 3) */
    {
        { 1, RSN_NULL_KEY, RSN_KEY_LEN },        /* Key ID 0 */
        { 2, RSN_NULL_KEY, RSN_KEY_LEN },        /* Key ID 1 */
        { 3, RSN_NULL_KEY, RSN_KEY_LEN },        /* Key ID 2 */
        { 4, RSN_NULL_KEY, RSN_KEY_LEN }         /* Key ID 3 */
    },
#else
    /* dot11WEPDefaultKeysTable (dot11smt 3) */
    {
        { 1, WEP_TEST_KEY },        /* Key ID 0 */
        { 2, WEP_TEST_KEY },        /* Key ID 1 */
        { 3, WEP_TEST_KEY },        /* Key ID 2 */
        { 4, WEP_TEST_KEY }         /* Key ID 3 */
    },
#endif /* SUPPORT_RSN */

    /* dot11PrivacyTable (dot11smt 5) */
//    FALSE,                  /* dot11PrivacyInvoked */
//    0,                      /* dot11WEPDefaultKeyID */
//    WEP_KEY_MAPPING_LEN,    /* dot11WEPKeyMappingLength */
//    TRUE,                   /* dot11ExcludeUnencrypted */
    0,                      /* dot11WEPICVErrorCount */
//    0,                      /* dot11WEPExcludedCount */
#if SUPPORT_IPNRSN
    FALSE,                  /* dot11RSNEnabled */
//    FALSE,                  /* dot11HaveGTK */
//    FALSE,                  /* dot11HavePTK */
//    FALSE,                  /* dot11WEPKeyMappingsKeyBroadcast */
    FALSE,                  /* dot11Pure1XEnabled */
#elif SUPPORT_WPA2
    FALSE,                    /* dot11RSNEnabled */
//    FALSE,                  /* dot11HaveGTK */
//    FALSE,                  /* dot11HavePTK */
//    FALSE,                  /* dot11WEPKeyMappingsKeyBroadcast */
    TRUE,                     /* dot11Pure1XEnabled */
#endif

    /* MAC Attributes (ieee802dot11 2) */
    /* dot11OperationTable (dot11mac 1) */
    NULL_MAC_ADDR,          /* dot11MACAddress */
    2347,       /* dot11RTSThreshold */
    7,          /* dot11ShortRetryLimit */
    4,          /* dot11LongRetryLimit */
    2346,       /* dot11FragmentationThreshold */
//    512,        /* dot11MaxTransmitMSDULifetime */
//    512,        /* dot11MaxReceiveLifetime */
//    "",         /* dot11ManufacturerID */
//    "",         /* dot11ProductID */

    /* dot11CountersTable (dot11mac 2) */
    0,          /* dot11CountersEntry 1 */
    0,          /* dot11CountersEntry 2 */
    0,          /* dot11CountersEntry 3 */
    0,          /* dot11CountersEntry 4 */
    0,          /* dot11CountersEntry 5 */
    0,          /* dot11CountersEntry 6 */
    0,          /* dot11CountersEntry 7 */
    0,          /* dot11CountersEntry 8 */
    0,          /* dot11CountersEntry 9 */
    0,          /* dot11CountersEntry 10 */
    0,          /* dot11CountersEntry 11 */
    0,          /* dot11CountersEntry 12 */
    0,          /* dot11CountersEntry 13 */
    0,          /* dot11CountersEntry 14 */
#if 0
    /* dot11GroupAddressesTable (dot11mac 3) */
    0,
    {
        {  1, NULL_MAC_ADDR, 1 }, {  2, NULL_MAC_ADDR, 1 },
        {  3, NULL_MAC_ADDR, 1 }, {  4, NULL_MAC_ADDR, 1 },
        {  5, NULL_MAC_ADDR, 1 }, {  6, NULL_MAC_ADDR, 1 },
        {  7, NULL_MAC_ADDR, 1 }, {  8, NULL_MAC_ADDR, 1 },
        {  9, NULL_MAC_ADDR, 1 }, { 10, NULL_MAC_ADDR, 1 },
        { 11, NULL_MAC_ADDR, 1 }, { 12, NULL_MAC_ADDR, 1 },
        { 13, NULL_MAC_ADDR, 1 }, { 14, NULL_MAC_ADDR, 1 },
        { 15, NULL_MAC_ADDR, 1 }, { 16, NULL_MAC_ADDR, 1 },
        { 17, NULL_MAC_ADDR, 1 }, { 18, NULL_MAC_ADDR, 1 },
        { 19, NULL_MAC_ADDR, 1 }, { 20, NULL_MAC_ADDR, 1 },
        { 21, NULL_MAC_ADDR, 1 }, { 22, NULL_MAC_ADDR, 1 },
        { 23, NULL_MAC_ADDR, 1 }, { 24, NULL_MAC_ADDR, 1 },
        { 25, NULL_MAC_ADDR, 1 }, { 26, NULL_MAC_ADDR, 1 },
        { 27, NULL_MAC_ADDR, 1 }, { 28, NULL_MAC_ADDR, 1 },
        { 29, NULL_MAC_ADDR, 1 }, { 30, NULL_MAC_ADDR, 1 },
        { 31, NULL_MAC_ADDR, 1 }, { 32, NULL_MAC_ADDR, 1 }
    },
#endif
    /* Resource Attributes (ieee802dot11 3) */
    /* Resource Type Attributes (dot11res 1) */
//    "RTID",                             /* dot11ResourceTypeIDName */
    /* dot11resAttribute 2 */
//    DEFAULT_MANUFACTURER_OUI,           /* dot11manufacturerOUI */
//    DEFAULT_MANUFACTURER_NAME,          /* dot11manufacturerName */
//    DEFAULT_MANUFACTURER_PRODUCT_NAME,  /* dot11manufacturerProductName */
//    DEFAULT_MANUFACTURER_PRODUCT_VER,   /* dot11manufacturerProductVersion */

    /* PHY Attributes (ieee802dot11 4) */
    /* dot11PhyOperationTable (dot11phy 1) */
//    MIB_PHY_TYPE_HRDSSS,        /* dot11PHYType */
    0,                          /* dot11CurrentRegDomain */
//    MIB_TEMP_TYPE_1,            /* dot11TempType */

    /* dot11PhyAntennaTable (dot11phy 2) */
    MIB_ANTENNA_PRIMARY,        /* dot11CurrentTxAntenna */
    MIB_DIVERSITY_FIXED_LIST,   /* dot11DiversitySupport */
    MIB_ANTENNA_DIVERSITY,      /* dot11CurrentRxAntenna */
#if 0
    /* dot11PhyTxPowerTable (dot11phy 3) */
    1,              /* dot11NumberSupportedPowerLevels */
    {
        50,             /* dot11TxPowerLevel1 */
        0,              /* dot11TxPowerLevel2 */
        0,              /* dot11TxPowerLevel3 */
        0,              /* dot11TxPowerLevel4 */
        0,              /* dot11TxPowerLevel5 */
        0,              /* dot11TxPowerLevel6 */
        0,              /* dot11TxPowerLevel7 */
        0               /* dot11TxPowerLevel8 */
    },
    1,              /* dot11CurrentTxPowerLevel */
#endif
    /* dot11PhyDSSSTable (dot11phy 5) */
    6,                          /* dot11CurrentChannel */
//    MIB_CCA_MODE_ED_ONLY | MIB_CCA_MODE_CS_ONLY | MIB_CCA_MODE_ED_AND_CS,   /* dot11CCAModeSupported */
//    MIB_CCA_MODE_ED_AND_CS,     /* dot11CurrentCCAMode */
//    0,                          /* dot11EDThreshold */
#if 0
    /* dot11RegDomainsSupportedTable (dot11phy 7) */
    7,              /* number of supported regulatory domains */
    {
        { 1, MIB_REG_DOMAIN_FCC },      /* FCC (US) */
        { 2, MIB_REG_DOMAIN_IC },       /* IC or DOC (Canada) */
        { 3, MIB_REG_DOMAIN_ETSI},      /* ETSI (Europe) */
        { 4, MIB_REG_DOMAIN_SPAIN },    /* Spain */
        { 5, MIB_REG_DOMAIN_FRANCE },   /* France */
        { 6, MIB_REG_DOMAIN_JAPAN_1 },  /* MPHPT (Japan) Channel 14 */
        { 7, MIB_REG_DOMAIN_JAPAN_2 },  /* MPHPT (Japan) Channel 1 to 13 */
        { 8, 0 }
    },
#endif
    /* dot11AntennasListTable (dot11phy 8) */
    2,              /* size of antennas list table */
    {
        { MIB_ANTENNA_PRIMARY, FALSE, FALSE, FALSE },   /* primary antenna */
        { MIB_ANTENNA_SECONDARY, FALSE, FALSE, FALSE }  /* secondary antenna */
    },
#if 0
    /* dot11SupportedDataRatesTxTable (dot11phy 9) */
    12,
    {
        { 1, RATE_1M },         /* 1 Mbit/s */
        { 2, RATE_2M },         /* 2 Mbit/s */
        { 3, RATE_5_5M },       /* 5.5 Mbit/s */
        { 4, RATE_11M },        /* 11 Mbit/s */
        { 5, RATE_6M },         /* 6 Mbit/s */
        { 6, RATE_9M },         /* 9 Mbit/s */
        { 7, RATE_12M },        /* 12 Mbit/s */
        { 8, RATE_18M },        /* 18 Mbit/s */
        { 9, RATE_24M },        /* 24 Mbit/s */
        { 10, RATE_36M },       /* 36 Mbit/s */
        { 11, RATE_48M },       /* 48 Mbit/s */
        { 12, RATE_54M },       /* 54 Mbit/s */
        { 13, 0 }, { 14, 0 }, { 15, 0 }, { 16, 0 }
    },

    /* dot11SupportedDataRatesRxTable (dot11phy 10) */
    12,
    {
        { 1, RATE_1M },         /* 1 Mbit/s */
        { 2, RATE_2M },         /* 2 Mbit/s */
        { 3, RATE_5_5M },       /* 5.5 Mbit/s */
        { 4, RATE_11M },        /* 11 Mbit/s */
        { 5, RATE_6M },         /* 6 Mbit/s */
        { 6, RATE_9M },         /* 9 Mbit/s */
        { 7, RATE_12M },        /* 12 Mbit/s */
        { 8, RATE_18M },        /* 18 Mbit/s */
        { 9, RATE_24M },        /* 24 Mbit/s */
        { 10, RATE_36M },       /* 36 Mbit/s */
        { 11, RATE_48M },       /* 48 Mbit/s */
        { 12, RATE_54M },       /* 54 Mbit/s */
        { 13, 0 }, { 14, 0 }, { 15, 0 }, { 16, 0 }
    },
#endif

#if SUPPORT_802_11A
    /* dot11PhyOFDMTable (dot11phy 11) */
    52,                         /* dot11CurrentFrequency */
    0,                          /* dot11TIThreshold */
    MIB_U_NII_BAND_LOWER | MIB_U_NII_BAND_MIDDLE | MIB_U_NII_BAND_UPPER,
                                /* dot11FrequencyBandsSupported */
#endif

    /* dot11PhyHRDSSSTable (dot11phy 12) */
//    TRUE,                       /* dot11ShortPreambleOptionImplemented */
//    FALSE,                      /* dot11PBCCOptionImplemented */
//    FALSE,                      /* dot11ChannelAgilityPresent */
//    FALSE,                      /* dot11ChannelAgilityEnabled */

    /* dot11PhyERPTable (dot11phy 14) */
//    FALSE,                      /* dot11ER-PBCCOptionImplemented */
//    FALSE,                      /* dot11DSSS-OFDMOptionImplemented */
//    FALSE,                      /* dot11DSSS-OFDMOptionEnabled */
//    TRUE,                       /* dot11ShortSlotTimeOptionImplemented */
//    FALSE,                      /* dot11ShortSlotTimeOptionEnabled */

#if SUPPORT_IPNRSN
    /* RSN Attributes (ieee802dot11 7) */
    /* dot11RSNConfigTable (ieee802dot11i 1) */
    {
//        0,            /* dot11RSNConfigIndex */
//        1,            /* dot11RSNConfigVersion */
//        WEP_KEY_MAPPING_LEN, /* dot11RSNConfigPairwiseKeysSupported */
//        {{0x02f25000}},     //{0x00,0x50,0xf2,2},  /* dot11RSNConfigMulticastCipher[0:4] */
        0,            /* dot11RSNConfigGroupRekeyMethod */
//        86400,        /* dot11RSNConfigGroupRekeyTime */
//        0,            /* dot11RSNConfigGroupRekeyPackets */
//        TRUE,         /* dot11RSNConfigGroupRekeyStrict */
        {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
         0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}, /* dot11RSNConfigPSKValue[RSN_KEY_LEN] */
        PSK_PHRASE,   /* dot11RSNConfigPSKPassPhrase[64] */
//        FALSE,        /* dot11RSNConfigTSNEnabled */
//        604800,       /* dot11RSNConfigGroupMasterRekeyTime */
        100,          /* dot11RSNConfigGroupUpdateTimeout */
        3,            /* dot11RSNConfigGroupUpdateCount */
        100,          /* dot11RSNConfigPairwiseUpdateTimeout */
        3             /* dot11RSNConfigPairwiseUpdateCount */
//        {0,0,0,0,0,0},/* dot11RSNConfigGroupKeyRSC */
//        0             /* dot11RSNConfigTKIPReplays */
    },

    /* dot11RSNConfigUnicsastCiphersTable (ieee802dot11i 2) */
    {
        {0, {{0x00f25000}}, FALSE }, /* None */
        {1, {{0x01f25000}}, FALSE}, /* WEP-40 */
        {2, {{0x02f25000}}, TRUE}, /* TKIP, default in WPA mode */
        {3, {{0x03f25000}}, FALSE}, /* WRAP */
        {4, {{0x04f25000}}, FALSE}, /* CCMP, default in RSN mode */
        {5, {{0x05f25000}}, FALSE}  /* WEP-104 */
        /* Meaning explain: */
        /* dot11RSNConfigUnicastCipherIndex */
        /* dot11RSNConfigUnicastCipher */
        /* dot11RSNConfigUnicastCipherEnabled */
    },

    /* dot11RSNConfigAuthenticationSuitesTable (ieee802dot11i 3) */
    {
        {0, {{0x00f25000}}, FALSE}, /* Reserved */
        {1, {{0x01f25000}}, FALSE }, /* 802.1X */
        {2, {{0x02f25000}}, TRUE}  /* PSK */
        /* Meaning explain: */
        /* dot11RSNConfigAuthenticationSuiteIndex */
        /* dot11RSNConfigAuthenticationSuite */
        /* dot11RSNConfigAuthenticationSuiteEnabled */
    },

#endif /* SUPPORT_RSN */

    /* QoS attribute*/
    /* SMT EDCF config table*/
    {MIB_CWMIN, MIB_CWMIN, ((MIB_CWMIN+1)/2-1), ((MIB_CWMIN+1)/4-1)},
    {MIB_CWMAX, MIB_CWMAX, MIB_CWMIN, ((MIB_CWMIN+1)/2-1)},
    {2,1,1,1},                       /* EDCF AIFSN */
    {0, 1500, 3000, 1500}               /* EDCF TXOP Limit */


};



/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/
void set_ap_timer(i2220mgmt_t *priv_p);

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*----------------------------------------------------------------*/
/*!
* \brief initialize the MIB structure.
*
* This routine is called to initialize all of the IEEE
* 802.11 MIB attributes in the Adapter structure.
*
* \param  priv_p  pointer to i2220mgmt_t, the device's private structure
* \return none
*/
/*----------------------------------------------------------------*/
void mib_initialize(i2220mgmt_t *priv_p)
{
        /* Initialize the IEEE 802.11 MIB */
        ipn_memcpy( &priv_p->mib, (void*) &initIeee80211Mib, sizeof(IEEE_802_11_MIB));
        mib_import(dot11CurrentChannel) = x_channel;
        priv_p->x_CfbBurst = x_CfbBurst;        /* Burst mode */
        priv_p->x_hide_ssid = x_hide_ssid;      /* Hidden SSID */

        /* Setup the PHY preamble mode and slot time */
        priv_p->phyPreambleType = phyShortPreamble;
        priv_p->phySlotType = phyShortSlot;

#if WLAN_INCLUDE_WDS
        priv_p->wds_tx_failure_limit = WDS_TX_FAIL_LIMIT;
        priv_p->wds_disable_limit = WDS_DISABLE_LIMIT * HZ;
#endif
}

/*----------------------------------------------------------------*/
/*!
* \brief  To look up the STA record by its MAC address.
*
* The STA record is looked up via the HASH table of MAC addresses. Then
* the STA record is linearly searched against corresponding the HASH table slot.
*
* \param  addr  the MAC address of the STA.
* \return the corresponding sta_info_t STA record.
*/
/*----------------------------------------------------------------*/
/* caller should have mutex for AP STA list handling */
sta_info_t* task_sta_lookup(i2220mgmt_t *priv_p, MacAddr_t addr)
{
        sta_info_t      *sta_p;

        sta_p = priv_p->ap.sta_hash[sta_hashfn(addr)];
        while (sta_p != NULL && ipn_memcmp(sta_p->asoc.adAddr, addr, MAC_ADDR_LEN) != 0)
                sta_p = sta_p->hash_next;

        return sta_p;
}

/*----------------------------------------------------------------*/
/*!
* \brief    add into the HASH table for the STA \a sta_p
*
* One slot of the HASH table for this STA will be calculated by the hash function
* and the \a sta_p will be added into the link list.
*
* \param  sta_p  pointer to the STA record of type sta_info_t.
* \return none
*/
/*----------------------------------------------------------------*/
void add_hash_sta(i2220mgmt_t *priv_p, sta_info_t *sta_p)
{
        int     h;

        h = sta_hashfn(sta_p->asoc.adAddr);
        sta_p->hash_next = priv_p->ap.sta_hash[h];
        priv_p->ap.sta_hash[h] = sta_p;
}

/*----------------------------------------------------------------*/
/*!
* \brief    delete from the HASH table for the STA \a sta_p
*
* The slot of the HASH table for this STA, calculated by the hash function,
* will have this STA \a sta_p to be removed from the link list.
*
* \param  sta_p  pointer to the STA record of type sta_info_t.
* \return none
*/
/*----------------------------------------------------------------*/
void del_hash_sta(i2220mgmt_t *priv_p, sta_info_t *sta_p)
{
        sta_info_t      *s;
        int             h;

        h = sta_hashfn(sta_p->asoc.adAddr);
        s = priv_p->ap.sta_hash[h];
        if (s == NULL) return;
        if (ipn_memcmp(s->asoc.adAddr, sta_p->asoc.adAddr, MAC_ADDR_LEN) == 0) {
                priv_p->ap.sta_hash[h] = s->hash_next;
                return;
        }

        while (s->hash_next != NULL &&
             ipn_memcmp(s->hash_next->asoc.adAddr, sta_p->asoc.adAddr, MAC_ADDR_LEN) != 0)
                s = s->hash_next;

        if (s->hash_next != NULL)
                s->hash_next = s->hash_next->hash_next;
        else
                KDBG_LOG_SDL(KDBG_WARN,"AP: could not remove STA " MACSTR " from hash table\n",
                       MAC2STR(sta_p->asoc.adAddr));
}

/*----------------------------------------------------------------*/
/*!
* \brief    Create a new STA record for STA of MAC address \a addr.
*
* The memory for the STA record is allocated and properly initialized.
* \a add_hash_sta is invoked to put pointer of the record into the HASH
* table. \a add_sta_proc is invoked to put the STA into the PROC file system,
* a means of accessing the STA's dynamic status.
*
* \param  addr  MAC address for this STA.
* \return pointer to the STA record of type sta_info_t.
*/
/*----------------------------------------------------------------*/
sta_info_t* create_sta(i2220mgmt_t *priv_p, MacAddr_t addr)
{
        sta_info_t      *sta_p;

        KDBG_ENTER;
        sta_p = (sta_info_t*)kal_mempool_alloc(sizeof(sta_info_t));
        if (sta_p == NULL) return NULL;

        /* initialize STA info data */
        ipn_memset(sta_p, 0, sizeof(sta_info_t));
        ipn_memcpy(sta_p->asoc.adAddr, addr, MAC_ADDR_LEN);
        sdlQ_emptyQ(&sta_p->psQ);

        sta_p->auth_state = ST_Auth_Rsp_Idle;
        sta_p->asoc.adPsm = SYN_sta_active;
        sta_p->actTime = kal_jiffies;

        sta_p->magicID = STA_MAGIC_ID;
        add_hash_sta(priv_p, sta_p);

        add_chain_tail(&priv_p->ap.sta_chain, sta_info_t, sta_p);

#if WLAN_INCLUDE_PROC && WLAN_INCLUDE_DEBUG
        /* We enable this function only in debugging mode because
         * embedded system crashes (slab.c) if successive proc files
         * are created in a short time.
         */
        add_sta_proc(priv_p, sta_p);
#endif

        KDBG_LOG_SDL(KDBG_DEBUG,"create STA " MACSTR " @0x%08lx\n",
                MAC2STR(sta_p->asoc.adAddr), (unsigned long)sta_p);

        KDBG_EXIT;
        return sta_p;
}

/*----------------------------------------------------------------*/
/*!
* \brief  release the association resource for this STA.
*
* \param  sta   pointer to the STA record of type sta_info_t.
* \return none
*/
/*----------------------------------------------------------------*/
void leave_assoc_state(i2220mgmt_t *priv_p, sta_info_t *sta_p)
{
        FragSdu_t       *fsdu;

        ASSERT(sta_p);

        while (sdlQ_Length(&sta_p->psQ) != 0) {
                sdlQ_dequeue_first(&sta_p->psQ, fsdu);
                sdlu_release_fsdu(fsdu);
        }

        /* Only process associated STA */
        if (sta_p->aid <= 0) return;

        priv_p->ap.sta_aid[sta_p->aid - 1] = NULL;
        sdl_reset_tmap(priv_p, sta_p->aid);
        if (sta_p->asoc.adPsm == SYN_power_save) {
                sta_p->asoc.adPsm = SYN_sta_active;
                priv_p->sta_sleeping--;
        }
        sta_p->aid = 0; /* Important */

        /* Update nonERP, preamble and slot time info */
        update_bss_operating(priv_p);

        /* Destroy association, indicate upper layer */
        kal_disassoc_indicate(priv_p->netdev, sta_p->asoc.adAddr);
}

/*----------------------------------------------------------------*/
/*!
* \brief  release the allocated resource for this STA.
*
* the memory and PROC file system resources, allocated beforehand, will
* be released here.
*
* \param  sta   pointer to the STA record of type sta_info_t.
* \return none
*/
/*----------------------------------------------------------------*/
void free_sta(i2220mgmt_t *priv_p, sta_info_t *sta_p)
{
        IPN_KEY_STRUC   k;

        KDBG_ENTER;

        ASSERT(sta_p);

	/* KenWang added 2005/06/24 -- begin -- */
	/* the client is disassociated, send signal to factoryreset to update iptables */
	struct task_struct *p;
	siginfo_t si;
	int pid = 0;
	pid = fac_pid;
	si.si_signo = SIGUSR2;
	si.si_errno = 0;
	p = find_task_by_pid(pid);
	send_sig_info(SIGUSR2, &si, p);
	/* KenWang added 2005/06/24 -- end -- */
#if WLAN_INCLUDE_PROC && WLAN_INCLUDE_DEBUG
        if (priv_p->proc_STA_dir != NULL) {
                char name[20];
                sprintf(name, MACSTR, MAC2STR(sta_p->asoc.adAddr));
                remove_proc_entry(name, priv_p->proc_STA_dir);
        }
#endif
        KDBG_LOG_SDL(KDBG_DEBUG,"free STA " MACSTR " @0x%08lx\n",
                MAC2STR(sta_p->asoc.adAddr), (unsigned long)sta_p);

        /* The HASH table entry for this STA must be cleared. */
        del_hash_sta(priv_p, sta_p);

        del_chain_member(&priv_p->ap.sta_chain, sta_info_t, sta_p);

        /* Clear the entry in security table */
        k.chkKeyId = TRUE;
        k.chkAddr2 = TRUE;
        k.chkAddr1 = FALSE;
        k.keyId = sta_p->keyId;
        ipn_memcpy(k.addr, sta_p->asoc.adAddr, 6);
        wlanoidSetClearKey(priv_p, &k);

        leave_assoc_state(priv_p, sta_p);

        /* Update info of neighbor BSS */
        if (sta_p->incNonERP == NEIGHBOR_NONERP) {
                priv_p->bss_nonERP--;
        }

        sta_p->magicID = 0;
        kal_mempool_free(sta_p);
        KDBG_EXIT;
}

/*----------------------------------------------------------------*/
/*!
* \brief  SDL task function, assigning an AID for a new associated STA.
*
* An unallocated AID value will be assigned for a new STA, with the STA record
* properly initialized from the information of the \a asRdu association request
* frame.
* Special care is taken when the AID table is full. We will accept the newly
* associated STA, and reject an older STA by one of these case:
*       1) relinquishing any STA which has already deauthenticated but the
*          STA record is still maintained within a prescribed interval.
*       2) relinquishing the STA which has longest inactivity time, i.e.,
*          the quietest one.
*
* \param  sig_code  SIG_ReasocReq or SIG_AsocReq
* \param  asRdu     pointer of the received (re)association request frame
* \param  ssta_p    pointer for the STA record
* \param  sts       pointer for the returned StatusCode_t value
* \return available AID
*/
/*----------------------------------------------------------------*/
int task_assign_AId(i2220mgmt_t *priv_p, sta_info_t *SSta_p)
{
        int             aid;
        sta_info_t      **sta_pp;

        KDBG_ENTER;
        ASSERT(SSta_p);

        /* check if the STA has already associated before. */
        if (SSta_p->aid > 0) {
                sta_info_t      *sta_p;

                sta_p = priv_p->ap.sta_aid[SSta_p->aid - 1];
                ASSERT(sta_p==SSta_p);
                return SSta_p->aid;
        }

        /* allocate a free slot of AID table for the new STA record. */
        sta_pp = &priv_p->ap.sta_aid[0];
        for (aid = 1; aid <= x_MAX_AID; aid++) {
                if (*sta_pp == NULL)
                        break;
                sta_pp++;
        }
        if (aid > x_MAX_AID) {
                /* too many STA */
                return 0;
        }

        SSta_p->aid = aid;
        *sta_pp = SSta_p;

        KDBG_EXIT;
        return aid;
}

/*----------------------------------------------------------------*/
/*!
* \brief  SDL utility function, query the STA record.
*
* The STA record of address \a addr will be looked up first. If empty result,
* a new STA record will be created. This is in particular required between
* the authenticated and associated state, where many per-station information
* required maintained.
*
* \param  addr  the MAC address for the STA
* \return pointer of the STA record
*/
/*----------------------------------------------------------------*/
sta_info_t* sdlu_query_sta(i2220mgmt_t *priv_p, MacAddr_t addr)
{
        sta_info_t *sta_p;
        KDBG_ENTER;

        sta_p = task_sta_lookup(priv_p, addr);
        if (!sta_p)
                sta_p = create_sta(priv_p, addr);
        KDBG_EXIT;
        return sta_p;
}

/*----------------------------------------------------------------*/
/*!
* \brief  SDL task function, to find authentication algorithm.
*
* To find in the MIB of a matching find authentication algorithm from a
* received authentication frame.
*
* \param  alg  authentication algorithm in a received authentication frame
* \return the index of the found dot11AuthenticationAlgorithmsTable[]
*/
/*----------------------------------------------------------------*/
int task_find_authAlg(i2220mgmt_t *priv_p, AuthType_t alg)
{
        int     i;
        KDBG_ENTER;

        alg++;  // algorithm number in MIB is one above as in auth-frame
        for (i=0; i<MAX_NUM_AUTH_ALGORITHM; i++) {
                if (mib_import(dot11AuthenticationAlgorithmsTable)[i].dot11AuthenticationAlgorithmsEnable
                && (alg == mib_import(dot11AuthenticationAlgorithmsTable)[i].dot11AuthenticationAlgorithm))
                        return i;
        }
        KDBG_EXIT;
        return -1;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void sdlf_age_send_null(i2220mgmt_t *priv_p, sta_info_t *sta_p, int chk_sleep)
{
        Frame_t         *nullPkt;
        sta_info_t      *send_sta_p;
        int             i;

        if (sta_p == NULL) return;

        /* Send 2 NULL data frame to test */
        for (i = 0; i < 2; i++) {
                nullPkt = (Frame_t*) sdl_mkFrame(priv_p, sizeof(Frame_t),
                                SYN_null_frame, sta_p->asoc.adAddr,
                                mib_import(dot11MACAddress));
                if (nullPkt == NULL) break;

                if (chk_sleep) {
                        nullPkt->mac_ctl |= MACCTL_CHECK_SLEEP;
                }
                send_sta_p = (i == 0) ? NULL : sta_p;
                sdu_send_request(priv_p, (Frame_t *)nullPkt, send_sta_p,
                                                SYN_contention);

                KDBG_LOG_MLME(KDBG_DEBUG,"Aging: send NULL data to "
                        MACSTR "\n", MAC2STR(sta_p->asoc.adAddr));
        }
}

/*----------------------------------------------------------------*/
/*!
* \brief  AP house-keeping routine.
*
* This is a timer function periodically executed at interval \a SDL_TIMER_INTERVAL.
* It will perform the following activities:
*    1) removal of deauthenticated STA's, after duration of STA_DEAUTH_TEMPORAL.
*       Before this interval, the wlan administrator can still get some
*       information for this STA, such as statistical data.
*    2) TBD.
*
* \param  data  the timer function cookie, pointer of the i2220mgmt_t struct.
* \return none
*/
/*----------------------------------------------------------------*/
void ap_timer_func(unsigned long data)
{
#define FREE_STA_MAX_NUM        10

        i2220mgmt_t             *priv_p = (i2220mgmt_t *)data;
        struct list_head        *lst;
        sta_info_t              *sta_p, *freeSTAs[FREE_STA_MAX_NUM];
        int                     i = 0, chk_sleep;
        UINT32                  tmp32, temp32;

        ipn_memset(freeSTAs, 0, sizeof(freeSTAs));

        /* clear inactive STA record and too old frames */
        list_for_each(lst, &priv_p->ap.sta_chain.head) {
                sta_p = list_entry(lst, sta_info_t, sw.link);

                /* Check power saving queue of every STA */
                chk_sleep = TRUE;
                if (!(priv_p->timeTick & 7) && sta_p->aid > 0) {
                        FragSdu_t       *fsdu;
                        while (sdlQ_Length(&sta_p->psQ) != 0) {
                                sdlQ_First(&sta_p->psQ, fsdu);
                                if (check_timeout(fsdu->eol)) {
                                        chk_sleep = FALSE;
                                        sdlQ_dequeue(&sta_p->psQ, fsdu);
                                        sdlu_release_fsdu(fsdu);

                                        KDBG_LOG_MLME(KDBG_DEBUG,"psQ aging for "
                                                MACSTR "\n", MAC2STR(sta_p->asoc.adAddr));
                                }
                                else    break;
                        }
                }

                /* Inactivity time out = 90 seconds */
                if (check_timeout(sta_p->actTime + STA_MAX_INACTIVITY)) {
                        if (sta_p->aid > 0 && !sta_p->rsnBlockingTraffic) {

                                /* Send 2 NULL data frame to test */
                                sdlf_age_send_null(priv_p, sta_p, chk_sleep);

                                /* free sta will be processed when tx fail,
                                 * please check function isr_process_tx_done()
                                 */
                        }
                        else if (i < FREE_STA_MAX_NUM && sta_p->aid >= 0) {
                                /* Note: free_sta() can not be used here because
                                 * the list loop still needs it for the next element.
                                 */
                                freeSTAs[i++] = sta_p;
                        }
                } /* end of check_timeout inactivity */

        #if RX_MIC_PARTIAL_CHECK
                sta_p->micTick = 0;
        #endif
        } /* end of list_for_each */

        /* free sta records, which are marked at the above list loop */
        for (i = 0; freeSTAs[i] && i < FREE_STA_MAX_NUM; i++) {
                free_sta(priv_p, freeSTAs[i]);
        }

        /* check defragment buffer to release aging skb */
        sdl_release_ArAge(priv_p->ap.defrag_buf, kal_jiffies);

#if !WLAN_INCLUDE_DEBUG
        /* Check if the UFIFO checksum error occurs. */
        NIC_MCR_WRITE(priv_p, MCR_PRBCR, 0xA01);
        NIC_MCR_READ(priv_p, MCR_PRBSR, &tmp32);
        if (tmp32 & BIT(0)) {
                KDBG_LOG_RX(KDBG_WARN,"Fatal Error: UFIFO checksum error\n");
                priv_p->macResetUFIFOCount++; /* debug */
                i2220hw_reset_mac(priv_p);
        }
#endif
        /* update rx statistics counts to avoid count overflow (32bits) */
        if (++priv_p->timeTick >= COUNTER_READ_INTERVAL / SDL_TIMER_INTERVAL) {
                priv_p->timeTick = 0;
                NIC_MCR_READ(priv_p, MCR_RFECR, &temp32);
                mib_import(dot11FCSErrorCount) += temp32;

                NIC_MCR_READ(priv_p, MCR_RFFCR, &temp32);
                priv_p->hwStatistics.RcvFifoFulls += temp32;
        }

        set_ap_timer(priv_p); /* periodically */

#if WLAN_INCLUDE_DEBUG
#if (90*HZ) > SDL_TIMER_INTERVAL && (90*HZ) < COUNTER_READ_INTERVAL
        if ((priv_p->timeTick % ((90*HZ)/SDL_TIMER_INTERVAL)) != 0) return;
#endif
        NIC_MCR_READ(priv_p, MCR_SCR, &tmp32);
        printk("MCR_SCR=0x%08lx, membase=0x%08lx\n", tmp32, priv_p->membase);

        NIC_MCR_READ(priv_p, MCR_IER, &tmp32);
        NIC_MCR_READ(priv_p, MCR_ISAR, &temp32);
        printk("MCR_IER=0x%08lx, MCR_ISAR=0x%08lx\n", tmp32, temp32);
        printk("intEnable=0x%08lx, qEnable=0x%08lx\n",
                priv_p->interruptEnable, priv_p->qInterruptEnable);

        NIC_MCR_READ(priv_p, MCR_QSR0, &tmp32);
        NIC_MCR_READ(priv_p, MCR_QSR1, &temp32);
        printk("MCR_QSR0=0x%08lx, MCR_QSR1=0x%08lx\n", tmp32, temp32);
#endif

        return;
}

/*----------------------------------------------------------------*/
/*!
* \brief  register to LINUX a request of timer service.
*
* A one-shot timer will be registered to the LINUX kernel. The periodicity is
* achieved via repeatedly invoking this function in the timer function
* \a ap_timer_func.
*
* \param  priv_p   pointer to i2220mgmt_t, the device's private structure
* \return none
*/
/*----------------------------------------------------------------*/
void set_ap_timer(i2220mgmt_t *priv_p)
{
        mgmt_timer_t    *tmr = &priv_p->ap.sdl_timer;

        mgmt_stop_timer(tmr);
        mgmt_start_timer(tmr, SDL_TIMER_INTERVAL,
                     (mgmt_timeout_func) ap_timer_func,
                     (unsigned long) priv_p);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param  priv_p   pointer to i2220mgmt_t, the device's private structure
* \return none
*/
/*----------------------------------------------------------------*/
void update_bss_operating(i2220mgmt_t *priv_p)
{
        struct list_head        *lst;
        sta_info_t              *sta_p;

        /* Re-calculate these operating parameters */
        priv_p->sta_nonERP = 0;
        priv_p->sta_LngPreamble = 0;
        priv_p->sta_LngSlotTime = 0;

        list_for_each(lst, &priv_p->ap.sta_chain.head) {
                sta_p = list_entry(lst, sta_info_t, sw.link);
                if (sta_p->aid <= 0) continue;

                /* Only assocated STA are included */
                if (sta_p->incNonERP > 0)
                        priv_p->sta_nonERP++;
                if (sta_p->incLngPreamble > 0)
                        priv_p->sta_LngPreamble++;
                if (sta_p->incLngSlotTime > 0)
                        priv_p->sta_LngSlotTime++;
        }

        /* Check if preamble type should be changed */
        if (priv_p->preambleMode == PREAMBLE_MODE_AUTO) {
                if (priv_p->sta_LngPreamble > 0 &&
                    priv_p->phyPreambleType == phyShortPreamble) {
                        priv_p->phyPreambleType = phyLongPreamble;
                        hw_set_rate_table(priv_p);
                }
                else if (priv_p->sta_LngPreamble == 0 &&
                         priv_p->phyPreambleType == phyLongPreamble) {
                        priv_p->phyPreambleType = phyShortPreamble;
                        hw_set_rate_table(priv_p);
                }
        }

        /* Check if slot time should be changed */
        if (priv_p->sta_LngSlotTime > 0 &&
            priv_p->phySlotType == phyShortSlot) {
                priv_p->phySlotType = phyLongSlot;
                pauChangeSlotTime(priv_p, SLOT_TIME_ERP_LONG,
                                priv_p->x_CfbBurst ? FALSE : TRUE);
        }
        else if ((priv_p->supportRates & RATE_SET_MASK_ERP_OFDM) &&
                 priv_p->sta_LngSlotTime == 0 &&
                 priv_p->phySlotType == phyLongSlot) {
                priv_p->phySlotType = phyShortSlot;
                pauChangeSlotTime(priv_p, SLOT_TIME_ERP_SHORT,
                                priv_p->x_CfbBurst ? FALSE : TRUE);
        }
}

/*----------------------------------------------------------------*/
/*!
* \brief  SDL utility function, to initialize the AP record structure.
*
* The AP record is initialized, the AID table is allocated from the module
* configurable parameter \a x_MAX_AID.
* The AP house-keeping function will also get started.
*
* \param  priv_p   pointer to i2220mgmt_t, the device's private structure
* \return none
*/
/*----------------------------------------------------------------*/
void sdlu_init_ap(i2220mgmt_t *priv_p)
{
        int             sz;
        ap_info_t       *ap = &priv_p->ap;

        sdlu_init_defrag(priv_p);

        sdlQ_emptyQ(&priv_p->bcQ);       /* broadcast power save queue */
        init_chain_list(&ap->sta_chain); /* the STA list chain */

        sz = x_MAX_AID * sizeof(sta_info_t*);
        ap->sta_aid = (sta_info_t**)kal_malloc(sz);
	if (ap->sta_aid == NULL) {
		printk("sdlu_init_ap: malloc fail\n");
		return;
	}
        ipn_memset(ap->sta_aid, 0, sz);

        /* Initialize timer structures */
        mgmt_init_timer(&priv_p->ap.sdl_timer);
        mgmt_init_timer(&priv_p->scan_timer);
        mgmt_init_timer(&priv_p->tx_timer);

#if SUPPORT_IPNRSN
        mgmt_init_timer(&priv_p->sm_timer);
#endif

#if SUPPORT_MSSID
        /* Enable primary SSID */
        priv_p->ap.ssid_list[0].enabled = TRUE;
        priv_p->ap.ssid_list[0].ssid_len = DEFAULT_SSID_LEN;
        ipn_memcpy(priv_p->ap.ssid_list[0].ssid, DEFAULT_SSID,DEFAULT_SSID_LEN);
#endif
        /* Start up the AP watchdog timer */
        set_ap_timer(priv_p);
}

/*----------------------------------------------------------------*/
/*!
* \brief  SDL utility function, to cleanup the AP record structure.
*
* Each STA record of this AP will be released one by one, and the AID
* table will also released.
* Note the AP record itself is part of the \a i2220mgmt_t structure
* and thus can't be freed here.
* The AP house-keeping function will also get stopped.
*
* \param  priv_p   pointer to i2220mgmt_t, the device's private structure
* \return none
*/
/*----------------------------------------------------------------*/
void sdlu_destroy_ap(i2220mgmt_t *priv_p)
{
        ap_info_t       *ap = &priv_p->ap;
        sta_info_t      *sta_p;
        FragSdu_t       *fsdu;
        int             i;

        /* delete timer first to prevent timer_bh invoking our
         * timer function while we are destroying our AP resources.
         */
#if SUPPORT_IPNRSN
        mgmt_del_timer(&priv_p->sm_timer);
#endif
        mgmt_del_timer(&priv_p->tx_timer);
        mgmt_del_timer(&priv_p->scan_timer);
        mgmt_del_timer(&priv_p->ap.sdl_timer);
        if (priv_p->chnl_info_p) {
                kal_free(priv_p->chnl_info_p);
                priv_p->chnl_info_p = NULL;
        }

        /* free defrag frame */
        sdlu_destroy_defrag(priv_p);

        /*----------------------------------------
         * here: free hash table and its PS queue
         *---------------------------------------*/

        /* cleanup the STA records and AID array */
        for (i = 0; !list_empty(&ap->sta_chain.head); i++) {
                peek_chain_head(&ap->sta_chain, sta_info_t, sta_p);
                if (!sta_p)  break;

                if (i >= x_MAX_AID) {
                        KDBG_LOG_INIT(KDBG_WARN,"More STAs than expected, %08lx\n",
                                (unsigned long)sta_p);
                }
                free_sta(priv_p, sta_p);
        }
        kal_free(ap->sta_aid);  /* free AID table */

        /* now clean up each SduQueue */
        while (sdlQ_Length(&priv_p->bcQ) != 0) {
                sdlQ_dequeue_first(&priv_p->bcQ, fsdu);
                sdlu_release_fsdu(fsdu);
        }
}

#ifdef IPN_NEW_KAL
/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
UINT32 __swap32(UINT32 i)
{
        UINT32  tmp = 0;

        tmp |= ((i & 0xFF000000) >> 24);
        tmp |= ((i & 0x00FF0000) >> 8);
        tmp |= ((i & 0x0000FF00) <<8);
        tmp |= ((i & 0x000000FF) <<24);

        return tmp;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
UINT16 __swap16(UINT16 i)
{
        UINT16  tmp = 0;

        tmp |= ((i & 0xFF00) >> 8);
        tmp |= ((i & 0x00FF) <<8);

        return tmp;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
#if 0  // Sam marked
void* ipn_memcpy(void* dest, void* src, int len)
{
        char *tmp = (char *) dest, *s = (char *) src;

        while (len--)
                *tmp++ = *s++;
        return dest;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void* ipn_memset(void* dest, int value, int len)
{
        char *xs = (char *) dest;

        while (len--)
                *xs++ = value;
        return dest;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int ipn_memcmp(void* dest, void* src, int len)
{
        const unsigned char     *su1, *su2;
        int                     res = 0;

        for( su1 = dest, su2 = src; 0 < len; ++su1, ++su2, len--)
                if ((res = *su1 - *su2) != 0)
                        break;
        return res;
}
#endif /* Sam marked */
#endif /* end of IPN_NEW_KAL */

