/*
** $Id: wep.c 1.19 2004/06/10 16:00:35 cmchang Exp $
*/

/*************************************************************************
* Copyright (c) 2002-2003 Inprocomm, Inc.
*
* All rights reserved. Copying, compilation, modification, distribution
* or any other use whatsoever of this material is strictly prohibited
* except in accordance with a Software License Agreement with
* Inprocomm, Inc.
**************************************************************************/
/*************************************************************************
GENERAL DESCRIPTION
********************
The WEP is implemented according to ANSI/IEEE Std 802.11, 1999 edition,
Clause 8.
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/

/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/
/*__________________________________________________________________________
** wepHwMpduEncapsulate
** description: this function encapsulate IEEE 802.11 frame body according
**              to ANSI/IEEE Std 802.11, 1999 edition, clause 8.
** parameters:  adapter_p - Pointer to the adapter object data area.
**              frameBody_p, pointer to input text
**              frameBodyLen, length of frame body (in bytes)
**              keyId, WEP key index number, 0~3: global key, 4~15: per-clinet
**                     mapping key
**              key_p, pointer to WEP key
**              keyLen, length of WEP key (RC4 key length)
**              resultFrameBodyLen_p, pointer to set the result length of
**                                    encapsulated frame
** return     : WEP_FRAME_FORMAT_ERROR,  if the frame body exceeds 2312 bytes.
**              WEP_ENCAPSULATE_SUCCESS, if WEP MPDU encapsulation success
** note:        1. After the frame is encapsulated, the frame length is expand
**                 by 8-bye, 4 for IV, and 4 for ICV.
**              2. The original PDU and expanded ICV are encrypted based on
**                 RSA RC4
**___________________________________________________________________________*/
UCHAR
wepHwMpduEncapsulate(IN PMP_ADAPTER adapter_p,
                     IN OUT PTFCB_STRUC  tfcb_p,
                     IN ULONG keyId,
                     IN PUCHAR key_p,
                     IN ULONG keyLen,
                     IN UCHAR cipher_suite)
{
    UCHAR   iv_KeyId;
    UINT    i;

    DEBUGFUNC("wepHwMpduEncapsulate");
    DBGPRINT(RSN, TRACE, ("wepHwMpduEncapsulate().\n"));

    /* Generate random IV to set into SCI of TFCB */
    tfcb_p->u_sciiv.sciIv = getRandomNumber(adapter_p);

    if (keyId < 4) { /* default key */
        /* add key ID and padding zeros to IV, in byte 3 */
        iv_KeyId= (UCHAR)keyId << 6;
    }
    else { /* mapping key */
        /* set key ID to 0 */
        iv_KeyId = 0; /* mapping key. so, set key ID to 0 */
    }
    tfcb_p->u_sciiv.s.keyId = iv_KeyId;
    tfcb_p->iv = tfcb_p->u_sciiv.sciIv;
    DBGPRINT(RSN, LOUD, (" IV = 0x%08X\n", tfcb_p->u_sciiv.sciIv));

    /* add length of DMA FIFO header with IV (TCI + MAC Header + IV). */
    tfcb_p->dmaFifoHeaderLen += 1;     /* size in unit of DWORD(4 bytes) */
    DBGPRINT(RSN, LOUD, (" Fifo Length = %d\n", tfcb_p->dmaFifoHeaderLen));

    /* set Security Operation Mode for SCI of TFCB */
    tfcb_p->secOpMode = cipher_suite;
    DBGPRINT(RSN, LOUD, (" Cipher Suite = 0x%02X\n", tfcb_p->secOpMode));

    /* fill up the Encryption Key(16 bytes) */
    DBGPRINT(RSN, LOUD, (" swTfcb_p->tfcb_p->encryptionKey ==>\n"));
    for (i = 0; i < keyLen; i++) {
        tfcb_p->encryptionKey[i] = (UCHAR) *(key_p + i);
        DBGPRINT_RAW(RSN, LOUD, ("[%02X]", tfcb_p->encryptionKey[i]));
    }
    DBGPRINT(RSN, LOUD, ("\n"));

    /* MPDU Length + IV(4 bytes) + ICV(4 bytes) */
    tfcb_p->mpduLen += 8;
    DBGPRINT(RSN, LOUD, (" MPDU Length = %d\n", tfcb_p->mpduLen));

    return WEP_ENCAPSULATE_SUCCESS;
}   /* End of wepHwMpduEncapsulate */


