/*
** $Id: wlan_kal.c 1.22 2004/09/23 19:28:08 cmchang Exp $
*/

/*! \file wlan_kal.c
    \brief subroutines to support IPN2220.o

    This is a set of subroutines.
*/

/******************************************************************************
Copyright (c) 2002-2003 Inprocomm, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with Inprocomm, Inc.
******************************************************************************/

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "wlan_os.h"
#include "wlan_kal.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/


/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/

/************************************************************************
*                             M A C R O S
*************************************************************************
*/
#if 0 // refer to the TC's eth.c
#define UINT32 unsigned long
#define SWAP32(x) ((UINT32)((((UINT32)(x)&(UINT32)0x000000ffUL)<<24)|(((UINT32)(x)&(UINT32) 0x0000ff00UL)<<8)|(((UINT32)(x) & (UINT32) 0x00ff0000UL)>>8)|(((UINT32)(x) & (UINT32) 0xff000000UL) >> 24) ))
#else
#define ROL(x,n)                (((x)<<(n))|((x)>>(32-(n))))
#define SWAP32(x)    ((ROL(x,8) & 0x00ff00ff) | ROL(x & 0x00ff00ff,24))
#endif 

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/
extern int linux_handle_8021X(struct net_device *dev, struct sk_buff* skb);

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

unsigned long kal_ticks(void)
{
        return jiffies;
}
/*----------------------------------------------------------------*/
/*!
* \brief  This function is called in both interrupt routine and
*         general routine.
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void* kal_malloc(int sz)
{
        /* kmalloc allocation flag. Please note this flag is used
         * in both interrupt routine and general routine.
         * In Linux, use GFP_ATOMIC.
         */
#if 0 //def CONFIG_TRENDCHIP_TC3162P
        return kmalloc(sz, GFP_KERNEL);
#else
        return /*KSEG1ADDR*/(kmalloc(sz, GFP_ATOMIC));
#endif
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_free(void* m)
{
        kfree(m);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void* kal_alloc_skb(int sz)
{
#ifdef CONFIG_CPU_S3C2510    /* Samsung ARM9, s3c2510x */
        return __dev_alloc_skb(sz, GFP_ATOMIC|GFP_DMA);
#else
        return dev_alloc_skb(sz);
#endif
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_free_skb(void* skb)
{
        dev_kfree_skb(skb);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
unsigned char* kal_skb_put(void* skb,int len)
{
        return skb_put(skb,len);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
unsigned char* kal_skb_push(void* skb,int len)
{
        return skb_push(skb,len);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
unsigned char* kal_skb_pull(void* skb,int len)
{
        return skb_pull(skb,len);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
unsigned char* kal_skb_data(void* skb)
{
        struct sk_buff* linux_skb = (struct sk_buff*)skb;
        return (unsigned char*)(linux_skb->data);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
unsigned int     kal_skb_len(void*skb)
{
        struct sk_buff* linux_skb = (struct sk_buff*)skb;
        return linux_skb->len;
}



/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_udelay(int t)
{
        udelay(t);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_mdelay(int t)
{
        mdelay(t);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void* kal_dev_priv(void *dev)
{
        struct net_device* linux_dev = (struct net_device*)dev;

        if(linux_dev)
                return (linux_dev)->priv;
        else
                return NULL;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_set_dev_macAddr(void *dev,unsigned char* mac)
{
        struct net_device* linux_dev = (struct net_device*)dev;

        if(linux_dev)
                memcpy((linux_dev)->dev_addr,mac,6);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void* kal_pci_alloc_sz(void* pci_p,int sz,unsigned long* dma_p)
{
        return pci_alloc_consistent(pci_p,(size_t)sz,(dma_addr_t *)dma_p);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_pci_free_sz(void* pci_p,int sz,void* m,unsigned long dma)
{
        pci_free_consistent(pci_p,(size_t)sz,m,(dma_addr_t )dma);
        return;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
unsigned long kal_pci_map_single(void* pcidev,void* mem,int size, int dir)
{
#ifdef CONFIG_CPU_S3C2510    /* Samsung ARM9, s3c2510x */
        return virt_to_bus(mem);
#else
        return pci_map_single((pcidev),(mem),(size),dir);
#endif
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_pci_unmap_single(void* pcidev,unsigned long bus,int size,int dir)
{
#ifdef CONFIG_CPU_S3C2510    /* Samsung ARM9, s3c2510x */
        return ;
#else
        pci_unmap_single((pcidev),(bus),(size),dir);
#endif
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
unsigned long kal_pci_readl(unsigned long membase, int offset){
#ifdef CONFIG_CPU_S3C2510    /* Samsung ARM9, s3c2510x */
        return PCIMemRead32(__mem_pci((unsigned long)(membase) + (offset)));
#else
#ifdef CONFIG_TRENDCHIP_TC3162P
//        return SWAP32(readl((unsigned long)(0xa0000000|membase) + (offset)));
	  return SWAP32(readl(KSEG1ADDR(membase+offset)));
#else
        return readl((unsigned long)(membase) + (offset));
#endif
#endif
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_pci_writel(unsigned long _value, unsigned long membase, int offset)
{
#ifdef CONFIG_CPU_S3C2510    /* Samsung ARM9, s3c2510x */
        PCIMemWrite32(__mem_pci((unsigned long)(membase) + (offset)), _value);
#else
#ifdef CONFIG_TRENDCHIP_TC3162P
        //writel(_value, (unsigned long)(membase) + (offset));
	// Sam add
//        writel(SWAP32(_value), (unsigned long)(0xa0000000|membase) + (offset));
	writel(SWAP32(_value), KSEG1ADDR(membase+offset));
#else 
        writel(_value, (unsigned long)(membase) + (offset));
#endif
#endif
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_netdev_rx(void* driver_skb, void * driver_dev)
{
        struct sk_buff* skb = (struct sk_buff*)driver_skb;
        struct net_device* dev = (struct net_device*)driver_dev;

//printk("Sam kal 387 kal_netdev_rx\n");
#if SUPPORT_IPNRSN
        /* 1x frame is processed by driver itself, and this 1x packet
         * will freed in linux_handle_8021X().
         */
        if (skb->data[12] == 0x88 && skb->data[13] == 0x8e) {
                linux_handle_8021X(dev, skb);
                return;
        }
#elif SUPPORT_WPA2
        if (skb->data[12] == 0x88 && (skb->data[13] == 0x8e || skb->data[13] == 0xc7)) {
                linux_handle_8021X(dev, skb);
                return;
        }
#endif
        skb->protocol = eth_type_trans(skb, dev);
        skb->dev = dev; /* Important, crash if not correct */
        netif_rx(skb);
}

#if !SUPPORT_IPNRSN & !SUPPORT_WPA2
/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_disassoc_indicate(void *dev, unsigned char *macAddr)
{
        return; /* Nothing to do in Linux */
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int kal_auth_request(void *dev, unsigned char *macAddr)
{
        return 0;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int kal_assoc_request(void *dev, unsigned char *macAddr,
                unsigned char *infoElem, int infoLen)
{
        return 0;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
void kal_tkip_micFailure(void *dev, unsigned char *macAddr)
{
        return;
}
#endif /* end of !SUPPORT_IPNRSN */

