/*
** $Id: wlan_os.c 1.28 2004/10/04 17:45:04 cmchang Exp $
*/

/*! \file wlan_os.c
    \brief subroutines to support Linux OS

    This is a set of subroutines.
*/

/******************************************************************************
Copyright (c) 2002-2003 Inprocomm, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with Inprocomm, Inc.
******************************************************************************/

/************************************************************************
*                     C O M P I L E R   F L A G S
*************************************************************************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "wlan_os.h"
#include "wlan_kal.h"
#include "wlan_lib.h"
#include "ap_wext.h"

#define TC3162P_LIGHT_WLAN_LED 1
#if TC3162P_LIGHT_WLAN_LED
#include "asm/tc3162/gpio.h"
unsigned int ledcount=0,dataflow=0;
#endif
#include <asm/tc3162/tc3162.h>


extern int x_MAX_AID;           /* Max number of supported STA */
                                /* Value Range: 1 ~ 128 */

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/* IPN2220/5320 802.11 54 Mb/s WLAN controller. */
#define PCIVENDOR_INPROCOMM     0x17FEUL
#define PCIDEVICE_IPN2220       0x2220UL
#define PCIDEVICE_IPN5320       0x5320UL

#define WLAN_MEM_MAPPED_LENGTH  0x800   /* memory-mapped length */

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/
typedef void (*do_isr_call_back)(unsigned long);

/* Tasklet mechanism is like buttom-half in Linux. We just want to
 * send a signal to OS for interrupt defer processing. All resources
 * are NOT allowed reentry, so txPacket, ISR-DPC and ioctl must avoid preempty.
 */
typedef struct wlandev_info_s {
        struct net_device       *dev;
        struct tasklet_struct   tasklet;
} wlandev_info_t;

/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/*----------------------------------------------------------------*/
/* --Module Parameters */

char    *x_ifname="wlan";       /*!< ifconfig interface name */
extern int tc_processor;

MODULE_PARM( x_ifname, "s");
MODULE_PARM( x_MAX_AID, "i");

MODULE_AUTHOR("Inprocomm brilliant AP team");
MODULE_DESCRIPTION("Inprocomm IPN2220 802.11ag AP, Build:"__DATE__" "__TIME__);
MODULE_SUPPORTED_DEVICE("Inprocomm IPN2220/5320");
#ifdef MODULE_LICENSE
MODULE_LICENSE("Inprocomm License Agreement");
#endif

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/
static struct pci_device_id pci_id_tbl[] __devinitdata = {
        {
                PCIVENDOR_INPROCOMM, PCIDEVICE_IPN2220,
                PCI_ANY_ID, PCI_ANY_ID,
                0, 0,
                /* Driver data, we just put the name here */
                (unsigned long)"Inprocomm IPN2220 802.11g WLAN Adapter"
        },
        {
                PCIVENDOR_INPROCOMM, PCIDEVICE_IPN5320,
                PCI_ANY_ID, PCI_ANY_ID,
                0, 0,
                /* Driver data, we just put the name here */
                (unsigned long)"Inprocomm IPN5320 802.11ag WLAN Adapter"
        },
        {
                0, 0, 0, 0, 0, 0, 0
        }
};
// Sam add 
MODULE_DEVICE_TABLE(pci,pci_id_tbl);
// End of Sam add

static wlandev_info_t           wlandev_array[MAX_WLAN_DEVICES];
static int                      ipn2220_wlan_bar = 2;
static int                      ipn2220_dev_num = 0;
static struct timer_list        ipn2220_timer_list;

/************************************************************************
*                             M A C R O S
*************************************************************************
*/
#ifdef CONFIG_CPU_S3C2510
#define pci_request_mem_region(a,b,c)   request_mem_pci_region(a,b,c)
#define pci_release_mem_region(a,b)     release_mem_pci_region(a,b)
#else
#define pci_request_mem_region(a,b,c)   request_mem_region(a,b,c)
#define pci_release_mem_region(a,b)     release_mem_region(a,b)
#endif

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/
extern int init_dev1x(struct net_device *apdev);
extern void rm_dev1x(struct net_device *apdev);
extern int wext_giwpriv(struct net_device *dev, struct iw_point *data);
extern void init_2220_proc(struct net_device *dev, char *dev_name);
extern void remove_2220_proc(struct net_device *dev, char *dev_name);

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*----------------------------------------------------------------*/
/*!
* \brief  Allocate a Interface index - unique device identifier, the
*         net_device::ifindex member for this wlan device.
*
* Allocates a device number and constructs the name for the given
* wlan dev
*
* \param   dev  pointer to kal_net_device_t
* \return  The device number on success, -1 otherwise
*/
/*----------------------------------------------------------------*/
static int wlan_get_index(struct net_device *dev)
{
        int     i;

        for  (i = 0; i < MAX_WLAN_DEVICES; i++) {
                if ( wlandev_array[i].dev == NULL ) {
                        /* Reserve 2 bytes space to store one digit of
                         * device number and NULL terminator
                         */
                        strncpy(dev->name, x_ifname, IFNAMSIZ - 2);
                        sprintf(dev->name + strlen(dev->name), "%d", i);
                        wlandev_array[i].dev = dev;
                        tasklet_init(&wlandev_array[i].tasklet,
                                (do_isr_call_back) do_isr_tasklet, 0);

                        ipn2220_dev_num++;
                        return i;
                }
        }

        return -1;
}

/*----------------------------------------------------------------*/
/*!
* \brief  Deallocate the Interface index - net_device::ifindex member.
*
* Frees a previously allocated device number.
*
* Arguments:
*       wlandev         ptr to the wlandev structure for the
*                       interface.
* \param   dev  pointer to kal_net_device_t
* \return  none
*/
/*----------------------------------------------------------------*/
static void wlan_clear_index(struct net_device *dev)
{
        int     i;

        for  (i = 0; i < MAX_WLAN_DEVICES; i++) {
                if ( wlandev_array[i].dev == dev ) {
                        tasklet_kill(&wlandev_array[i].tasklet);
                        wlandev_array[i].dev = NULL;

                        ipn2220_dev_num--;
                }
        }

        return;
}

/*----------------------------------------------------------------*/
/*!
* \brief  Register the allocated struct net_device object to the Linux NET4
*         protocol suite.
*
* Roughly matches the functionality of register_netdev.  This function
* is called after the driver has successfully probed and set up the
* resources for the device.  It's now ready to become a named device
* in the Linux system.
*
* First we allocate a name for the device (if not already set), then
* we call the Linux function register_netdevice.
*
* \param   dev  pointer to kal_net_device_t
* \return  zero on success, non-zero otherwise.
*/
/*----------------------------------------------------------------*/
static int wlan_net_register(struct net_device *dev)
{
        int             i;

        rtnl_lock();

        if ((i = wlan_get_index(dev)) < 0) {
                rtnl_unlock();
                return -EDQUOT;
        }

        if (register_netdevice(dev)) {
                wlan_clear_index(dev);
                rtnl_unlock();
                return -EIO;
        }
        rtnl_unlock();
        return i; /* success */
}

/*----------------------------------------------------------------*/
/*!
* \brief  Register the allocated struct net_device object to the Linux NET4
*         protocol suite.
*
* Roughly matches the functionality of unregister_netdev.  This
* function is called to remove a named device from the system.
*
* First we tell linux that the device should no longer exist.
* Then we remove it from the list of known wlan devices.
*
* \param   priv_p  pointer to i2220mgmt_t
* \return  zero on success, non-zero otherwise.
*/
/*----------------------------------------------------------------*/
static void wlan_net_unregister(struct net_device *dev)
{
        rtnl_lock();

        if (dev) {
                wlan_clear_index(dev);
                unregister_netdevice(dev);
        }
        rtnl_unlock();
}

/*----------------------------------------------------------------*/
/*!
* \brief  A method of struct net_device, initiates the netdevice
*
* Called in response to register_netdev.
*
* This is usually the probe function. At system boot, the network
* initialization code loops through all the net_device structures and
* calls their probing (>init ) functions by passing them a pointer to the
* device itself. If the probe function succeeds, the kernel initializes the
* next available net_device structure to use that interface. This way of
* setting up drivers permits incremental assignment of devices to the names
* eth0 , eth1 , and so on, without changing the name field of each device.
*
* However, since the probe has already been done elsewhere. All we do here
* is return success.
*
* \param   netdev   Linux netdevice.
* \return  0    for success
*/
/*----------------------------------------------------------------*/
int wlan_init( struct net_device *netdev)
{
        return 0;
}

/*----------------------------------------------------------------*/
/*!
* \brief  A method of struct net_device, to open the network interface
*         for normal operation.
*
* Opens the interface. The interface is opened whenever ifconfig activates it.
* The open method should register any system resource it needs (I/O ports,
* IRQ, DMA, etc.), turn on the hardware, and increment the module usage
* count.
* However, in wlan situation that may not be entirely true, since we require
* to provide the driver of the wlan MLME-START inforamtion to start-up the
* device.
*
* \param   netdev   Linux netdevice.
* \return  zero on success, non-zero otherwise
*/
/*----------------------------------------------------------------*/
static int wlan_open(struct net_device *dev)
{
        MOD_INC_USE_COUNT;
        netif_start_queue(dev);

        return 0; /* success */
}

/*----------------------------------------------------------------*/
/*!
* \brief  A method of struct net_device, to close the network interface.
*
* Stops the interface. The interface is stopped when it is brought down;
* operations performed at open time should be reversed.
* Following this call, no frames should go up or down through this interface.
*
* \param   netdev   Linux netdevice.
* \return  zero on success, non-zero otherwise
*/
/*----------------------------------------------------------------*/
static int wlan_stop(struct net_device *dev)
{
        MOD_DEC_USE_COUNT;
        netif_stop_queue(dev);

        return 0; /* success */
}

/*----------------------------------------------------------------*/
/*!
* \brief  A method of struct net_device, indicates the MTU of the
*         interface has been changed.
*
* This function is in charge of taking action if there is a change in the MTU
* (maximum transfer unit) for the interface.
*
* \param  dev     netdevice struct
* \param  new_mtu the new MTU value
* \return  zero on success, a negative errno on failure.
*/
/*----------------------------------------------------------------*/
static int wlan_change_mtu(struct net_device *dev, int new_mtu)
{
        /* RX_SKB_BUF_SZ is rx buffer, 84 is overhead,
         * max mac_header(32)+llc(8)+cipher(20)+fcs(4)+reserved(20)
         */
        if (new_mtu < 0 || new_mtu > (RX_SKB_BUF_SZ - 84))
                return -EINVAL;
        dev->mtu = new_mtu;
        return 0;
}

/*----------------------------------------------------------------*/
/*!
* \brief  A method of struct net_device, indicates the interface MAC
*         address has been changed.
*
* Handles the ioctl for changing the MACAddress of a netdevice
*
* references: linux/netdevice.h and drivers/net/net_init.c
*
* \param  dev     netdevice struct
* \param  addr    the new MACAddress (a struct)
* \return  zero on success, a negative errno on failure.  Possible values:
*               -EBUSY  device is bussy (cmd not possible)
*               -and errors returned by: p80211req_dorequest(..)
*/
/*----------------------------------------------------------------*/
static int wlan_set_mac_address(struct net_device *dev, void *addr)
{
        struct sockaddr *new_addr = addr;
        return i2220ap_set_mac_address(dev, new_addr->sa_data);
}

/*----------------------------------------------------------------*/
/*!
* \brief  A method of struct net_device, to get the network interface
*         statistical information.
*
* Whenever an application needs to get statistics for the interface,
* this method is called. This happens, for example, when ifconfig or
* netstat -i is run.
*
* Here we're reporting the Linux i/f level statistics.  Hence, for the
* primary numbers, we don't want to report the numbers from the MIB.
* Eventually, it might be useful to collect some of the error counters though.
*
* \param   netdev   Linux netdevice.
* \return  the address of the statistics structure
*/
/*----------------------------------------------------------------*/
struct net_device_stats* wlan_get_stats(struct net_device *dev)
{
        static struct net_device_stats result;
        struct net_stat tmp_stat;
        memset(&tmp_stat, 0 ,sizeof(tmp_stat));

        param_get_Statistic(dev, &tmp_stat);
        result.rx_packets  = tmp_stat.rxfragments;
        result.rx_bytes    = tmp_stat.rxbytes;
        result.tx_packets  = tmp_stat.txfragments;
        result.tx_bytes    = tmp_stat.txbytes;
        result.rx_dropped  = tmp_stat.rxdiscardsnobuffer;
        result.rx_errors   = tmp_stat.rxdiscardswepundecr;

        return &result;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int wlan_do_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
        int             result = 0;
        struct iwreq    *iwr = (struct iwreq *)ifr;
//printk("Sam 433 cmd is %08x SIOCGIWPRIV=%08x\n",cmd,SIOCGIWPRIV);
        if (cmd == SIOCGIWPRIV) {
                return wext_giwpriv(dev, &iwr->u.data );
        }

#if SUPPORT_IPNRSN | SUPPORT_WPA2
        if((cmd >= WLAN_IOCTL_WLAN_PARAM) &&  (cmd <= WLAN_IOCTL_HOSTAPD)){
                return do_hostap_ioctl(dev, iwr,cmd);
        }
#endif

        if ((cmd >= SIOCIWFIRSTPRIV) && (cmd < SIOCIW_MAXIOCTL)) {
                return private_support_ioctl(dev, iwr, cmd);
        }

        return result;
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
int wlan_hard_start_xmit(struct sk_buff *skb, struct net_device *netdev)
{
#ifdef CONFIG_CPU_S3C2510    /* Samsung ARM9, s3c2510x */
        struct sk_buff  *skb_new;
#endif
        if ( netif_queue_stopped(netdev) ) {
                return 1;
        }

#ifdef CONFIG_CPU_S3C2510    /* Samsung ARM9, s3c2510x */
        /* copy to non-cache sk_buff */
        skb_new = kal_alloc_skb(skb->len);
        if (skb_new == NULL) {
                return 1;
        }
        kal_skb_put(skb_new, skb->len);
        memcpy(skb_new->data, skb->data, skb->len);
        kal_free_skb(skb);
        skb = skb_new;
#else
        /* We will modify the skb->data, so unshare it.
         * skb_unshare() can check if skb needs to be duplicated.
         */
        if ( (skb = skb_unshare(skb, GFP_ATOMIC)) == NULL ){
                return 1; 
        }
#endif
        netif_stop_queue(netdev);
        netdev->trans_start = jiffies;

        /* In this function, skb will be tx out or discarded */
        i2220ap_hard_start_xmit(netdev, skb);

        netif_wake_queue(netdev);

#if TC3162P_LIGHT_WLAN_LED
	dataflow = 1;
#endif

        return 0;
}

/*----------------------------------------------------------------*/
/*!
* \brief  Creating the private data structure associated with the
*         Linux NET4 struct net_device object.
*
* \return
*/
/*----------------------------------------------------------------*/
static struct net_device *wlan_net_create(struct pci_dev *pdev, unsigned long membase)
{
        struct net_device        *dev;
      
	dev = kal_malloc(sizeof(struct net_device));
        if (dev == NULL) { printk("wlan_net_create: malloc fail\n"); return NULL;}

        /* Setup netdevice's ability to report resources */
        memset(dev, 0, sizeof(struct net_device));
// Add by Sam
	membase = 0xA0000000 | membase;
// End Sam add
        dev->irq = pdev->irq;
        dev->mem_start = membase;
        dev->mem_end = membase + pci_resource_len(pdev, ipn2220_wlan_bar);
        dev->priv = i2220ap_priv_create(pdev, dev, membase);
        if (kal_dev_priv(dev) == NULL) {
                kal_free(dev);
                return NULL;
        }

        /* Initialize the network device objects. */
        ether_setup(dev);
        dev->init               = &wlan_init;
        dev->open               = &wlan_open;
        dev->stop               = &wlan_stop;
        dev->change_mtu         = &wlan_change_mtu;
        dev->get_stats          = &wlan_get_stats;
        dev->set_mac_address    = &wlan_set_mac_address;
        dev->do_ioctl           = &wlan_do_ioctl;
        dev->hard_start_xmit    = &wlan_hard_start_xmit;

        netif_stop_queue(dev);
        return dev;
}

/*----------------------------------------------------------------*/
/*!
* \brief  Destroying the Linux NET4 struct net_device object and the AP object.
*
* Called at module cleanup time. It should
* be called after mgmt_net_unregister.  Basically, all it does is
* free the 'struct device' that's associated with the priv_p.
*
* \param   priv_p  pointer to i2220mgmt_t
* \return  zero on success, non-zero otherwise.
*/
/*----------------------------------------------------------------*/
static void wlan_net_destroy(struct net_device *dev)
{
        if (dev) {
                i2220ap_priv_destroy(dev);
                kal_free(dev);
        }
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
static void wlan_timer_periodic(unsigned long data)
{
        struct timer_list *tmr = &ipn2220_timer_list;

        do_timer_check();
        tmr->expires = jiffies + HZ / 10;
        tmr->function = wlan_timer_periodic;
        add_timer(tmr);
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
static void wlan_timer_create(void)
{
        struct timer_list *tmr = &ipn2220_timer_list;

        /* Only create a periodic timer even though
         * multiple cards exist simultaneously.
         */
        if (ipn2220_dev_num == 1) {
                init_timer(tmr);
                tmr->expires = jiffies + HZ / 10;
                tmr->function = wlan_timer_periodic;
                add_timer(tmr);
        }
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
static void wlan_timer_destroy(void)
{
        if (ipn2220_dev_num <= 0) {
                del_timer(&ipn2220_timer_list);
        }
}

/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
static void wlan_hw_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
        void                    *priv_p;
        struct tasklet_struct   *tasklet_p;

        #if TC3162P_LIGHT_WLAN_LED
		ledcount ++;
                if(ledcount > 10 && dataflow == 1)
                {
                        unsigned short gpiodata;
                        gpiodata = *(volatile unsigned short*)TC3162_GPDATA_REG;
                        gpiodata ^= (1<<(GPIO_WIRELESS_LED/2));
                        *(volatile unsigned short*)TC3162_GPDATA_REG = gpiodata;
                        ledcount = dataflow =0;
                }
		else if(ledcount > 30)
		{
			unsigned short gpiodata;
                        gpiodata = *(volatile unsigned short*)TC3162_GPDATA_REG;
                        gpiodata &=~(1<<(GPIO_WIRELESS_LED/2));
                        *(volatile unsigned short*)TC3162_GPDATA_REG = gpiodata;
		}
        #endif

        priv_p    = kal_dev_priv(((wlandev_info_t*)dev_id)->dev);
        tasklet_p = &((wlandev_info_t*)dev_id)->tasklet;

        /* Check if this interrupt is ours. If it is true, issue
         * the event to defer process.
         */
        if ( i2220hw_interrupt(priv_p)) {
                tasklet_p->data = (unsigned long) priv_p;
                tasklet_schedule(tasklet_p);
        }
}

void tc3162_wlan_poll(void)
{
	int i;
	wlandev_info_t          *wlandev_info_p;
	if (tc_processor != 1) // processor is not TC3162P, get out here
		return;
	for (i=0; i< MAX_WLAN_DEVICES; ++i)
	{
		wlandev_info_p = &wlandev_array[i];
		if (wlandev_info_p->dev)
		        wlan_hw_interrupt(PCI_A_INT, (void *)wlandev_info_p, 0);
	}
}

EXPORT(tc3162_wlan_poll);
/*----------------------------------------------------------------*/
/*!
* \brief
*
* \param
* \return
*/
/*----------------------------------------------------------------*/
static void wlan_pci_latency(struct pci_dev *pdev)
{
        unsigned char   pci_latency;
        //unsigned char   new_latency = 8;
        unsigned char   new_latency = 40;

        /* Check the PCI latency value. It is better to set lhe latency timer
         * be larger than 8.
         */
        pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &pci_latency);
        if (pci_latency < new_latency) {
                pci_write_config_byte(pdev, PCI_LATENCY_TIMER, new_latency);
        }
}

/*----------------------------------------------------------------*/
/*!
* \brief  PCI probe call back function, called when a PCI device of matching
*         manufacture/device ID is found.
*
* This function initializes the device it is passed. Actually it is this place,
* the per instance of IPN2220 NIC card get started execution.
*
* \param   pdev            pointer to PCI device structure containing info about
*                          PCI configuration.
* \param   id              pointer to the device ID entry that matched this device.
* \return  zero when success, negative as failure
*/
/*----------------------------------------------------------------*/
static int __devinit wlan_probe_pci(struct pci_dev *pdev, const struct pci_device_id *id)
{
        int                     result = 0; /* success */
        int                     bar;        /* base address register */
        unsigned long           phymem, membase;
        struct net_device       *dev;
        int                     devIndex;
        wlandev_info_t          *wlandev_info_p;


        /* ----------------------------------------
         * Probe PCI and allocate system resources
         * ----------------------------------------
         */

        /* First, enable the PCI device */
//printk("Sam wireless 693 wlan_probe_pci\n");
        if (pci_enable_device(pdev)) {
                result = -EIO;
                goto fail;
        }
        // Sam resume following line from commenting out for adding latency
         wlan_pci_latency(pdev);

        /* Figure out our resources */
        for (bar=0; bar<6; bar++) {
                if ((pci_resource_flags(pdev,bar) & IORESOURCE_MEM) &&
                     pci_resource_len(pdev,bar) == WLAN_MEM_MAPPED_LENGTH)
                        break;
        }
        if (bar >= 6) {
                result = -EIO;
                goto fail;
        }
        phymem = pci_resource_start(pdev, bar);
//printk("Sam 713 phymem=%08x\n",phymem);
        if (!pci_request_mem_region(phymem, pci_resource_len(pdev, bar), "IPN2220")) {
                result = -EIO;
//printk("Sam 716\n");
                goto fail;
        }

        membase = (unsigned long)ioremap(phymem, pci_resource_len(pdev, bar));
	// debug membase = 0xbfba1000;
//printk("Sam 721 membase remap is %08x\n",membase);
        if ( membase == 0 ) {
                result = -EIO;
//printk("Sam 723\n");
                goto fail;
        }
        ipn2220_wlan_bar = bar; /* For pci_release_mem_region() */

        /* Enable PCI bus-mastering */
        pci_set_master(pdev);

        /* ----------------------------------------
         * Create Linux network and register
         * ----------------------------------------
         */
        dev = wlan_net_create(pdev, membase);
        if (dev == NULL) {
                result = -ENOMEM;
//printk("Sam 738\n");
                goto fail;
        }
        if ((devIndex = wlan_net_register(dev)) < 0) {
                wlan_net_destroy(dev);
                result = -EIO;
//printk("Sam 744\n");
                goto fail;
        }
#if SUPPORT_IPNRSN | SUPPORT_WPA2
        /* This function should be invoked after wlan_net_create()
         * because private struct of apdev is used in it.
         */
        init_dev1x(dev);
#endif
        /* -----------------------------------------
         * Initialize IPN2220 hardware
         * -----------------------------------------
         */
        if (i2220ap_pci_probe(dev) != 0) {
        #if SUPPORT_IPNRSN | SUPPORT_WPA2
                /* This function should be invoked before wlan_net_destroy()
                * because private struct of apdev is used in it.
                */
                rm_dev1x(dev);
        #endif
                wlan_net_unregister(dev);
                wlan_net_destroy(dev);

                /* We cannot access chip after iounmap() and release_mem */
                iounmap((void*)membase);
                pci_release_mem_region(pci_resource_start(pdev, ipn2220_wlan_bar),
                               pci_resource_len(pdev, ipn2220_wlan_bar));

                result = -EIO;
                goto fail;
        }

        /* ----------------------------------------
         * Create a periodic timer (100ms)
         * ----------------------------------------
         */
        wlan_timer_create();

#if WLAN_INCLUDE_PROC
        init_2220_proc(dev, dev->name);
#endif

        /* Build association between pdev and net_device struct */
        wlandev_info_p = &wlandev_array[devIndex];
        pci_set_drvdata(pdev, wlandev_info_p);

        /* Shouldn't actually hook up the IRQ until we
         * _know_ things are alright.  A test routine would help.
         */
        request_irq(pdev->irq, wlan_hw_interrupt,
                SA_SHIRQ, dev->name, (void *) wlandev_info_p);
fail:
        return result;
}

/*----------------------------------------------------------------*/
/*!
* \brief  PCI remove call back function, called when the PCI device should
*         be stopped.
*
* The remove method is used to tell the device driver that it should shut down
* the device and stop dealing with it, releasing any associated storage. The
* function is called either when the device is removed from the system or when
* the driver calls pci_unregister_driver in order to be unloaded from the system.
* Unlike probe, this method is specific to one PCI device, not to the whole set
* handled by this driver; the specific device is passed as an argument.
*
* \param   pdev            pointer to PCI device structure
* \return  none
*/
/*----------------------------------------------------------------*/
static void wlan_remove_pci(struct pci_dev *pdev)
{
        wlandev_info_t          *wlandev_info_p;
        struct net_device       *dev;
        unsigned long           mem_start;

        wlandev_info_p = pci_get_drvdata(pdev);
        if (wlandev_info_p == NULL) return;

        dev = wlandev_info_p->dev;
        if (dev == NULL) return;

 #if WLAN_INCLUDE_PROC
        remove_2220_proc(dev, dev->name);
#endif

       /* Free (Disable) IRQ to system */
        if (pdev->irq) {
                free_irq(pdev->irq, wlandev_info_p);
        }

        /* Remove wlan network by unregister. wlan_timer_destroy() should
         * be invoked after wlan_net_unregister() for multiple cards.
         */
        wlan_net_unregister(dev);

#if SUPPORT_IPNRSN | SUPPORT_WPA2
        /* This function should be invoked before wlan_net_destroy()
         * because private struct of apdev is used in it.
         */
        rm_dev1x(dev);
#endif
        wlan_timer_destroy();
        mem_start = dev->mem_start;

        /* Stop driver and remove all private data structs */
        i2220ap_remove_pci(dev);

        /* Free the netdev and netdev->priv memory */
        wlan_net_destroy(dev);

        /* We cannot access chip after iounmap() and release_mem */
        iounmap((void*)mem_start);
        pci_release_mem_region(pci_resource_start(pdev, ipn2220_wlan_bar),
                               pci_resource_len(pdev, ipn2220_wlan_bar));

        pci_disable_device(pdev);
        pci_set_drvdata(pdev, NULL);
}

/*----------------------------------------------------------------*/
/*!
* \brief  Driver entry point when the driver is configured as a Linux Module.
*
* Module initialization routine, called once at module load time, by the user-
* level modutils application: insmod or modprobe.
* In particular the invokation of \a pci_register_driver with the \a i2220_pci_drv_id
* will tell the PCI controller that when it detects our own PCI device, calls it
* via the \a probe call back function, from where every interesting stuffs of
* our driver started to life.
*
* \returns zero when success, non-zero for failure, module will be unloaded.
*/
/*----------------------------------------------------------------*/
static struct pci_driver i2220_pci_drv_id = {
        name:           "IPN_WLAN",
        probe:          wlan_probe_pci,
        remove:         __devexit_p(wlan_remove_pci),
        id_table:       pci_id_tbl,
};

static int __init init_module(void)
{
        /* We use printk only here and in cleanup_module, so that the module
         * loading/unloading will always get alerted no matter what the KDBG_
         * debugging setting is.
         */
	
	if (tc_processor != 1) // processor is not TC3162P, get out here
	{
		return -ENODEV;
	}
        printk(KERN_ALERT "\n"
               KERN_ALERT "\n"
               KERN_ALERT "=================================================================\n"
               KERN_ALERT "Loading ... %s %s %s %s %s\n", ipn2220_drv_ver,
                        " Build:", __DATE__, " ", __TIME__);

        /* Autogen of symbol reference for koops script */
        printk(KERN_ALERT "\n"
               KERN_ALERT "%08lx wlan_open [IPN_WLAN]\n" KERN_ALERT "\n",
               (unsigned long)   wlan_open);

        memset(wlandev_array, 0, sizeof(wlandev_array));

        /* This call will result in a call to wlan_probe_pci
         * if there is a matched PCI card present (ie., which
         * vendor+device id are matched)
         */
        if (pci_register_driver(&i2220_pci_drv_id) <= 0) {
                printk(KERN_ALERT "IPN2220AP: No device found.\n");
                pci_unregister_driver(&i2220_pci_drv_id);
                return -ENODEV;
        }
#if TC3162P_LIGHT_WLAN_LED
        else   // Sam add for light WLAN LED
        {
                unsigned int gpiomode;
                unsigned short gpiodata;                                                                                                                               
                // Step 0. set gpio 11 as output mode
                gpiomode = *(volatile unsigned int*)(TC3162_GPCTL_REG);
                gpiomode |= (1 << GPIO_WIRELESS_LED);
                *(volatile unsigned int*)TC3162_GPCTL_REG = gpiomode;

                // Step 1. enable open drain
                gpiodata = *(volatile unsigned short*)(TC3162_GPOPENDRAIN_REG);
                gpiodata |= (1 <<(GPIO_WIRELESS_LED/2));
                *(volatile unsigned short *)TC3162_GPOPENDRAIN_REG = gpiodata;

                //Step2. light wlan led on
                gpiodata = *(volatile unsigned short*)TC3162_GPDATA_REG;
                gpiodata &= ~(1<<(GPIO_WIRELESS_LED/2));
                *(volatile unsigned short*)TC3162_GPDATA_REG = gpiodata;
        }
#endif
        return 0;
}


/*----------------------------------------------------------------*/
/*!
* \brief  Driver exit point when the driver as a Linux Module is removed.
*
* Called at module unload time, by the user level modutils application: rmmod.
* This is our last chance to clean up after ourselves.
*
* \returns none
*/
/*----------------------------------------------------------------*/
static void __exit cleanup_module(void)
{
        pci_unregister_driver(&i2220_pci_drv_id);
        printk(KERN_ALERT "Unload .... %s %s %s %s %s\n", ipn2220_drv_ver,
                " Build:", __DATE__, " ", __TIME__);

        return;
}
/*********Merlin 20041104*********/
module_init(init_module);
module_exit(cleanup_module);
/********************************/
