/*
 *	Suppliment to iwpriv 
 *
 *		Jane Chang, Inprocomm, Inc.
 *      jane@inprocomm.com
 */

/******************************************************************************
Copyright (c) 2000-2003 Integrated Programmable Communications, Inc.

All rights reserved. Copying, compilation, modification, distribution
or any other use whatsoever of this material is strictly prohibited
except in accordance with a Software License Agreement with
Integrated Programmable Communications, Inc.
******************************************************************************/
/* Compile: 
     1. make wireless_tools
     2. gcc iwsta.c libiw.a -lm     (libiw.a is provided by wireless_tools)
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/

#include "iwlib.h"
/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/* Backward compatibility */
#ifndef IW_PRIV_TYPE_ADDR
#define IW_PRIV_TYPE_ADDR	0x6000
#endif	/* IW_PRIV_TYPE_ADDR */

#ifndef MAC2STR
#define MAC2STR(a) (a)[0], (a)[1], (a)[2], (a)[3], (a)[4], (a)[5]
#define MACSTR "%02x:%02x:%02x:%02x:%02x:%02x"
#endif

#define IOCTL_GET_STRUCT	SIOCIWFIRSTPRIV + 9

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/
typedef struct sta_status_ {
        short           aid;         /* STA's unique AID (1 .. 2007) or 0 if not yet assigned */
        unsigned char   adAddr[6];      /* address of associated station */
        unsigned char   auth_state;  /* station status */
        unsigned char   ps_state;    /* station power save status */
        unsigned long assoc_time;
} sta_status_t;

/*
 * Structure used by wlan to store the log.
 */
struct net_stat
{
        unsigned long   txunicastframes;
        unsigned long   txmulticastframes;
        unsigned long   txfragments	;
        unsigned long   txsingleretryframes;
        unsigned long   txmultipleretryframes;
        unsigned long   txretrylimitexceeded;
        unsigned long   txdiscards;
	unsigned long   txbytes;
        unsigned long   rxmulticastframes;
	unsigned long   rxbytes;
        unsigned long   rxfragments	;
        unsigned long   rxfcserrors;
        unsigned long   rxdiscardsnobuffer;
        unsigned long   rxdiscardswepundecr;
};

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/
int st_info_ioctl(int skfd,  char *	ifname)
{
        struct iwreq iwr;
        sta_status_t *staS;
        int i=0, n=10; /* Default number of station--10 should be enough */

        staS = malloc(sizeof(sta_status_t)*n);
	memset(staS,0,sizeof(sta_status_t)*n);
        memset(&iwr, 0, sizeof(iwr));

        strcpy(iwr.ifr_name,ifname);
        iwr.u.data.flags = 1; /* subcmd = StaList*/
        iwr.u.data.pointer = (caddr_t)staS;
        iwr.u.data.length = sizeof(sta_status_t)*n;

        if (ioctl(skfd, IOCTL_GET_STRUCT, &iwr) < 0) {
              perror("ioctl[IOCTL_GET_STRUCT]");
              return -1;
        }

        for(i=0;i<n;i++)
        {
                if (staS[i].aid) {
                        printf("MAC Addr:"MACSTR"\n",MAC2STR(staS[i].adAddr));
                }
                else
                        break;
        }
        return 0;
}

int ap_stat_ioctl(int skfd,  char *	ifname)
{
        struct iwreq iwr;
        struct net_stat tmp_stat;
        int i=0;
        memset(&tmp_stat,0,sizeof(tmp_stat));
        memset(&iwr, 0, sizeof(iwr));

        strcpy(iwr.ifr_name,ifname);
        iwr.u.data.flags = 50; /* subcmd = ap_statistic*/
        iwr.u.data.pointer = (caddr_t)&tmp_stat;
        iwr.u.data.length = sizeof(tmp_stat);

        if (ioctl(skfd, IOCTL_GET_STRUCT, &iwr) < 0) {
              perror("ioctl[IOCTL_GET_STRUCT]");
              return -1;
        }

        printf("txbyes: %d\n",tmp_stat.txbytes);
        printf("txunicastframes: %d\n",tmp_stat.txunicastframes);
        printf("txmulticastframes: %d\n",tmp_stat.txmulticastframes);
        printf("txfragments: %d\n",tmp_stat.txfragments);
        printf("txsingleretryframes: %d\n",tmp_stat.txsingleretryframes);
        printf("txmultipleretryfram: %d\n",tmp_stat.txmultipleretryframes);
        printf("txretrylimitexceede: %d\n",tmp_stat.txretrylimitexceeded);
        printf("txdiscards: %d\n",tmp_stat.txdiscards);
        printf("rxbyes: %d\n",tmp_stat.rxbytes);
        printf("rxmulticastframes: %d\n",tmp_stat.rxmulticastframes);
        printf("rxfragments: %d\n",tmp_stat.rxfragments);
        printf("rxfcserrors: %d\n",tmp_stat.rxfcserrors);
        printf("rxdiscardswepundecr: %d\n",tmp_stat.rxdiscardswepundecr);

        return 0;
}

int main(int    argc, char **    argv){
        int skfd;		/* generic raw socket desc.	*/
        int goterr = 0;

        /* Create a channel to the NET kernel. */
        if((skfd = iw_sockets_open()) < 0) {
                printf("open socket fail\n");
                return(-1);
        }

        if(!argv[1] || !argv[2]){
                printf("Please use ""iwsta [WLAN interface] [command] ""!.\n");
                printf("Available commands:stalist, apstat.\n");
                return -1;
        }
        
        if(!strncmp(argv[2], "stalist", 7))
                goterr = st_info_ioctl(skfd, argv[1]);
        else if(!strncmp(argv[2], "apstat", 6))
                goterr = ap_stat_ioctl(skfd, argv[1]);
        else{
                printf("Please use ""iwsta [WLAN interface] [command] ""!.\n");
                printf("Available commands:stalist, apstat.\n");
                return -1;
        }

        /* Close the socket. */
        close(skfd);

        return(goterr);
}

