/*
** $Id: pau.c 2.1 2004/10/04 18:07:03 cmchang Exp $
*/

/*************************************************************************
* Copyright (c) 2002-2004 Inprocomm, Inc.
*
* All rights reserved. Copying, compilation, modification, distribution
* or any other use whatsoever of this material is strictly prohibited
* except in accordance with a Software License Agreement with
* Inprocomm, Inc.
**************************************************************************/
/*************************************************************************
GENERAL DESCRIPTION
********************
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"


/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/
#define CCA_SHORT_SLOT_TIME_WITH_ED           0x0C
#define CCA_LONG_SLOT_TIME_WITH_ED            0x0A
#define CCA_SHORT_SLOT_TIME_NO_ED             0x04
#define CCA_LONG_SLOT_TIME_NO_ED              0x00


#define LED_BLINKING_ON_TIME_1          20
#define LED_BLINKING_OFF_TIME_1         6
#define LED_BLINKING_ON_TIME_2          20
#define LED_BLINKING_OFF_TIME_2         3

#define LCR_DEFAULT_VALUE       (LCR_LED_MODE_TX_RX | \
                                LCR_LED_POLARITY_POSITIVE | \
                                ((LED_BLINKING_ON_TIME_1 << 8) & LCR_LED_ON_CNT) | \
                                (LED_BLINKING_OFF_TIME_1 & LCR_LED_OFF_CNT))

#define CRFCR0_DEFAULT_VALUE    (CRFCR0_RSP_SAME_PREAMBLE | \
                                CRFCR_CTRL_RSP_TX_RATE_1M | \
                                (CRFCR_CTRL_RSP_TX_RATE_2M << 5) | \
                                (CRFCR_CTRL_RSP_TX_RATE_5_5M << 10) | \
                                (CRFCR_CTRL_RSP_TX_RATE_11M << 15) | \
                                (CRFCR_CTRL_RSP_TX_RATE_6M << 20) | \
                                (CRFCR_CTRL_RSP_TX_RATE_6M << 25))

#define CRFCR1_DEFAULT_VALUE    (CRFCR_CTRL_RSP_TX_RATE_12M | \
                                (CRFCR_CTRL_RSP_TX_RATE_12M << 5) | \
                                (CRFCR_CTRL_RSP_TX_RATE_24M << 10) | \
                                (CRFCR_CTRL_RSP_TX_RATE_24M << 15) | \
                                (CRFCR_CTRL_RSP_TX_RATE_24M << 20) | \
                                (CRFCR_CTRL_RSP_TX_RATE_24M << 25))


/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/


/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/
DEBUGFILE("pau.c");


const NIC_MCR_CONFIG_ENTRY pauInitTableCommon[] = {
#if MAC_ONLY_MODE
    {MCR_CTPR1,     0x00000001},
#endif
    {MCR_DRCR,      0x00010801},
    {MCR_EPDDR,     0x00640172},
    ////{MCR_DTR,       0x907b5cd7},
    //{MCR_DTR,       0x9078c8d7},
    {MCR_DTR,       0x9078f0d7},
    {MCR_RFBCR,     0x00027372},
    {MCR_BCWR,      0x0000001f},
    {MCR_AIFSR,     0x00e21112},
    {MCR_TQCWR,     0x03ff000f},
    {MCR_AC0CWR,    0x03ff000f},
    {MCR_AC1CWR,    0x03ff000f},
    {MCR_AC2CWR,    0x01ff000f},
    {MCR_AC3CWR,    0x00ff0007},
    {MCR_MRBS,      WLAN_MAX_PACKET_SIZE},
    {MCR_SCPR0,     0xc78f000F},
    {MCR_LCR,       LCR_DEFAULT_VALUE},
    {MCR_NRTR,      0x000f00ff},
    {0x008c,        0x80000000}
};


/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/


/*_________________________________________________________________________
**  pauSoftwareReset
**
**  descriptions: It resets the IPN2220. After reset, DMA, PAU are diable.
**                It also init MAC registers to its default value.
**                After reset the clock is enabled
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**  return:       (none)
**  note:
**__________________________________________________________________________*/
VOID
pauSoftwareReset(IN PMP_ADAPTER adapter_p)
{
    DEBUGFUNC("pauSoftwareReset");

    INITLOG(("\n"));

    //NIC_DISABLE_CLOCK_GATE(adapter_p);

    NIC_MCR_WRITE(adapter_p, MCR_SCR, SCR_GPIO_PD_N | SCR_CLK_ALL_EN |
        SCR_MAC_RST);

    NdisStallExecution(1);

    NIC_MCR_WRITE(adapter_p, MCR_SCR, SCR_GPIO_PD_N | SCR_CLK_ALL_EN | SCR_BB_MRST |
        SCR_BB_SIORST);

    NdisStallExecution(1);

    nicDisableInterrupt(adapter_p);
}   /* pauSoftwareReset */

/*_________________________________________________________________________
**  pauDisable
**
**  descriptions: This routine disable PAU
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**
**  return:       TRUE, if channel setting successful
**                FALSE, if channel setting false
**  note:
**__________________________________________________________________________*/
BOOLEAN
pauDisable(IN PMP_ADAPTER adapter_p)
{
    UINT_32 temp32;

    DEBUGFUNC("pauDisable");

    DBGLOG(HW, TRACE, ("\n"));


    NIC_MCR_READ(adapter_p, MCR_SCR, &temp32);
    //temp32 &= ~(PPMCR_PAU_ENABLE | PPMCR_POWER_SAVE_SWITCH);
    temp32 &= ~SCR_PAU_ENABLE;
    NIC_MCR_WRITE(adapter_p, MCR_SCR, temp32);

    NdisStallExecution(2);

    return TRUE;
}   /* pauDisable */

/*_________________________________________________________________________
**  pauEnable
**
**  descriptions: This routine enable PAU
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**
**  return:       TRUE, if channel setting successful
**                FALSE, if channel setting false
**  note:
**__________________________________________________________________________*/
BOOLEAN
pauEnable(IN PMP_ADAPTER adapter_p)
{
    UINT temp32;

    DEBUGFUNC("pauEnable");

    DBGLOG(HW, TRACE, ("\n"));

    NIC_MCR_READ(adapter_p, MCR_SCR, &temp32);
    //temp32 |= PPMCR_PAU_ENABLE | PPMCR_POWER_SAVE_SWITCH;
    temp32 |= SCR_PAU_ENABLE;
    NIC_MCR_WRITE(adapter_p, MCR_SCR, temp32);

    return TRUE;
}   /* pauEnable */

/*_________________________________________________________________________
**  pauMCRInit
**
**  descriptions: This routine initialize MCR
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**  return:       (none)
**  note:
**__________________________________________________________________________*/
VOID
pauMCRInit(IN PMP_ADAPTER adapter_p)
{
    UINT i;
    PNIC_MCR_CONFIG_ENTRY mcrList_p;
    UINT mcrListSize;
#if DBG
    UINT_32 temp32;
#endif

    DEBUGFUNC("pauMCRInit");

    INITLOG(("\n"));

    mcrListSize = sizeof(pauInitTableCommon) / sizeof(NIC_MCR_CONFIG_ENTRY);

    INITLOG(("Initialize PAU with common PAU init table\n"));

	for (i = 0; i < mcrListSize; i++) {
        DBGLOG(INIT, INFO, ("MCR write: offset=0x%04x, value=0x%08x\n",
            pauInitTableCommon[i].offset, pauInitTableCommon[i].value));

        NIC_MCR_WRITE(adapter_p, pauInitTableCommon[i].offset,
            pauInitTableCommon[i].value);
	}

#if DBG
    for (i = 0; i < mcrListSize; i++) {
        NIC_MCR_READ(adapter_p, pauInitTableCommon[i].offset, &temp32);

        DBGLOG(INIT, INFO,
            ("   MCR read-back: offset=0x%04x, read_value=0x%08x\n",
            pauInitTableCommon[i].offset, temp32));
    }
#endif

    if (adapter_p->rfSpecificConfig_p == NULL) {
        ERRORLOG(("Unsupported RF Type\n"));
        return;
    }

#if SUPPORT_802_11A
    if (adapter_p->currentPhyType == PHY_TYPE_OFDM) {
        mcrList_p = adapter_p->rfSpecificConfig_p->pauInitTable5_p;
        mcrListSize = adapter_p->rfSpecificConfig_p->pauInitTableLen5;
    } else
#endif
    {
        mcrList_p = adapter_p->rfSpecificConfig_p->pauInitTable24_p;
        mcrListSize = adapter_p->rfSpecificConfig_p->pauInitTableLen24;
    }

#if DBG
    NIC_MCR_READ(adapter_p, MCR_SCR, &temp32);
    DBGLOG(INIT, INFO, ("SCR: 0x%08x\n", temp32));
#endif

    if (mcrList_p) {
        INITLOG(("Initialize PAU with RF-specific PAU init table\n"));

        for (i = 0; i < mcrListSize; i++) {
            DBGLOG(INIT, INFO, ("MCR write: offset=0x%04x, value=0x%08x\n",
                mcrList_p[i].offset, mcrList_p[i].value));
            NIC_MCR_WRITE(adapter_p, mcrList_p[i].offset, mcrList_p[i].value);
	    }

#if DBG
        for (i = 0; i < mcrListSize; i++) {
            NIC_MCR_READ(adapter_p, mcrList_p[i].offset, &temp32);
            DBGLOG(INIT, INFO,
                ("   MCR read-back: offset=0x%04x, read_value=0x%08x\n",
                mcrList_p[i].offset, temp32));
        }
#endif
    }


#if DBG
    NIC_MCR_WRITE(adapter_p, MCR_PRBCR, adapter_p->MACDebugPort);
#endif
}   /* pauMCRInit */

/*_________________________________________________________________________
**  pauSetOwnMacAddress
**
**  descriptions: This routine will write a given individual address to the
**                Own MAC Address Registers (OMARs) of the INPROCOMM
**                IPN2220 chip.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      ownMacAddr_p - Pointer to the own MAC address to be written to the
**                     OMRs
**  return:       (none)
**  note:
**__________________________________________________________________________*/
VOID
pauSetOwnMacAddress(IN PMP_ADAPTER adapter_p,
                    IN PUINT_8 ownMacAddr_p)
{
    UINT_32 regValue;

    DEBUGFUNC("pauSetOwnMacAddress");

    regValue = ((UINT_32) ownMacAddr_p[0]) | ((UINT_32) ownMacAddr_p[1] << 8) |
        ((UINT_32) ownMacAddr_p[2] << 16) | ((UINT_32) ownMacAddr_p[3] << 24);

    NIC_MCR_WRITE(adapter_p, MCR_OMAR0, regValue);

#if DBG
    NIC_MCR_READ(adapter_p, MCR_OMAR0, &regValue);
    DBGLOG(HW, INFO, ("OMAR0 = 0x%08x\n", regValue));
#endif

    regValue = ((UINT_32) ownMacAddr_p[4]) | ((UINT_32) ownMacAddr_p[5] << 8);

    NIC_MCR_WRITE(adapter_p, MCR_OMAR1, regValue);

#if DBG
    NIC_MCR_READ(adapter_p, MCR_OMAR1, &regValue);
    DBGLOG(HW, INFO, ("OMAR1 = 0x%08x\n", regValue));
#endif
}   /* pauSetOwnMacAddress */

/*_________________________________________________________________________
**  pauSetCurrentBssid
**
**  descriptions: This routine will write a given BSSID to the Current
**                BSSID Registers (CBRs) of the INPROCOMM IPN2220 chip.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      currentBssid_p - Pointer to the current BSSID to be written to
**                       the CBRs
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
pauSetCurrentBssid(IN PMP_ADAPTER adapter_p,
                   IN PUINT_8 currentBssid_p)
{
    UINT_32 regValue;

    DEBUGFUNC("pauSetCurrentBssid");

    regValue = ((UINT_32) currentBssid_p[0]) |
        ((UINT_32) currentBssid_p[1] << 8) |
        ((UINT_32) currentBssid_p[2] << 16) |
        ((UINT_32) currentBssid_p[3] << 24);

    NIC_MCR_WRITE(adapter_p, MCR_CBR0, regValue);

    regValue = ((UINT_32) currentBssid_p[4]) |
        ((UINT_32) currentBssid_p[5] << 8);

    NIC_MCR_WRITE(adapter_p, MCR_CBR1, regValue);

#if DBG
    NIC_MCR_READ(adapter_p, MCR_CBR0, &regValue);
    DBGLOG(HW, INFO, ("CBR0 = 0x%08x\n", regValue));

    NIC_MCR_READ(adapter_p, MCR_CBR1, &regValue);
    DBGLOG(HW, INFO, ("CBR1 = 0x%08x\n", regValue));
#endif
}   /* pauSetCurrentBssid */

/*_________________________________________________________________________
**  pauEnablePromiscuousMode
**
**  descriptions: This routine is called to enable the promiscuous mode of
**                the NIC.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
pauEnablePromiscuousMode(IN PMP_ADAPTER adapter_p)
{
    UINT_32 regValue;

    DEBUGFUNC("pauEnablePromiscuousMode");

    DBGLOG(HW, TRACE, ("Enable NIC promiscuous mode\n"));

    NIC_MCR_READ(adapter_p, MCR_DRCR, &regValue);

    regValue &= ~DRCR_ADDR_1_VALIDATION;

    NIC_MCR_WRITE(adapter_p, MCR_DRCR, regValue);
}   /* pauEnablePromiscuousMode */

/*_________________________________________________________________________
**  pauDisablePromiscuousMode
**
**  descriptions: This routine is called to disable the promiscuous mode of
**                the NIC.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
pauDisablePromiscuousMode(IN PMP_ADAPTER adapter_p)
{
    UINT_32 regValue;

    DEBUGFUNC("pauDisablePromiscuousMode");

    DBGLOG(HW, TRACE, ("Disable NIC promiscuous mode\n"));

    NIC_MCR_READ(adapter_p, MCR_DRCR, &regValue);

    regValue |= DRCR_ADDR_1_VALIDATION;

    NIC_MCR_WRITE(adapter_p, MCR_DRCR, regValue);
}   /* pauDisablePromiscuousMode */

/*_________________________________________________________________________
**  pauEnableSnifferMode
**
**  descriptions: This routine is called to enable the sniffer (wireless
**                traffic monitoring) mode of the NIC. In this mode, all
**                of the frames, including MAC control frames, are passed
**                to the software.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
pauEnableSnifferMode(IN PMP_ADAPTER adapter_p)
{
    UINT_32 regValue;

    DEBUGFUNC("pauEnableSnifferMode");

    DBGLOG(HW, TRACE, ("Enable NIC sniffer mode\n"));

    NIC_MCR_READ(adapter_p, MCR_DRCR, &regValue);

    regValue |= DRCR_PROMISCUOUS_MODE;

    regValue &= ~DRCR_FCS_ERROR_FRAME_DISCARD;

    NIC_MCR_WRITE(adapter_p, MCR_DRCR, regValue);
}   /* pauEnableSnifferMode */

/*_________________________________________________________________________
**  pauDisableSnifferMode
**
**  descriptions: This routine is called to disable the sniffer (wireless
**                traffic monitoring) mode of the NIC.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
pauDisableSnifferMode(IN PMP_ADAPTER adapter_p)
{
    UINT_32 regValue;

    DEBUGFUNC("pauDisableSnifferMode");

    DBGLOG(HW, TRACE, ("Disable NIC sniffer mode\n"));

    NIC_MCR_READ(adapter_p, MCR_DRCR, &regValue);

    regValue &= ~DRCR_PROMISCUOUS_MODE;

    regValue |= DRCR_FCS_ERROR_FRAME_DISCARD;

    NIC_MCR_WRITE(adapter_p, MCR_DRCR, regValue);
}   /* pauDisableSnifferMode */

/*_________________________________________________________________________
**  pauSetupReceiverAddrMatching
**
**  descriptions: This routine is called to set up the receiver address
**                matching in the PAU.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
pauSetupReceiverAddrMatching(IN PMP_ADAPTER adapter_p,
                             IN UINT_32 addrMatchingMask)
{
    UINT_32 regValue;

    DEBUGFUNC("pauSetupReceiverAddrMatching");

    DBGLOG(HW, TRACE, ("Disable NIC sniffer mode\n"));

    NIC_MCR_READ(adapter_p, MCR_DRCR, &regValue);

    regValue &= ~(DRCR_MGT_BSSID_VALIDATION | DRCR_ADDR_3_OWN_SA_FILTERING |
        DRCR_ADDR_3_BSSID_VALIDATION | DRCR_ADDR_2_BSSID_VALIDATION);

    regValue |= addrMatchingMask & (DRCR_MGT_BSSID_VALIDATION |
        DRCR_ADDR_3_OWN_SA_FILTERING | DRCR_ADDR_3_BSSID_VALIDATION |
        DRCR_ADDR_2_BSSID_VALIDATION);

    NIC_MCR_WRITE(adapter_p, MCR_DRCR, regValue);
}   /* pauSetupReceiverAddrMatching */

/*_________________________________________________________________________
**  pauSetAllowedMcAddr
**
**  descriptions: This routine will write an multicast address to the
**                Allowed Multicast Address Registers (AMARs).
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      mcAddrIndex - The index of the allowed multicast address to which
**                  the given multicast address will be written
**      mcAddr_p - Pointer to the multicast address to be written
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
pauSetAllowedMcAddr(IN PMP_ADAPTER adapter_p,
                    IN UINT_32 mcAddrIndex,
                    IN PUINT_8 mcAddr_p)
{
    UINT_32 regOffset;
    UINT_32 regValue;

    DEBUGFUNC("pauSetAllowedMcAddr");

    if (mcAddrIndex >= 6) {
        return;
    }

    regOffset = (mcAddrIndex * 6) & 0xfffffffc;

    if (mcAddrIndex & 1) {
        NIC_MCR_READ(adapter_p, MCR_AMAR0 + regOffset, &regValue);

        regValue &= BITS(0,15);

        regValue |= ((UINT_32) mcAddr_p[0] << 16) | ((UINT_32) mcAddr_p[1] << 24);

        NIC_MCR_WRITE(adapter_p, MCR_AMAR0 + regOffset, regValue);

        DBGLOG(HW, LOUD, ("MC Addr %d: Write AMAR0 + %d: 0x%08x\n",
            mcAddrIndex, regOffset, regValue));

        regOffset += 4;

        regValue = (UINT_32) mcAddr_p[2] | ((UINT_32) mcAddr_p[3] << 8) |
            ((UINT_32) mcAddr_p[4] << 16) | ((UINT_32) mcAddr_p[5] << 24);

        NIC_MCR_WRITE(adapter_p, MCR_AMAR0 + regOffset, regValue);

        DBGLOG(HW, LOUD, ("MC Addr %d: Write AMAR0 + %d: 0x%08x\n",
            mcAddrIndex, regOffset, regValue));
    } else {
        regValue = (UINT_32) mcAddr_p[0] | ((UINT_32) mcAddr_p[1] << 8) |
            ((UINT_32) mcAddr_p[2] << 16) | ((UINT_32) mcAddr_p[3] << 24);

        NIC_MCR_WRITE(adapter_p, MCR_AMAR0 + regOffset, regValue);

        DBGLOG(HW, LOUD, ("MC Addr %d: Write AMAR0 + %d: 0x%08x\n",
            mcAddrIndex, regOffset, regValue));

        regOffset += 4;

        NIC_MCR_READ(adapter_p, MCR_AMAR0 + regOffset, &regValue);

        regValue &= BITS(16,31);

        regValue |= (UINT_32) mcAddr_p[4] | ((UINT_32) mcAddr_p[5] << 8);

        NIC_MCR_WRITE(adapter_p, MCR_AMAR0 + regOffset, regValue);

        DBGLOG(HW, LOUD, ("MC Addr %d: Write AMAR0 + %d: 0x%08x\n",
            mcAddrIndex, regOffset, regValue));
    }
}   /* pauSetAllowedMcAddr */

/*_________________________________________________________________________
**  pauSetMacOperationMode
**
**  descriptions: This routine will write the current MAC operation mode
**                to the MAC Protocol Timing Control Register (MPTCR).
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      macOperationMode - The current MAC operation mode
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
pauSetMacOperationMode(IN PMP_ADAPTER adapter_p,
                       IN UINT_32 macOperationMode)
{
    UINT_32 regValue;

    DEBUGFUNC("pauSetMacOperationMode");

    NIC_MCR_READ(adapter_p, MCR_MPTCR, &regValue);

    DBGLOG(HW, INFO, ("MPTCR = 0x%08x (original)\n", regValue));

    regValue &= ~MPTCR_OPERATION_MODE;

    switch (macOperationMode) {
        case MAC_OPERATION_MODE_ESS_STA:
             regValue |= MPTCR_OPERATION_MODE_ESS_STA;
             break;

        case MAC_OPERATION_MODE_IBSS:
             regValue |= MPTCR_OPERATION_MODE_IBSS;
             break;

        case MAC_OPERATION_MODE_ESS_AP:
             regValue |= MPTCR_OPERATION_MODE_ESS_AP;
             break;
    }

    NIC_MCR_WRITE(adapter_p, MCR_MPTCR, regValue);

#if DBG
    NIC_MCR_READ(adapter_p, MCR_MPTCR, &regValue);
    DBGLOG(HW, INFO, ("MPTCR = 0x%08x\n", regValue));
#endif
}   /* pauSetMacOperationMode */

/*_________________________________________________________________________
**  pauGetCtrlRspTxRateCode
**
**  descriptions: This routine is called to get the code of the rate that
**                will be used to set the Control Response Frame Control
**                Registers (CRFCRs).
**  parameters:
**      rate - The rate to transmit control response frames
**  result:
**      rateCode_p - Pointer to the code of the control response
**                   transmission rate that can be used to set the CRFCRs.
**  return:
**      TRUE - We successfully got the code of the given rate.
**      FALSE - The given rate is not recognized by the routine.
**  note:
**__________________________________________________________________________*/
BOOLEAN
pauGetCtrlRspTxRateCode(IN UINT_8 rate,
                        OUT PUINT_32 rateCode_p)
{
    PDATA_RATE_INFO rateInfo_p;

    if (!rateGetDataRateInfo(rate, &rateInfo_p)) {
        return FALSE;
    }

    *rateCode_p = rateInfo_p->rateCode;

    if (rateInfo_p->phyMode & PHY_MODULATION_OFDM) {
        *rateCode_p |= CRFCR_CTRL_RSP_TX_RATE_OFDM_MODE;
    } else if (rateInfo_p->phyMode & PHY_MODULATION_DSSS_CCK) {
        *rateCode_p |= CRFCR_CTRL_RSP_TX_RATE_CCK_MODE;
    }

    return TRUE;
}   /* pauGetCtrlRspTxRateCode */

/*_________________________________________________________________________
**  pauSetCtrlRspTxRate
**
**  descriptions: This routine is called to set the Control Response Frame
**                Control Registers (CRFCRs) by a given basic rate set.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      macOperationMode - The current MAC operation mode
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
pauSetCtrlRspTxRate(IN PMP_ADAPTER adapter_p,
                    IN UINT_32 basicRateSet)
{
    UINT_8 ctrlRspTxRate;
    UINT_32 rateCode;
    UINT_32 regValue;

    DEBUGFUNC("pauSetCtrlRspTxRate");

    NIC_MCR_READ(adapter_p, MCR_CRFCR0, &regValue);

    regValue &= CRFCR0_RSP_SAME_PREAMBLE;

    regValue |= CRFCR_CTRL_RSP_TX_RATE_1M & CRFCR0_CTRL_RSP_TX_RATE_FOR_1M;

    if (rateGetHighestRate(basicRateSet & RATE_SET_MASK_ERP_DSSS_CCK,
        RATE_2M, &ctrlRspTxRate, NULL)) {
        if (!pauGetCtrlRspTxRateCode(ctrlRspTxRate, &rateCode)) {
            rateCode = CRFCR_CTRL_RSP_TX_RATE_1M;
        }
    } else {
        rateCode = CRFCR_CTRL_RSP_TX_RATE_2M;
    }
    regValue |= (rateCode << 5) & CRFCR0_CTRL_RSP_TX_RATE_FOR_2M;

    if (rateGetHighestRate(basicRateSet & RATE_SET_MASK_ERP_DSSS_CCK,
        RATE_5_5M, &ctrlRspTxRate, NULL)) {
        if (!pauGetCtrlRspTxRateCode(ctrlRspTxRate, &rateCode)) {
            rateCode = CRFCR_CTRL_RSP_TX_RATE_1M;
        }
    } else {
        rateCode = CRFCR_CTRL_RSP_TX_RATE_5_5M;
    }
    regValue |= (rateCode << 10) & CRFCR0_CTRL_RSP_TX_RATE_FOR_5_5M;

    if (rateGetHighestRate(basicRateSet & RATE_SET_MASK_ERP_DSSS_CCK,
        RATE_11M, &ctrlRspTxRate, NULL)) {
        if (!pauGetCtrlRspTxRateCode(ctrlRspTxRate, &rateCode)) {
            rateCode = CRFCR_CTRL_RSP_TX_RATE_1M;
        }
    } else {
        rateCode = CRFCR_CTRL_RSP_TX_RATE_11M;
    }
    regValue |= (rateCode << 15) & CRFCR0_CTRL_RSP_TX_RATE_FOR_11M;

    if (rateGetHighestRate(basicRateSet & RATE_SET_MASK_ERP_OFDM, RATE_6M,
        &ctrlRspTxRate, NULL)) {
        if (!pauGetCtrlRspTxRateCode(ctrlRspTxRate, &rateCode)) {
            rateCode = CRFCR_CTRL_RSP_TX_RATE_6M;
        }
    } else {
        rateCode = CRFCR_CTRL_RSP_TX_RATE_6M;
    }
    regValue |= (rateCode << 20) & CRFCR0_CTRL_RSP_TX_RATE_FOR_6M;

    if (rateGetHighestRate(basicRateSet & RATE_SET_MASK_ERP_OFDM, RATE_9M,
        &ctrlRspTxRate, NULL)) {
        if (!pauGetCtrlRspTxRateCode(ctrlRspTxRate, &rateCode)) {
            rateCode = CRFCR_CTRL_RSP_TX_RATE_6M;
        }
    } else {
        rateCode = CRFCR_CTRL_RSP_TX_RATE_6M;
    }
    regValue |= (rateCode << 25) & CRFCR0_CTRL_RSP_TX_RATE_FOR_9M;

    NIC_MCR_WRITE(adapter_p, MCR_CRFCR0, regValue);

#if DBG
    NIC_MCR_READ(adapter_p, MCR_CRFCR0, &regValue);
    DBGLOG(HW, LOUD, ("CRFCR0 = 0x%08x\n", regValue));
#endif

    if (rateGetHighestRate(basicRateSet & RATE_SET_MASK_ERP_OFDM, RATE_12M,
        &ctrlRspTxRate, NULL)) {
        if (!pauGetCtrlRspTxRateCode(ctrlRspTxRate, &rateCode)) {
            rateCode = CRFCR_CTRL_RSP_TX_RATE_6M;
        }
    } else {
        rateCode = CRFCR_CTRL_RSP_TX_RATE_12M;
    }
    regValue = rateCode & CRFCR1_CTRL_RSP_TX_RATE_FOR_12M;

    if (rateGetHighestRate(basicRateSet & RATE_SET_MASK_ERP_OFDM, RATE_18M,
        &ctrlRspTxRate, NULL)) {
        if (!pauGetCtrlRspTxRateCode(ctrlRspTxRate, &rateCode)) {
            rateCode = CRFCR_CTRL_RSP_TX_RATE_6M;
        }
    } else {
        rateCode = CRFCR_CTRL_RSP_TX_RATE_12M;
    }
    regValue |= (rateCode << 5) & CRFCR1_CTRL_RSP_TX_RATE_FOR_18M;

    if (rateGetHighestRate(basicRateSet & RATE_SET_MASK_ERP_OFDM, RATE_24M,
        &ctrlRspTxRate, NULL)) {
        if (!pauGetCtrlRspTxRateCode(ctrlRspTxRate, &rateCode)) {
            rateCode = CRFCR_CTRL_RSP_TX_RATE_6M;
        }
    } else {
        rateCode = CRFCR_CTRL_RSP_TX_RATE_24M;
    }
    regValue |= (rateCode << 10) & CRFCR1_CTRL_RSP_TX_RATE_FOR_24M;

    if (rateGetHighestRate(basicRateSet & RATE_SET_MASK_ERP_OFDM, RATE_36M,
        &ctrlRspTxRate, NULL)) {
        if (!pauGetCtrlRspTxRateCode(ctrlRspTxRate, &rateCode)) {
            rateCode = CRFCR_CTRL_RSP_TX_RATE_6M;
        }
    } else {
        rateCode = CRFCR_CTRL_RSP_TX_RATE_24M;
    }
    regValue |= (rateCode << 15) & CRFCR1_CTRL_RSP_TX_RATE_FOR_36M;

    if (rateGetHighestRate(basicRateSet & RATE_SET_MASK_ERP_OFDM, RATE_48M,
        &ctrlRspTxRate, NULL)) {
        if (!pauGetCtrlRspTxRateCode(ctrlRspTxRate, &rateCode)) {
            rateCode = CRFCR_CTRL_RSP_TX_RATE_6M;
        }
    } else {
        rateCode = CRFCR_CTRL_RSP_TX_RATE_24M;
    }
    regValue |= (rateCode << 20) & CRFCR1_CTRL_RSP_TX_RATE_FOR_48M;

    if (rateGetHighestRate(basicRateSet & RATE_SET_MASK_ERP_OFDM, RATE_54M,
        &ctrlRspTxRate, NULL)) {
        if (!pauGetCtrlRspTxRateCode(ctrlRspTxRate, &rateCode)) {
            rateCode = CRFCR_CTRL_RSP_TX_RATE_6M;
        }
    } else {
        rateCode = CRFCR_CTRL_RSP_TX_RATE_24M;
    }
    regValue |= (rateCode << 25) & CRFCR1_CTRL_RSP_TX_RATE_FOR_54M;

    NIC_MCR_WRITE(adapter_p, MCR_CRFCR1, regValue);

#if DBG
    NIC_MCR_READ(adapter_p, MCR_CRFCR1, &regValue);
    DBGLOG(HW, LOUD, ("CRFCR1 = 0x%08x\n", regValue));
#endif
}   /* pauSetCtrlRspTxRate */

/*_________________________________________________________________________
**  pauSetCtrlRspTxRateNew
**
**  descriptions: This routine is called to set the Control Response Frame
**                Control Registers (CRFCRs) by an array of supported rates
**                information.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      supRatesArray_p - Pointer to the array of the BSS supported rate
**                        information structures. If a value of NULL is
**                        given, the default values of CRFCRs are set.
**      numSupRates - The number of BSS supported rates. If a value of 0 is
**                    given, the default values of CRFCRs are set.
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
pauSetCtrlRspTxRateNew(IN PMP_ADAPTER adapter_p,
                       IN PSUPPORTED_RATE_INFO supRatesArray_p,
                       IN UINT numSupRates)
{
    UINT i;
    UINT_32 rateCode;
    UINT_32 regValue0, regValue1;
    PDATA_RATE_INFO rateInfo_p;

    DEBUGFUNC("pauSetCtrlRspTxRateNew");

    NIC_MCR_READ(adapter_p, MCR_CRFCR0, &regValue0);

    regValue0 = (regValue0 & (CRFCR0_RSP_SAME_PREAMBLE | CRFCR0_RSP_PWR_MGT)) |
        (CRFCR0_DEFAULT_VALUE & ~(CRFCR0_RSP_SAME_PREAMBLE |
        CRFCR0_RSP_PWR_MGT));

    regValue1 = CRFCR1_DEFAULT_VALUE;

    if (supRatesArray_p) {
        for (i = 0; i < numSupRates; i++) {
            if (!rateGetDataRateInfo(supRatesArray_p[i].ackTxRate,
                &rateInfo_p)) {
                continue;
            }

            rateCode = rateInfo_p->rateCode;

            if (rateInfo_p->phyMode & PHY_MODULATION_OFDM) {
                rateCode |= CRFCR_CTRL_RSP_TX_RATE_OFDM_MODE;
            } else if (rateInfo_p->phyMode & PHY_MODULATION_DSSS_CCK) {
                rateCode |= CRFCR_CTRL_RSP_TX_RATE_CCK_MODE;
            }

            switch (supRatesArray_p[i].rate) {
                case RATE_1M:
                     regValue0 = (regValue0 & ~CRFCR0_CTRL_RSP_TX_RATE_FOR_1M) |
                         (rateCode & CRFCR0_CTRL_RSP_TX_RATE_FOR_1M);
                     break;

                case RATE_2M:
                     regValue0 = (regValue0 & ~CRFCR0_CTRL_RSP_TX_RATE_FOR_2M) |
                         ((rateCode << 5) & CRFCR0_CTRL_RSP_TX_RATE_FOR_2M);
                     break;

                case RATE_5_5M:
                     regValue0 = (regValue0 & ~CRFCR0_CTRL_RSP_TX_RATE_FOR_5_5M) |
                         ((rateCode << 10) & CRFCR0_CTRL_RSP_TX_RATE_FOR_5_5M);
                     break;

                case RATE_11M:
                     regValue0 = (regValue0 & ~CRFCR0_CTRL_RSP_TX_RATE_FOR_11M) |
                         ((rateCode << 15) & CRFCR0_CTRL_RSP_TX_RATE_FOR_11M);
                     break;

                case RATE_6M:
                     regValue0 = (regValue0 & ~CRFCR0_CTRL_RSP_TX_RATE_FOR_6M) |
                         ((rateCode << 20) & CRFCR0_CTRL_RSP_TX_RATE_FOR_6M);
                     break;

                case RATE_9M:
                     regValue0 = (regValue0 & ~CRFCR0_CTRL_RSP_TX_RATE_FOR_9M) |
                         ((rateCode << 25) & CRFCR0_CTRL_RSP_TX_RATE_FOR_9M);
                     break;

                case RATE_12M:
                     regValue1 = (regValue1 & ~CRFCR1_CTRL_RSP_TX_RATE_FOR_12M) |
                         (rateCode & CRFCR1_CTRL_RSP_TX_RATE_FOR_12M);
                     break;

                case RATE_18M:
                     regValue1 = (regValue1 & ~CRFCR1_CTRL_RSP_TX_RATE_FOR_18M) |
                         ((rateCode << 5) & CRFCR1_CTRL_RSP_TX_RATE_FOR_18M);
                     break;

                case RATE_24M:
                     regValue1 = (regValue1 & ~CRFCR1_CTRL_RSP_TX_RATE_FOR_24M) |
                         ((rateCode << 10) & CRFCR1_CTRL_RSP_TX_RATE_FOR_24M);
                     break;

                case RATE_36M:
                     regValue1 = (regValue1 & ~CRFCR1_CTRL_RSP_TX_RATE_FOR_36M) |
                         ((rateCode << 15) & CRFCR1_CTRL_RSP_TX_RATE_FOR_36M);
                     break;

                case RATE_48M:
                     regValue1 = (regValue1 & ~CRFCR1_CTRL_RSP_TX_RATE_FOR_48M) |
                         ((rateCode << 20) & CRFCR1_CTRL_RSP_TX_RATE_FOR_48M);
                     break;

                case RATE_54M:
                     regValue1 = (regValue1 & ~CRFCR1_CTRL_RSP_TX_RATE_FOR_54M) |
                         ((rateCode << 25) & CRFCR1_CTRL_RSP_TX_RATE_FOR_54M);
                     break;
            }
        }
    }

    NIC_MCR_WRITE(adapter_p, MCR_CRFCR0, regValue0);
    NIC_MCR_WRITE(adapter_p, MCR_CRFCR1, regValue1);

#if DBG
    NIC_MCR_READ(adapter_p, MCR_CRFCR0, &regValue0);
    DBGLOG(HW, LOUD, ("CRFCR0 = 0x%08x\n", regValue0));

    NIC_MCR_READ(adapter_p, MCR_CRFCR1, &regValue1);
    DBGLOG(HW, LOUD, ("CRFCR1 = 0x%08x\n", regValue1));
#endif
}   /* pauSetCtrlRspTxRateNew */

/*_________________________________________________________________________
**  pauSetBeaconContent
**
**  descriptions: This routine is called to set the beacon content to the
**                hardware through the SW Beacon Content Control and Data
**                Registers.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      offset - The offset at which the beacon content locates from the
**               Beacon Interval field in the generated Beacon frame
**      content_p - Pointer to the starting address of the beacon content
**                  that we want to write to the PAU
**      contentLen - The length of the beacon content that we want to write
**                   to the PAU
**  return:
**      TRUE - We have successfully setup the beacon content.
**      FALSE - We cannot setup the beacon content since the PAU were
**              always accessing the BGRs for more than 10 microseconds.
**  note:
**__________________________________________________________________________*/
BOOLEAN
pauSetBeaconContent(IN PMP_ADAPTER adapter_p,
                    IN UINT_32 offset,
                    IN PUINT_8 content_p,
                    IN UINT_32 contentLen)
{
    UINT_32 i;
    BOOLEAN succeeded;
    UINT_32 index;
    UINT_32 bcData;
    UINT_32 regValue;

    DEBUGFUNC("pauSetBeaconContent");

    DBGLOG(HW, LOUD, ("\n"));

    NIC_MCR_READ(adapter_p, MCR_BGCR1, &regValue);

    if (regValue & BGCR1_BC_OWN) {
        regValue &= ~BGCR1_BC_OWN;

        NIC_MCR_WRITE_AND_WAIT(adapter_p, MCR_BGCR1, regValue,
            BGCR1_BC_OWN, 1, 10, succeeded);

        if (!succeeded) {
            WARNLOG(("Cannot own the beacon content\n"));

            return FALSE;
        }
    }

    NIC_MCR_READ_AND_WAIT(adapter_p, MCR_SWBCCR, &regValue,
        (regValue & SWBCCR_BC_ACCESS_START) == 0, 2, 10, succeeded);

    if (!succeeded) {
        WARNLOG(("Cannot read BC due to busy BC access\n"));

        return FALSE;
    }

    regValue = SWBCCR_BC_ACCESS_START | SWBCCR_BC_READ | ((offset >> 2) &
        SWBCCR_BC_INDEX);

    NIC_MCR_WRITE_AND_WAIT(adapter_p, MCR_SWBCCR, regValue,
        SWBCCR_BC_ACCESS_START, 2, 10, succeeded);

    if (!succeeded) {
        WARNLOG(("Failed to read Beacon Content [%d]\n", offset >> 2));

        return FALSE;
    }

    NIC_MCR_READ(adapter_p, MCR_SWBCDR, &bcData);

    DBGLOG(HW, INFO, ("Read Beacon Content [%d]: 0x%08x\n", offset >> 2,
        bcData));

    DBGLOG(HW, INFO, ("offset = %d, contentLen = %d\n", offset, contentLen));

    for (i = 0; i < contentLen; i++) {
        index = offset + i;

        switch (index & 3) {
            case 0:
                 bcData = (bcData & (UINT_32) ~BITS(0,7)) |
                     ((UINT_32) content_p[i] & BITS(0,7));
                 break;

            case 1:
                 bcData = (bcData & (UINT_32) ~BITS(8,15)) |
                     (((UINT_32) content_p[i] << 8) & BITS(8,15));
                 break;

            case 2:
                 bcData = (bcData & (UINT_32) ~BITS(16,23)) |
                     (((UINT_32) content_p[i] << 16) & BITS(16,23));
                 break;

            default:
                 bcData = (bcData & (UINT_32) ~BITS(24,31)) |
                     (((UINT_32) content_p[i] << 24) & BITS(24,31));

                 index >>= 2;

                 DBGLOG(HW, INFO, ("Write Beacon Content [%d]: 0x%08x\n",
                     index, bcData));

                 NIC_MCR_READ_AND_WAIT(adapter_p, MCR_SWBCCR, &regValue,
                     (regValue & SWBCCR_BC_ACCESS_START) == 0, 2, 10,
                     succeeded);

                 if (!succeeded) {
                     WARNLOG(("Cannot write BC due to busy BC access\n"));

                     return FALSE;
                 }

                 NIC_MCR_WRITE(adapter_p, MCR_SWBCDR, bcData);

                 regValue = SWBCCR_BC_ACCESS_START | SWBCCR_BC_WRITE |
                     (index & SWBCCR_BC_INDEX);

                 NIC_MCR_WRITE_AND_WAIT(adapter_p, MCR_SWBCCR, regValue,
                     SWBCCR_BC_ACCESS_START, 2, 10, succeeded);

                 if (!succeeded) {
                     WARNLOG(("Failed to write Beacon Content [%d]\n", index));

                     return FALSE;
                 }

                 bcData = 0;
        }
    }

    i = offset + contentLen;

    if (i & 3) {
        index = i >> 2;

        if (index != offset >> 2) {
            NIC_MCR_READ_AND_WAIT(adapter_p, MCR_SWBCCR, &regValue,
                (regValue & SWBCCR_BC_ACCESS_START) == 0, 2, 10, succeeded);

            if (!succeeded) {
                WARNLOG(("Cannot read BC due to busy BC access\n"));

                return FALSE;
            }

            regValue = SWBCCR_BC_ACCESS_START | SWBCCR_BC_READ |
                (index & SWBCCR_BC_INDEX);

            NIC_MCR_WRITE_AND_WAIT(adapter_p, MCR_SWBCCR, regValue,
                SWBCCR_BC_ACCESS_START, 2, 10, succeeded);

            if (!succeeded) {
                WARNLOG(("Failed to read Beacon Content [%d]\n", index));

                return FALSE;
            }

            NIC_MCR_READ(adapter_p, MCR_SWBCDR, &regValue);

            DBGLOG(HW, INFO, ("Read Beacon Content [%d]: 0x%08x\n", index,
                regValue));

            switch (i & 3) {
                case 1:
                     bcData = (bcData & BITS(0,7)) |
                         (regValue & BITS(8,31));
                     break;

                case 2:
                     bcData = (bcData & BITS(0,15)) |
                         (regValue & BITS(16,31));
                     break;

                case 3:
                     bcData = (bcData & BITS(0,23)) |
                         (regValue & BITS(24,31));
                     break;
            }
        }

        DBGLOG(HW, INFO, ("Write Beacon Content [%d]: 0x%08x\n",
            index, bcData));

        NIC_MCR_READ_AND_WAIT(adapter_p, MCR_SWBCCR, &regValue,
            (regValue & SWBCCR_BC_ACCESS_START) == 0, 2, 10, succeeded);

        if (!succeeded) {
            WARNLOG(("Cannot write BC due to busy BC access\n"));

            return FALSE;
        }

        NIC_MCR_WRITE(adapter_p, MCR_SWBCDR, bcData);

        regValue = SWBCCR_BC_ACCESS_START | SWBCCR_BC_WRITE |
            (index & SWBCCR_BC_INDEX);

        NIC_MCR_WRITE_AND_WAIT(adapter_p, MCR_SWBCCR, regValue,
            SWBCCR_BC_ACCESS_START, 2, 10, succeeded);

        if (!succeeded) {
            WARNLOG(("Failed to write Beacon Content [%d]\n", index));

            return FALSE;
        }
    }

#if DBG
    for (i = offset >> 2; i <= index; i++) {
        NIC_MCR_READ_AND_WAIT(adapter_p, MCR_SWBCCR, &regValue,
            (regValue & SWBCCR_BC_ACCESS_START) == 0, 2, 10, succeeded);

        if (!succeeded) {
            WARNLOG(("Cannot read BC due to busy BC access\n"));

            return FALSE;
        }

        regValue = SWBCCR_BC_ACCESS_START | SWBCCR_BC_READ |
            (i & SWBCCR_BC_INDEX);

        NIC_MCR_WRITE_AND_WAIT(adapter_p, MCR_SWBCCR, regValue,
            SWBCCR_BC_ACCESS_START, 2, 10, succeeded);

        if (!succeeded) {
            WARNLOG(("Failed to read Beacon Content [%d]\n", i));

            return FALSE;
        }

        NIC_MCR_READ(adapter_p, MCR_SWBCDR, &bcData);

        DBGLOG(HW, INFO, ("Beacon Content [%d]: 0x%08x\n", i, bcData));
    }
#endif

    return TRUE;
}   /* pauSetBeaconContent */

#if SUPPORT_802_11G
/*_________________________________________________________________________
**  pauChangeSlotTime
**
**  descriptions: This routine is called to change the slot time setting.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      slotTime - The new slot time to be set
**  return:
**      TRUE - The slot time is changed successfully.
**      FALSE - The given slot time value is not supported.
**  note:
**__________________________________________________________________________*/
BOOLEAN
pauChangeSlotTime(IN PMP_ADAPTER adapter_p,
                  IN UINT_8 slotTime,
                  IN UINT_32 useED)
{
    UINT_32   tntdrValue, stdrValue;
    UINT_8    ccaShortSlot, ccaLongSlot;
#if DBG
    UINT_32   temp32;
#endif

    DEBUGFUNC("pauChangeSlotTime");

    if (BCIRead8(adapter_p, BBCR_CR2, &ccaShortSlot, TRUE, 0xff)) {
        ccaShortSlot &= (~CR2_CCA_METHOD);
        ccaLongSlot = ccaShortSlot;
    } else {
        ccaShortSlot = 0;
        ccaLongSlot = ccaShortSlot;
    }

    if (useED) {
        ccaShortSlot |= CCA_SHORT_SLOT_TIME_WITH_ED;
        ccaLongSlot |= CCA_LONG_SLOT_TIME_WITH_ED;
    } else {
        ccaShortSlot |= CCA_SHORT_SLOT_TIME_NO_ED;
        ccaLongSlot |= CCA_LONG_SLOT_TIME_NO_ED;
    }

    if (adapter_p->rfSpecificConfig_p == NULL) {
        ERRORLOG(("Unsupported RF Type\n"));
        return FALSE;
    }

    tntdrValue = adapter_p->rfSpecificConfig_p->pauTntdrValue;
    stdrValue = adapter_p->rfSpecificConfig_p->pauStdrValue;

    switch (slotTime) {
        case SLOT_TIME_SHORT:
             DBGLOG(HW, TRACE, ("Change to short slot time\n"));
             tntdrValue -= 0xdc0000;
             stdrValue -= 0xdc00dc;
             if (!BCIWrite8(adapter_p, BBCR_CR2, ccaShortSlot, TRUE, 0xff)) {
                 return FALSE;
             }
             break;

        case SLOT_TIME_ERP_LONG:
             DBGLOG(HW, TRACE, ("Change to long slot time\n"));
             if (!BCIWrite8(adapter_p, BBCR_CR2, ccaLongSlot, TRUE, 0xff)) {
                 return FALSE;
             }
             break;

        default:
             ERRORLOG(("Invalid slot time: %d us\n", slotTime));
             return FALSE;
    }

    NIC_MCR_WRITE(adapter_p, MCR_TNTDR, tntdrValue);
    NIC_MCR_WRITE(adapter_p, MCR_STDR, stdrValue);

#if DBG
    NIC_MCR_READ(adapter_p, MCR_TNTDR, &temp32);
    DBGLOG(HW, TRACE, ("SNTDR: 0x%x\n", temp32));
    NIC_MCR_READ(adapter_p, MCR_STDR, &temp32);
    DBGLOG(HW, TRACE, ("STDR (slot time): 0x%x\n", temp32));
#endif

    return TRUE;
}   /* pauChangeSlotTime */
#endif

/*_________________________________________________________________________
**  pauSetACParameters
**
**  descriptions: This routine is called to set access category (AC)
**                parameters, CWmin, CWmax, AIFSN, and TXOP limit, if
**                exists, for AC0-AC3 and TQ0/TQ1 queues.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      cwmin - The CWmin value of the given transmit queue
**      cwmax - The CWmax value of the given transmit queue
**      aifsn - The AIFSN value of the given transmit queue
**      txopLimit - The TXOP limit of the given transmit queue, in units of
**                  32 us. This parameter has no effect on TQ0/TQ1.
**  return:
**      TRUE - The AC parameters of the given transmit queue are set
**             successfully.
**      FALSE - The given transmit queue has no settings about the AC
**              parameters.
**  note:
**__________________________________________________________________________*/
BOOLEAN
pauSetACParameters(IN PMP_ADAPTER adapter_p,
                   IN TX_QUEUE_NUM txQueueNum,
                   IN UINT_32 cwmin,
                   IN UINT_32 cwmax,
                   IN UINT_32 aifsn,
                   IN UINT_32 txopLimit)
{
    UINT_32 tmp32;

    DEBUGFUNC("pauSetACParameters");

    if (cwmin > 0xFF) {
        cwmin = 0xFF;
    }

    if (cwmax > 0xFFFF) {
        cwmax = 0xFFFF;
    }

    NIC_MCR_READ(adapter_p, MCR_AIFSR, &tmp32);

    switch (txQueueNum) {
        case TXQ_AC0:
             DBGLOG(HW, TRACE,
                 ("AC0: CWmin=%d, CWmax=%d, AIFSN=%d, TXOP_Limit=%d\n", cwmin,
                 cwmax, aifsn, txopLimit));

             NIC_MCR_WRITE(adapter_p, MCR_AC0CWR,
                 (cwmin & AC0CWR_AC0_CW_MIN) | (cwmax << 16));

             tmp32 = (tmp32 & ~AIFSR_AIFS_0) | (aifsn & AIFSR_AIFS_0);
             if (aifsn < 2) {
                 tmp32 |= AIFSR_AC0_RN_ADD_1_EN;
             } else {
                 tmp32 &= ~AIFSR_AC0_RN_ADD_1_EN;
             }
             NIC_MCR_WRITE(adapter_p, MCR_AIFSR, tmp32);

             NIC_MCR_READ(adapter_p, MCR_ACTXOPR0, &tmp32);
             tmp32 = (tmp32 & ACTXOPR0_AC1_LIMIT) |
                 (txopLimit & ACTXOPR0_AC0_LIMIT);
             NIC_MCR_WRITE(adapter_p, MCR_ACTXOPR0, tmp32);
             break;

        case TXQ_AC1:
             DBGLOG(HW, TRACE,
                 ("AC1: CWmin=%d, CWmax=%d, AIFSN=%d, TXOP_Limit=%d\n", cwmin,
                 cwmax, aifsn, txopLimit));

             NIC_MCR_WRITE(adapter_p, MCR_AC1CWR,
                 (cwmin & AC1CWR_AC1_CW_MIN) | (cwmax << 16));

             tmp32 = (tmp32 & ~AIFSR_AIFS_1) | ((aifsn << 4) & AIFSR_AIFS_1);
             if (aifsn < 2) {
                 tmp32 |= AIFSR_AC1_RN_ADD_1_EN;
             } else {
                 tmp32 &= ~AIFSR_AC1_RN_ADD_1_EN;
             }
             NIC_MCR_WRITE(adapter_p, MCR_AIFSR, tmp32);

             NIC_MCR_READ(adapter_p, MCR_ACTXOPR0, &tmp32);
             tmp32 = (tmp32 & ACTXOPR0_AC0_LIMIT) | (txopLimit << 16);
             NIC_MCR_WRITE(adapter_p, MCR_ACTXOPR0, tmp32);
             break;

        case TXQ_AC2:
             DBGLOG(HW, TRACE,
                 ("AC2: CWmin=%d, CWmax=%d, AIFSN=%d, TXOP_Limit=%d\n", cwmin,
                 cwmax, aifsn, txopLimit));

             NIC_MCR_WRITE(adapter_p, MCR_AC2CWR,
                 (cwmin & AC2CWR_AC2_CW_MIN) | (cwmax << 16));

             tmp32 = (tmp32 & ~AIFSR_AIFS_2) | ((aifsn << 8) & AIFSR_AIFS_2);
             if (aifsn < 2) {
                 tmp32 |= AIFSR_AC2_RN_ADD_1_EN;
             } else {
                 tmp32 &= ~AIFSR_AC2_RN_ADD_1_EN;
             }
             NIC_MCR_WRITE(adapter_p, MCR_AIFSR, tmp32);

             NIC_MCR_READ(adapter_p, MCR_ACTXOPR1, &tmp32);
             tmp32 = (tmp32 & ACTXOPR1_AC3_LIMIT) |
                 (txopLimit & ACTXOPR1_AC2_LIMIT);
             NIC_MCR_WRITE(adapter_p, MCR_ACTXOPR1, tmp32);
             break;

        case TXQ_AC3:
             DBGLOG(HW, TRACE,
                 ("AC3: CWmin=%d, CWmax=%d, AIFSN=%d, TXOP_Limit=%d\n", cwmin,
                 cwmax, aifsn, txopLimit));

             NIC_MCR_WRITE(adapter_p, MCR_AC3CWR,
                 (cwmin & AC3CWR_AC3_CW_MIN) | (cwmax << 16));

             tmp32 = (tmp32 & ~AIFSR_AIFS_3) | ((aifsn << 12) & AIFSR_AIFS_3);
             if (aifsn < 2) {
                 tmp32 |= AIFSR_AC3_RN_ADD_1_EN;
             } else {
                 tmp32 &= ~AIFSR_AC3_RN_ADD_1_EN;
             }
             NIC_MCR_WRITE(adapter_p, MCR_AIFSR, tmp32);

             NIC_MCR_READ(adapter_p, MCR_ACTXOPR1, &tmp32);
             tmp32 = (tmp32 & ACTXOPR1_AC2_LIMIT) | (txopLimit << 16);
             NIC_MCR_WRITE(adapter_p, MCR_ACTXOPR1, tmp32);
             break;

        case TXQ_TQ0:
        case TXQ_TQ1:
             DBGLOG(HW, TRACE, ("TQ0/TQ1: CWmin=%d, CWmax=%d, AIFSN=%d\n",
                 cwmin, cwmax, aifsn));

             NIC_MCR_WRITE(adapter_p, MCR_TQCWR,
                 (cwmin & TQCWR_TQ_CW_MIN) | (cwmax << 16));

             tmp32 = (tmp32 & ~AIFSR_AIFS_TQ) | ((aifsn << 16) & AIFSR_AIFS_TQ);
             if (aifsn < 2) {
                 tmp32 |= AIFSR_TQ_RN_ADD_1_EN;
             } else {
                 tmp32 &= ~AIFSR_TQ_RN_ADD_1_EN;
             }
             NIC_MCR_WRITE(adapter_p, MCR_AIFSR, tmp32);
             break;

        default:
             WARNLOG(("No CWmin/CWmax/AIFSN/TXOP_Limit for tx queue: %d\n",
                 txQueueNum));
             return FALSE;
    }

    return TRUE;
}   /* pauSetACParameters */

/*_________________________________________________________________________
**  pauUpdateNAV
**
**  descriptions: This routine is called to update the NAV.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      nav - The NAV used to update
**  return:
**      TRUE - The slot time is changed successfully.
**      FALSE - The given slot time value is not supported.
**  note:
**__________________________________________________________________________*/
BOOLEAN
pauUpdateNAV(IN PMP_ADAPTER adapter_p,
             IN UINT_32 nav)
{
    BOOLEAN succeeded;
    UINT_32 regValue;

    NIC_MCR_READ_AND_WAIT(adapter_p, MCR_NSUR, &regValue,
        (regValue & NSUR_NAV_UPDATE) == 0, 1, 10, succeeded);

    if (!succeeded) {
        WARNLOG(("NAV update always in progress\n"));

        return FALSE;
    }

    regValue = NSUR_NAV_UPDATE | (nav & NSUR_NAV_UPDATE_VALUE);

    NIC_MCR_WRITE_AND_WAIT(adapter_p, MCR_NSUR, regValue, NSUR_NAV_UPDATE,
        2, 10, succeeded);

    if (!succeeded) {
        WARNLOG(("Busy in updating NAV\n"));
    }

    return succeeded;
}   /* pauUpdateNAV */


/*_________________________________________________________________________
**  pauClearHwKeyTable
**
**  description: This function clear the IPN2220 H/W key table
**  parameters:  adapter_p - Pointer to the adapter object data area.
**  return:
**  note:
**_________________________________________________________________________*/
VOID
pauClearHwKeyTable(IN PMP_ADAPTER adapter_p)
{
    IPN_KEY_CFG_ENTRY ipnKey;

    if (NIC_NOT_ACCESSIBLE(adapter_p)) {
        return;
    }

    NdisZeroMemory(&ipnKey, sizeof(ipnKey));

    for (ipnKey.entryIndex = 0;
         ipnKey.entryIndex < (UINT_32)adapter_p->swKeyTableSize;
         ipnKey.entryIndex++)
    {
        pauAccessHwKey(adapter_p, &ipnKey, TRUE);
    }

} /* pauClearHwKeyTable */


/*_________________________________________________________________________
**  pauAccessHwKey
**
**  description: This function write/read one key entry to/from the H/W key table
**  parameters:  adapter_p - Pointer to the adapter object data area.
**               key_p - Pointer to the IPN2220 key structure
**               writeKey - TURE: add (or clear) one key entry
**                          FALSE: read one key entry
**  return:      TRUE, successful
**               FALSE, otherwise
**  note:
**_________________________________________________________________________*/
BOOLEAN
pauAccessHwKey(IN PMP_ADAPTER adapter_p,
               IN PIPN_KEY_CFG_ENTRY key_p,
               IN BOOLEAN writeKey)
{
    UINT_32 regSkacr, regSkadr0, regSkadr1, regSkadr2, regSkadr3, regSkadr4, regSkadr5;
    UINT_32 regValue;


    DEBUGFUNC("pauAccessHwKey");

    if (NIC_NOT_ACCESSIBLE(adapter_p)) {
        WARNLOG(("NIC is not accessible now\n"));
        return FALSE;
    }

    regSkacr = regSkadr0 = regSkadr1 = regSkadr2 = regSkadr3 = regSkadr4 = regSkadr5 = 0;

    if (key_p->entryIndex > adapter_p->swKeyTableSize - 1)
    {
        ERRORLOG(("Error, invalid H/W key entry: 0x%08lx", key_p->entryIndex));
        return FALSE;
    }

    if (!adapter_p->swSecurityEngine.swDecryptionEnabled)
    {
        regSkacr = (adapter_p->swKeyTableSize << 8) & 0x0000FF00;
    }
    regSkacr |= SKACR_ACCESS_START;
    regSkacr |= key_p->entryIndex & 0x000000FF;

    if (writeKey)
    {
        if (key_p->key.keyId > 3)
        {
            ERRORLOG(("Error, invalid key ID: 0x%02x (H/W)", key_p->key.keyId));
            return FALSE;
        }

        regSkacr |= SKACR_READ_WRITE;

        if (key_p->key.chkKeyId)
        {
            regSkadr0 |= ((UINT_32)key_p->key.keyId) << 30;
            regSkadr0 |= SKADR0_CHECK_KEY_ID;
        }

        regSkadr0 |= (key_p->key.chkAddr1) ? SKADR0_CHECK_A1 : 0;
        regSkadr0 |= (key_p->key.chkAddr2) ? SKADR0_CHECK_A2 : 0;
        if (key_p->key.chkAddr1 | key_p->key.chkAddr2)
        {
            regSkadr0 |= ((UINT_32) key_p->key.addr[3])             |
                         ((UINT_32) key_p->key.addr[4] << 8)        |
                         ((UINT_32) key_p->key.addr[5] << 16);

            regSkadr1 = ((UINT_32) key_p->key.addr[0])              |
                        ((UINT_32) key_p->key.addr[1] << 8)         |
                        ((UINT_32) key_p->key.addr[2] << 16);
        }

        regSkadr0 |= ((UINT_32) key_p->key.cipher) << 24;

        regSkadr2 = ((UINT_32) key_p->key.keyMaterial[0])           |
                    ((UINT_32) key_p->key.keyMaterial[1] << 8)      |
                    ((UINT_32) key_p->key.keyMaterial[2] << 16)     |
                    ((UINT_32) key_p->key.keyMaterial[3] << 24);
        regSkadr3 = ((UINT_32) key_p->key.keyMaterial[4])           |
                    ((UINT_32) key_p->key.keyMaterial[5] << 8)      |
                    ((UINT_32) key_p->key.keyMaterial[6] << 16)     |
                    ((UINT_32) key_p->key.keyMaterial[7] << 24);
        regSkadr4 = ((UINT_32) key_p->key.keyMaterial[8])           |
                    ((UINT_32) key_p->key.keyMaterial[9] << 8)      |
                    ((UINT_32) key_p->key.keyMaterial[10] << 16)    |
                    ((UINT_32) key_p->key.keyMaterial[11] << 24);
        regSkadr5 = ((UINT_32) key_p->key.keyMaterial[12])          |
                    ((UINT_32) key_p->key.keyMaterial[13] << 8)     |
                    ((UINT_32) key_p->key.keyMaterial[14] << 16)    |
                    ((UINT_32) key_p->key.keyMaterial[15] << 24);

        if (pauCheckSKACRAccessStart(adapter_p))
        {
            NIC_MCR_WRITE(adapter_p, MCR_SKADR0, regSkadr0);
            NIC_MCR_WRITE(adapter_p, MCR_SKADR1, regSkadr1);
            NIC_MCR_WRITE(adapter_p, MCR_SKADR2, regSkadr2);
            NIC_MCR_WRITE(adapter_p, MCR_SKADR3, regSkadr3);
            NIC_MCR_WRITE(adapter_p, MCR_SKADR4, regSkadr4);
            NIC_MCR_WRITE(adapter_p, MCR_SKADR5, regSkadr5);

            NIC_MCR_WRITE(adapter_p, MCR_SKACR, regSkacr);

            DBGLOG(RSN, INFO, ("SKACR  = 0x%08lx\n", regSkacr));
            DBGLOG(RSN, INFO, ("SKADR0 = 0x%08lx\n", regSkadr0));
            DBGLOG(RSN, INFO, ("SKADR1 = 0x%08lx\n", regSkadr1));
            DBGLOG(RSN, INFO, ("SKADR2 = 0x%08lx\n", regSkadr2));
            DBGLOG(RSN, INFO, ("SKADR3 = 0x%08lx\n", regSkadr3));
            DBGLOG(RSN, INFO, ("SKADR4 = 0x%08lx\n", regSkadr4));
            DBGLOG(RSN, INFO, ("SKADR5 = 0x%08lx\n", regSkadr5));
            DBGLOG(RSN, INFO, ("Add key into H/W key table\n"));
        }
        else
        {
            ERRORLOG(("SKACR access is not available\n"));
            return FALSE;
        }
    }
    else
    {
        if (pauCheckSKACRAccessStart(adapter_p))
        {
            NIC_MCR_WRITE(adapter_p, MCR_SKACR, regSkacr);
        }

        if (pauCheckSKACRAccessStart(adapter_p))
        {
            NIC_MCR_READ(adapter_p, MCR_SKADR0, &regValue);
            key_p->key.keyId = (UINT_8)((regValue >> 30) & 0x00000003);
            key_p->key.chkAddr2 = (regValue & SKADR0_CHECK_A2) ? TRUE : FALSE;
            key_p->key.chkAddr1 = (regValue & SKADR0_CHECK_A1) ? TRUE : FALSE;
            key_p->key.chkKeyId = (regValue & SKADR0_CHECK_KEY_ID) ? TRUE : FALSE;
            key_p->key.cipher = (UINT_8)((regValue >> 24) & 0x00000007);

            key_p->key.addr[3] = (UINT_8) ((regValue) & 0x000000FF);
            key_p->key.addr[4] = (UINT_8) ((regValue >> 8) & 0x000000FF);
            key_p->key.addr[5] = (UINT_8) ((regValue >> 16) & 0x000000FF);
            NIC_MCR_READ(adapter_p, MCR_SKADR1, &regValue);
            key_p->key.addr[0] = (UINT_8) ((regValue) & 0x000000FF);
            key_p->key.addr[1] = (UINT_8) ((regValue >> 8) & 0x000000FF);
            key_p->key.addr[2] = (UINT_8) ((regValue >> 16) & 0x000000FF);

            NIC_MCR_READ(adapter_p, MCR_SKADR2, &regValue);
            key_p->key.keyMaterial[0] = (UINT_8) ((regValue) & 0x000000FF);
            key_p->key.keyMaterial[1] = (UINT_8) ((regValue >> 8) & 0x000000FF);
            key_p->key.keyMaterial[2] = (UINT_8) ((regValue >> 16) & 0x000000FF);
            key_p->key.keyMaterial[3] = (UINT_8) ((regValue >> 24) & 0x000000FF);
            NIC_MCR_READ(adapter_p, MCR_SKADR3, &regValue);
            key_p->key.keyMaterial[4] = (UINT_8) ((regValue) & 0x000000FF);
            key_p->key.keyMaterial[5] = (UINT_8) ((regValue >> 8) & 0x000000FF);
            key_p->key.keyMaterial[6] = (UINT_8) ((regValue >> 16) & 0x000000FF);
            key_p->key.keyMaterial[7] = (UINT_8) ((regValue >> 24) & 0x000000FF);
            NIC_MCR_READ(adapter_p, MCR_SKADR4, &regValue);
            key_p->key.keyMaterial[8] = (UINT_8) ((regValue) & 0x000000FF);
            key_p->key.keyMaterial[9] = (UINT_8) ((regValue >> 8) & 0x000000FF);
            key_p->key.keyMaterial[10] = (UINT_8) ((regValue >> 16) & 0x000000FF);
            key_p->key.keyMaterial[11] = (UINT_8) ((regValue >> 24) & 0x000000FF);
            NIC_MCR_READ(adapter_p, MCR_SKADR5, &regValue);
            key_p->key.keyMaterial[12] = (UINT_8) ((regValue) & 0x000000FF);
            key_p->key.keyMaterial[13] = (UINT_8) ((regValue >> 8) & 0x000000FF);
            key_p->key.keyMaterial[14] = (UINT_8) ((regValue >> 16) & 0x000000FF);
            key_p->key.keyMaterial[15] = (UINT_8) ((regValue >> 24) & 0x000000FF);
        }
    }

    return TRUE;

} /* pauAccessHwKey */


/**_________________________________________________________________________
** pauCheckSKACRAccessStart
**
** description: This function check the Access Start bit of SKACR
** parameters:  adapter_p - Pointer to the adapter object data area.
**
** return:      TRUE, Ready to access command or previous access
**                    command completed.
**              FALSE, Previous access command cycle in progress.
** note:
**__________________________________________________________________________*/
BOOLEAN
pauCheckSKACRAccessStart(IN PMP_ADAPTER adapter_p)
{
    INT     i = 0;
    UINT_32   inProgress;

    DEBUGFUNC("pauCheckSKACRAccessStart");

    if (NIC_NOT_ACCESSIBLE(adapter_p)) {
        WARNLOG(("NIC is not accessible now\n"));
        return FALSE;
    }

    for (i = 0; i < 10; i++) {
        NIC_MCR_READ(adapter_p, MCR_SKACR, &inProgress);
        if (!(inProgress & SKACR_ACCESS_START)) {
            return TRUE;
        }
    }

    return FALSE;
}   /* End of pauCheckSKACRAccessStart */

/*_________________________________________________________________________
**  pauConfigureLED
**
**  descriptions: This routine is called to configure the LED pin.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      ledOutputMode - The new LED output mode to be set
**      ledNoTriggerSetting - The new LED setting when nothing triggers
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
pauConfigureLED(IN PMP_ADAPTER adapter_p,
                IN UINT_32 ledOutputMode,
                IN UINT_32 ledNoTriggerSetting)
{
    UINT_32 tmp32 = 0;

    DEBUGFUNC("pauConfigureLED");

    switch (ledOutputMode) {
        case NIC_LED_OUTPUT_TX:
             tmp32 = LCR_LED_MODE_TX;
             break;
        case NIC_LED_OUTPUT_RX:
             tmp32 = LCR_LED_MODE_RX;
             break;
        case NIC_LED_OUTPUT_TX_RX:
             tmp32 = LCR_LED_MODE_TX_RX;
             break;
        default:
             tmp32 = LCR_LED_MODE_GPIO;
    }

    if (ledNoTriggerSetting == NIC_LED_NO_TRIGGER_OFF) {
        tmp32 &= ~LCR_LED_POLARITY;
        tmp32 |= ((LED_BLINKING_ON_TIME_1 << 8) & LCR_LED_ON_CNT) |
            (LED_BLINKING_OFF_TIME_1 & LCR_LED_OFF_CNT);
    } else {
        tmp32 |= LCR_LED_POLARITY;
        tmp32 |= ((LED_BLINKING_ON_TIME_2 << 8) & LCR_LED_ON_CNT) |
            (LED_BLINKING_OFF_TIME_2 & LCR_LED_OFF_CNT);
    }

    NIC_MCR_WRITE(adapter_p, MCR_LCR, tmp32);

#if DBG
    NIC_MCR_READ(adapter_p, MCR_LCR, &tmp32);
    DBGLOG(HW, INFO, ("LCR = 0x%08x\n", tmp32));
#endif
}   /* pauConfigureLED */

