/*
** $Id: tsftimer.c 1.19 2004/10/04 18:11:54 cmchang Exp $
*/

/*************************************************************************
* Copyright (c) 2002 Inprocomm, Inc.
*
* All rights reserved. Copying, compilation, modification, distribution
* or any other use whatsoever of this material is strictly prohibited
* except in accordance with a Software License Agreement with
* Inprocomm, Inc.
**************************************************************************/
/*************************************************************************
GENERAL DESCRIPTION
********************
This file contains the MAC protocol timing control routines for INPROCOMM
802.11 Wireless LAN Adapters.
*/

/************************************************************************
*                E X T E R N A L   R E F E R E N C E S
*************************************************************************
*/
#include "precomp.h"

/************************************************************************
*                          C O N S T A N T S
*************************************************************************
*/

/************************************************************************
*                         D A T A   T Y P E S
*************************************************************************
*/

/************************************************************************
*                        P U B L I C   D A T A
*************************************************************************
*/

/************************************************************************
*                       P R I V A T E   D A T A
*************************************************************************
*/
DEBUGFILE("tsftimer.c");

/************************************************************************
*                             M A C R O S
*************************************************************************
*/

/************************************************************************
*              F U N C T I O N   D E C L A R A T I O N S
*************************************************************************
*/

/************************************************************************
*                          F U N C T I O N S
*************************************************************************
*/

/*_________________________________________________________________________
**  tsftimerInitializeTsfTimer
**
**  descriptions: This routine is called to initialize the TSF timer to 0.
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
tsftimerInitializeTsfTimer(IN PMP_ADAPTER adapter_p)
{
    NIC_MCR_WRITE(adapter_p, MCR_TTAR1, 0);
    NIC_MCR_WRITE(adapter_p, MCR_TTAR2, 0);
    NIC_MCR_WRITE(adapter_p, MCR_TTAR0, TTAR0_TSF_TIMER_VALUE_CHANGE);
}   /* tsftimerInitializeTsfTimer */

/*_________________________________________________________________________
**  tsftimerEnableTbttTimer
**
**  descriptions: This routine is called to enable the TBTT timer (event
**                generator).
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      nextTbttTime - The next TBTT timer event generation time, in units
**                     of TUs
**      tbttPeriod - The TBTT timer event generation period, in units of
**                   TUs
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
tsftimerEnableTbttTimer(IN PMP_ADAPTER adapter_p,
                        IN UINT_32 nextTbttTime,
                        IN UINT_32 tbttPeriod)
{
    UINT_32 mptcrValue;

    DEBUGFUNC("tsftimerEnableTbttTimer");

    NIC_MCR_READ(adapter_p, MCR_MPTCR, &mptcrValue);

    if (mptcrValue & (MPTCR_EN_COMP_NEXT_TBTT_TIME | MPTCR_TBTT_ENABLE)) {
        mptcrValue &= ~(MPTCR_EN_COMP_NEXT_TBTT_TIME | MPTCR_TBTT_ENABLE);

        NIC_MCR_WRITE(adapter_p, MCR_MPTCR, mptcrValue);
    }

    NIC_MCR_WRITE(adapter_p, MCR_NTTCR, (nextTbttTime &
        NTTCR_NEXT_TBTT_TIME) | (tbttPeriod << 16));

    mptcrValue |= MPTCR_EN_COMP_NEXT_TBTT_TIME | MPTCR_TBTT_ENABLE;

    NIC_MCR_WRITE(adapter_p, MCR_MPTCR, mptcrValue);
}   /* tsftimerEnableTbttTimer */

/*_________________________________________________________________________
**  tsftimerDisableTbttTimer
**
**  descriptions: This routine is called to disable the TBTT timer (event
**                generator).
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
tsftimerDisableTbttTimer(IN PMP_ADAPTER adapter_p)
{
    UINT_32 mptcrValue;

    DEBUGFUNC("tsftimerDisableTbttTimer");

    NIC_MCR_READ(adapter_p, MCR_MPTCR, &mptcrValue);

    mptcrValue &= ~(MPTCR_EN_COMP_NEXT_TBTT_TIME | MPTCR_TBTT_ENABLE);

    NIC_MCR_WRITE(adapter_p, MCR_MPTCR, mptcrValue);
}   /* tsftimerDisableTbttTimer */

/*_________________________________________________________________________
**  tsftimerEnablePreTbttTimer
**
**  descriptions: This routine is called to enable the pre-TBTT timer
**                (event generator).
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      nextPreTbttTime - The next pre-TBTT timer event generation time,
**                        in units of microseconds
**      preTbttPeriod - The pre-TBTT timer event generation period, in
**                      units of TUs
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
tsftimerEnablePreTbttTimer(IN PMP_ADAPTER adapter_p,
                           IN UINT_32 nextPreTbttTime,
                           IN UINT_32 preTbttPeriod)
{
    UINT_32 mptcrValue;

    DEBUGFUNC("tsftimerEnablePreTbttTimer");

    if (adapter_p->interruptEnable & IER_PRE_TBTT_INT) {
        NIC_MASK_INTERRUPT(adapter_p, IER_PRE_TBTT_INT);
    }

    NIC_MCR_READ(adapter_p, MCR_MPTCR, &mptcrValue);

    if (mptcrValue & MPTCR_EN_COMP_NEXT_PRE_TBTT_TIME) {
        mptcrValue &= ~MPTCR_EN_COMP_NEXT_PRE_TBTT_TIME;

        NIC_MCR_WRITE(adapter_p, MCR_MPTCR, mptcrValue);
    }

    NIC_MCR_WRITE(adapter_p, MCR_NPTCR, nextPreTbttTime &
        NPTCR_NEXT_PRE_TBTT_TIME);

    mptcrValue = (mptcrValue & ~MPTCR_PRE_TBTT_GEN_PERIOD) | (preTbttPeriod &
        MPTCR_PRE_TBTT_GEN_PERIOD) | MPTCR_EN_COMP_NEXT_PRE_TBTT_TIME;

    NIC_MCR_WRITE(adapter_p, MCR_MPTCR, mptcrValue);

    NIC_MCR_WRITE(adapter_p, MCR_ISAR, ISAR_PRE_TBTT_INT);

    NIC_UNMASK_INTERRUPT(adapter_p, IER_PRE_TBTT_INT);
}   /* tsftimerEnablePreTbttTimer */

/*_________________________________________________________________________
**  tsftimerDisablePreTbttTimer
**
**  descriptions: This routine is called to disable the pre-TBTT timer
**                (event generator).
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
tsftimerDisablePreTbttTimer(IN PMP_ADAPTER adapter_p)
{
    UINT_32 mptcrValue;

    DEBUGFUNC("tsftimerDisablePreTbttTimer");

    if (adapter_p->interruptEnable & IER_PRE_TBTT_INT) {
        NIC_MASK_INTERRUPT(adapter_p, IER_PRE_TBTT_INT);
    }

    NIC_MCR_READ(adapter_p, MCR_MPTCR, &mptcrValue);

    mptcrValue &= ~MPTCR_EN_COMP_NEXT_PRE_TBTT_TIME;

    NIC_MCR_WRITE(adapter_p, MCR_MPTCR, mptcrValue);
}   /* tsftimerDisablePreTbttTimer */

/*_________________________________________________________________________
**  tsftimerEnableDtimTimer
**
**  descriptions: This routine is called to enable the DTIM timer (event
**                generator).
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**      nextDtimTime - The next DTIM timer event generation time, in units
**                     of TUs
**      dtimPeriod - The DTIM timer event generation period, in units of
**                   TUs
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
tsftimerEnableDtimTimer(IN PMP_ADAPTER adapter_p,
                        IN UINT_32 nextDtimTime,
                        IN UINT_32 dtimPeriod)
{
    UINT_32 mptcrValue;

    DEBUGFUNC("tsftimerEnableDtimTimer");

    NIC_MCR_READ(adapter_p, MCR_MPTCR, &mptcrValue);

    if (mptcrValue & (MPTCR_EN_COMP_NEXT_DTIM_TIME | MPTCR_DTIM_ENABLE)) {
        mptcrValue &= ~(MPTCR_EN_COMP_NEXT_DTIM_TIME | MPTCR_DTIM_ENABLE);

        NIC_MCR_WRITE(adapter_p, MCR_MPTCR, mptcrValue);
    }

    NIC_MCR_WRITE(adapter_p, MCR_NDTCR, (nextDtimTime &
        NDTCR_NEXT_DTIM_TIME) | (dtimPeriod << 16));

    mptcrValue |= MPTCR_EN_COMP_NEXT_DTIM_TIME | MPTCR_DTIM_ENABLE;

    NIC_MCR_WRITE(adapter_p, MCR_MPTCR, mptcrValue);
}   /* tsftimerEnableDtimTimer */

/*_________________________________________________________________________
**  tsftimerDisableDtimTimer
**
**  descriptions: This routine is called to disable the DTIM timer (event
**                generator).
**  parameters:
**      adapter_p - Pointer to the Adapter structure
**  return:         (none)
**  note:
**__________________________________________________________________________*/
VOID
tsftimerDisableDtimTimer(IN PMP_ADAPTER adapter_p)
{
    UINT_32 mptcrValue;

    DEBUGFUNC("tsftimerDisableDtimTimer");

    NIC_MCR_READ(adapter_p, MCR_MPTCR, &mptcrValue);

    mptcrValue &= ~(MPTCR_EN_COMP_NEXT_DTIM_TIME | MPTCR_DTIM_ENABLE);

    NIC_MCR_WRITE(adapter_p, MCR_MPTCR, mptcrValue);
}   /* tsftimerDisableDtimTimer */

